(**************************************************************************
 * module BogenGenerator                                                  *
 **************************************************************************)
(** Stellt die Funktionen zur Verfuegung, die aus einer internen
 *  Grammatikrepraesentation vom Typ [BogenGrammar.t] einen Parser
 *  in Form von C-Programmcode erzeugen.
 *)


(**************************************************************************
 * Globale Deklarationen:                                                 *
 **************************************************************************)

(** Flag, ob die Kommandozeilenoption "-d" beim Programmaufruf angegeben
 *  wurde.
 *  (Erzeugt Datei "<Ausgabedateiname>.h" mit Tokentypen fuer Lex.)
 *)
let option_d = ref false


(** Stringrefernz fuer den Namen einer eventuellen Ausgabedatei fuer die
 *  Tokentypen, wie Lex sie benoetigt.
 *)
let token_filename =
  ref ""


(** Laenge des Lookahead. *)
let lookahead_length = ref 10


(** Zaehler fuer die Anzahl der bisher insgesamt erzeugen Items bzw.
 *  fuer die hoechste schon verbrauchte Itemnummer.
 *  Die Werte 0 und 1 sind reserviert fuer die Items \[S' -> .S\] und
 *  \[S' -> S.\], die fuer Beginn und Ende des Parsevorganges stehen.
 *)
let item_count = ref 1


(**************************************************************************
 * Datentypdeklarationen:                                                 *
 **************************************************************************)

(** Datentyp fuer Items mit folgenden Feldern:
 *  - [number]: Laufende Nummer, ueber die das Item angesprochen wird.
 *  - [dot_symbol]: Welches Symbol steht direkt hinter dem
 *    Bearbeitungspunkt?
 *  - [dot_pos] : Zu Diagnosezwecken Position des Bearbeitungspunktes.
 *  - [rule]: Zu Diagnosezwecken Stringrepraesentation der Grammatikregel,
 *    auf der das Item basiert.
 *  - [first_k]: FIRST_k-Menge der rechten Seite des Items,
 *    repraesentiert als Liste von Listen von Strings. Jede Liste steht
 *    dabei fuer ein Element der FIRST_k-Menge, jeder String in solch
 *    einer Liste steht fuer ein Terminalsymbol.
 *)
type item_t = {
  number      : int;
  dot_symbol  : BogenGrammar.symbol_t;
  dot_pos     : int;
  rule_number : int;
  rule        : string;
  first_k     : string list list
}


(**************************************************************************
 * Funktionen - Verwaltung der Terminal- und Nichtterminalsymbole:        *
 **************************************************************************)

(** Die Funktion [create_c_tokentypes ig] erzeugt einen C-Aufzaehlungstyp 
 *  fuer die Tokentypen in der Grammatik [ig].
 *  @param ig Interne Grammatikrepraesentation vom Typ [BogenGrammar.t].
 *  @return String mit C-Code.
 *)
let create_c_tokentypes ig =
  (* Die Hilfsfunktion [code_for_token key data code_so_far] erzeugt in
   * der Definition des C-Aufzaehlungstypen fuer die Terminalsymbole den
   * Eintrag fuer das Symbol [key], versehen mit der eindeutigen Nummer
   * [data]. Dahinter wird der String [code_so_far] mit dem bisher
   * generierten C-Code angehaengt.
   * @param key Name des Terminalsymbols, fuer das der Eintrag erzeugt 
   *  wird.
   * @param data Nummer des Terminalsymbols.
   * @param code_so_far Bisher generierter C-Code.
   * @return String aus C-Code.
   *)
  let code_for_token key data code_so_far =
    match code_so_far with
      "" ->
        (key ^ " = ") ^ (string_of_int data);
    | _ ->
        (key ^ " = ") ^ (string_of_int data) ^ ",\n     " ^ code_so_far

  in let token_string = Hashtbl.fold code_for_token ig.BogenGrammar.terminals ""

  in let out_string =
    "#ifndef YYTOKENTYPE\n" ^
    "# define YYTOKENTYPE\n" ^
    "   enum yytokentype {\n     " ^ token_string ^ "\n   };\n" ^
    "#endif\n"
  in out_string


(** [create_c_nonterminals ig] erzeugt aus der Nichtterminal-Hashtabelle
 *  von [ig] C-Code, der die Nichtterminalmenge der Grammatik
 *  repraesentiert.
 *  @param ig Aktuelle Grammatik vom Typ [BogenGrammar.t].
 *    Diese enthaelt unter anderem eine Hashtabelle fuer die
 *    Nichtterminalsymbole, in deren Schluesselfeldern die Symbole als
 *    String abgelegt sind, und in deren Datenfeldern fuer jedes Symbol
 *    abgelegt ist:
 *    - eine eindeutige Nummer fuer das Symbol und
 *    - nach Generierung der Items eine Liste von Nummern aller Items,
 *      die bei Expansion anhand dieses Symbols neu im Graphen entstehen
 *      koennen.
 *  @return Ein String aus C-Code fuer die Nichtterminalsymbole.
 *)
let create_c_nonterminals ig =

  (* Die Hilfsfunktion [item_arrays key data code_so_far] erzeugt fuer
   * das Nichtterminalsymbol [key] ein C-Array, das die Nummern
   * aller Items aufnimmt, die bei einer Expansion anhand [key] neu im
   * Graph entstehen koennen.
   * @param key Name des Nichtterminalsymbols, fuer das das Array erzeugt
   *  werden soll.
   * @param data Datenfeld des Nichtterminalsymbols vom Typ
   *  [BogenGrammar.nonterminal_t].
   * @param code_so_far Bisher generierter C-Code, der vor den neu
   *  erzeugte angehaengt wird.
   * @return String mit C-Code.
   *)
  let item_arrays key data code_so_far =
    "\n"
    ^ "int nt_items" ^ (string_of_int data.BogenGrammar.number) ^ "[] =\n"
    ^ "\t{ "
    ^ (List.fold_left
        ( fun items_so_far n -> (items_so_far ^ (string_of_int n) ^ ",") )
        ""
        data.BogenGrammar.items
      )
    ^ " };"
    ^ code_so_far

  (* Die Hilfsfunktion [add_to_nt_list] generiert aus den Dateneintraegen
   * der Nichtterminalsymbole zunaechst eine Liste, um diese nach Nummern
   * sortieren zu koennen.
   *
   * Der Aufruf [add_to_nt_list key data list_so_far] erzeugt fuer das
   * Nichtterminalsymbol [key] ein 4-Tupel aus
   * - Nummer der Symbols in der eindeutigen Nummerierung,
   * - Name des Symbols als String,
   * - Anzahl der Items \[A -> . \alpha\] mit dem Symbol [key] auf der
   *   linken Seite, konvertiert in einen String, und
   * - Name des Arrays, in dem die Nummern dieser Items abgelegt sind.
   * @param key Name des Nichtterminalsymbols als String.
   * @param data Dateneintrag des Symbols aus der Hashtabelle aller
   *  Nichtterminalsymbole.
   * @param list_so_far Bisherige Liste der Nichtterminaleintraege
   * @return Bisherige Liste der Nichtterminaleintraege, ergaenzt um
   *  einen Eintrag fuer das Symbol [key].
   *)
  and add_to_nt_list key data list_so_far =
    ( data.BogenGrammar.number,
      key,
      string_of_int (List.length data.BogenGrammar.items),
      "nt_items" ^ string_of_int data.BogenGrammar.number
    )
    :: list_so_far

  (* Hilfsfunktion zum Sortieren einer Liste von 4-Tupeln.
   *
   * Ein Aufruf [lsort liste] sortiert [liste] anhand des ersten
   * Tupelelements mit der Vergleichsfunktion [Pervasives.compare].
   * @param liste Zu sortierende Liste.
   * @return Sortierte Version von [liste].
   *)
  in let lsort = List.sort
    (fun (a1, _, _, _) (a2, _, _, _) -> compare a1 a2)
  
  (* Hilfsfunktion, um aus einer Liste von Nichtterminaleintraegen einen
   * String aus C-Code zu erzeugen.
   *
   * Ein Aufruf [nt_structs nt_list] sortiert die Liste [nt_list] zunaechst
   * nach der Symbolnummer. Dann wird erzeugt fuer jedes Listenelement
   * C-Code fuer einen Eintrag in das Array [nonterminals\[\]] erzeugt.
   * Arrayindex ist Symbolnummer - 1.
   * Inhalt eines Eintrages ist:
   * - ein Kommentar, der Nummer und Name des Symbols enthaelt,
   * - die Anzahl der Items, die bei der Expansion anhand dieses Symbols
   *   entstehen koennen, sowie
   * - ein Zeiger auf das Array dieser Items, wie es in der Funktion
   *   [item_arrays] generiert wird.
   * @param nt_list Liste von Nichtterminaleintraegen, wie sie von
   *  [add_nt_to_list] erzeugt werden.
   * @return String mit C-Code.
   *)
  in let nt_structs nt_list =
    List.fold_left
      ( fun code_so_far
          ( nt_num, nt_name, nt_item_count, nt_item_array ) ->
          code_so_far
          ^ "\n\t/* " ^ (string_of_int nt_num) ^ ": "  ^ nt_name ^ " */"
          ^ "\n\t{"
            ^ nt_item_count ^ ", "
            ^ nt_item_array ^ ", "
						^ "nt_name" ^ (string_of_int nt_num)
          ^ "},"
      )
      ""
      (lsort nt_list)

	(* Hilfsfunktion, um aus den Nichtterminaleintraegen C-Strings
	 * fuer die Namen der Nichtterminalsymbole zu erzeugen.
	 * @param, nt_list Liste von Nichtterminaleintraegen, wie sie
	 * 	von [add_nt_to_list] erzeugt werden.
	 * @return String mit C-Code.
	 *)
	in let nt_names nt_list =
		List.fold_left
			( fun code_so_far (nt_num, nt_name, _, _) ->
					code_so_far
					^ "char nt_name"
					^ (string_of_int nt_num) ^ "[] = { \""
					^ nt_name ^ "\" };\n"
			)
			""
			(lsort nt_list)

  in let out_string =

    (* Fuer jedes NT-Symbol wird zuerst das C-Array erzeugt, das die
     * Nachfolgeitems festlegt:
     *)
    (Hashtbl.fold item_arrays ig.BogenGrammar.nonterminals "\n\n")

    (* Datentypdeklaration fuer die Nichtterminalsymbole in C: *)
    ^ "\nstruct nonterminal_t {\n"
    ^ "\tint item_count;\n"
    ^ "\tint *items;\n"
		^ "\tchar *name;\n"
    ^ "};\n\n\n"

		(* Strings mit den Namen aller Nichtterminalsymbole: *)
		^ nt_names ( (Hashtbl.fold add_to_nt_list
									ig.BogenGrammar.nonterminals []) )
		^ "\n\n"

    (* Schliesslich wird das C-Array aller NT-Eintraege definiert: *)    
    ^ "struct nonterminal_t nonterminals[] = {"
    ^ (nt_structs ( Hashtbl.fold add_to_nt_list
                    ig.BogenGrammar.nonterminals [])
      )
    ^ "\n};\n"

  in
    out_string

(**************************************************************************
 * Funktionen - Hilfsfunktionen auf Listen:                               *
 **************************************************************************)

(** Der Funktionsaufruf [minlength l] berechnet fuer eine Liste [l] von
 *  Listen die Laenge der kuerzesten enthaltenen Liste.
 *  @param l Liste von Listen.
 *  @return Laenge der kuerzesten Listen in l.
 *)
let rec minlength l =
  match l with
    [] -> 0 (* Falls keine Elementliste existiert *)
  | a :: tail ->
      ( match tail with
          [] -> List.length a
        | _ -> min (List.length a) (minlength tail)
      )


(** [take n l] liefert exakt die ersten [n] Elemente der Liste [l].
 *  Sollte die Liste zu wenig Elemente enthalten, werfen [hd] oder [tl]
 *  eine Exception.
 *  @param n Laenge der Resultatliste.
 *  @param l Liste, aus der die Elemente genommen werden sollen.
 *  @return Liste mit den ersten n Elementen von [l].
  *)
let rec take n l =
  match n with
      0 -> 
        []
    | _ ->
        (List.hd l) :: ( take (n-1) (List.tl l))

    
(** [take_le n l] liefert die ersten [n] Elemente der Liste [l].
 *  Sollte die Liste weniger Elemente enthalten, wird die gesamte Liste
 *  zurueckgegeben.
 *  @param n Maximale Laenge der Resultatliste.
 *  @param l Liste, aus der die Elemente genommen werden sollen.
 *  @return Liste mit den ersten <= [n] Elementen von [l].
 *)
let rec take_le n l =
  match n with
      0 ->
        []
    | _ ->
        ( match l with
            [] -> l
          | _ ->
              (List.hd l) :: (take_le (n-1) (List.tl l))
        )


(** [take_uniq i l] nimmt eine Liste [l] von Liste und liefert eine neue
 *  Liste, in der alle Elementlisten aus bei Laenge i abgeschnitten sind.
 *  Das Resulat ist duplikatfrei.
 *  @param i Laenge der Elementlisten im Resultat.
 *  @param l Liste von Listen.
 *  @return Liste mit den Elementlisten von [l], abgeschnitten bei
 *    Laenge i und duplikatfrei.
 *)
let take_uniq i l =
  match l with
    [] -> []
  | _ ->
      ( match i with
          0 -> [[]]
        | _ ->
            let t = Hashtbl.create 7
            in begin
              List.iter (fun ll -> Hashtbl.replace t (take_le i ll) ()) l;
              Hashtbl.fold
                (fun key data init -> key :: init)
                t []; 
            end
      )
(* Frage: muss ich die Hashtafel vor Verlassen der Funktion noch loeschen,
 * oder geschieht das automatisch?
 *)


(** [remove n l] entfernt die ersten [n] Elemente der Liste [l].
 *  Sollte die Liste weniger als [n] Elemente enthalten, wird eine
 *  Exception geworfen.
 *  @param n Anzahl zu entfernender Elemente.
 *  @param l Liste, aus der die Elemente entfernt werden sollen.
 *  @return Liste [l] ohne die ersten [n] Elemente.
 *)
let rec remove n l =
  match n with
      0 ->
        l
  | _ ->
      (  match l with
           [] -> raise (Failure "BogenGenerator.remove")
         | _ ->
             remove (n-1) (List.tl l)
      )
   

(** Die Hilfsfunktion [oplus] vereinigt zwei Listen von Listen von Strings
 *  folgendermassen:
 *  [oplus l1 l2] konkateniert jede Subliste von l1 mit jeder Subliste von
 *  l2.
 *  @param l1 Erste zu vereinigende Liste.
 *  @param l2 Zweite zu vereinigende Liste.
 *  @return Neue Liste, in der jede Liste in [l1] mit jeder Liste in [l2]
 *    konkateniert ist.
 *)
let oplus l1 l2 =
match l1 with
    [] ->
    l2
  | _ ->
    List.flatten (
      List.map
        ( fun e1 -> List.map
            (fun e2 -> e1 @ e2)
            l2
        )
        l1
      )


(** [oplus_uniq i l1 l2] liefert eine Liste [l], in der alle Elementlisten
 *  aus [l1] mit allen Elementlisten aus [l2] konkateniert sind,
 *  wobei die Elementlisten von [l] bei einer Laenge von [i] abgeschnitten
 *  sind und [l] keine Duplikate enthaelt.
 *)
let oplus_uniq i l1 l2 =
  match l1 with
    [] ->
      take_uniq i l2
  | _ ->
      take_uniq i (oplus l1 (take_uniq (i - minlength l1) l2))


(**************************************************************************
 * Funktionen - Vorausberechnung der FIRST_k-Mengen:                      *
 **************************************************************************)

(** [compute_f0]: Hilfsfunktion zum Berechnen der F_0-Mengen fuer alle
 *  Nichtterminale im Initialisierungsteil von [compute_firstk_nt]:
 *  [compute_f0 ig f_table key data] fuegt zur Hashtabelle der F_0-Werte
 *  [f_table] fuer das Nichterminalsymbol [key] den Wert F_0(key) hinzu.
 *  @param ig Grammatik vom Typ [BogenGrammar.t].
 *  @param f_table Hashtabelle der F_i-Werte, in die die neu berechneten
 *    F_0-Mengen der Nichtterminalsymbole eingetragen werden.
 *  @param key Nichtterminalsymbol, fuer das die F_0-Menge berechnet werden
 *    soll.
 *  @param data Datenfeld im Hashtabelleneintrag fuer das Symbol [key].
 *  @return [()], da das Ergebnis direkt in die Hashtabelle [f_table]
 *    eingetragen wird.
 *)
let compute_f0 ig f_table key data =
  (* Mit der internen Hilfsfunktion [f0_for_rule f_list (lhs, rhs)] wird
   * getestet, ob die Grammatikregel [(lhs, rhs)] zur Vergroesserung der
   * Menge F_0(key), gespeichert in der Liste [f_list], beitragen kann.
   * @param f_list Bisheriges F_0(key)
   * @param rule  Aktuell betrachtete Grammatikregel der Form [(lhs, rhs)].
   * @return Modifizierte Liste fuer F_0(key)
   *) 
  let f0_for_rule f_list (num, lhs, rhs) =
    match lhs with
      (* Betrachte nur Regeln mit [key] als linker Seite: *)
      x when x = key ->
        let ok = ref true
        and j =
          if (List.length rhs < !lookahead_length) then
            (List.length rhs)
          else
            !lookahead_length;

        in begin
          (* Teste, ob die ersten j Symbole auf der rechten Regelseite
           * Terminale sind.
           *)
          List.iter
            (fun x -> match x with
              BogenGrammar.Nonterminal y -> ok := false
            | _ -> ())
            (take j rhs);

          (* Wenn die ersten j Symbole Terminale sind, kann die Liste
           * dieser Terminalsymbole zu F_0(key) hinzugenommen werden.
           * Ansonsten wird diese Regel verworfen und die bisherige
           * F_0-Liste unveraendert zurueckgegeben.
           *)
          if !ok then
            (List.rev_map BogenGrammar.string_of_symbol (take j rhs))
              :: f_list
          else
            f_list;
        end
    | _ ->
        f_list

  (* Die interne Funktion [f0_for_rule] wird fuer jede Produktion in
   * [ig.productions] aufgerufen. Resultat ist eine Liste von
   * Terminalsymbol-Listen, die die Menge F_0(key) repraesentiert.
   *)
  in let res = List.fold_left f0_for_rule [] ig.BogenGrammar.productions

  (* Trage die soeben berechnete F_0-Menge [res] in den Hashtabellen-Eintrag
   * des Symbols [key] ein:
   *)
  in
    Hashtbl.add f_table (BogenGrammar.Nonterminal key) (res, [])


(** [test_fi] ist eine Hilfsfunktion fuer [compute_fi]:
 *  [test_fi f_table x rhs_rest f_so_far] ueberprueft, ob sich die bisher
 *  bestimmte Folge [x] von Terminalsymbolen weiter zu einem Eintrag
 *  in F_i(A) fuer ein Nichtterminalsymbol A fortsetzen laesst, indem
 *  nacheinander alle Symbole auf der rechten Seite der aktuell betrachteten
 *  Regel fuer A beruecksichtigt werden.
 *  @param x Bisher hergeleiteter Teilstring
 *  @param rhs_rest Liste von Symbolen, deren F_{i-1}-Mengen noch auf ihren
 *    moeglichen Beitrag zu x untersucht werden muessen.
 *  @param f_so_far Liste mit den bisher erzeugten Eintraegen.
 *  @return Paar aus changeflag und, ggf. ergaenzter, Liste von herleitbaren
 *    Terminal-Listen.
 *)
let rec test_fi f_table x rhs_rest f_so_far =
  let changed = ref false
  and new_f = ref f_so_far in
    
    (* Falls der Rest der rechten Regelseite leer ist, ist x fertig
     * hergeleitet:
     *)
    if rhs_rest = [] then
      ( if List.mem x !new_f then
          ()
        else
          begin
            changed := true;
            new_f := x :: !new_f
          end;
      )

    (* Andernfalls sind die naechsten Symbole der Regel zu untersuchen: *)
    else
      ( let (_, all_y) = Hashtbl.find f_table (List.hd rhs_rest)
        in
          List.iter
            ( fun y ->

              (* Falls |xy| >= k, so ist der String fertig hergeleitet: *)
              if (List.length x + List.length y >= !lookahead_length)
              then
                let xy = take !lookahead_length (x @ y)
                in
                  ( if List.mem xy !new_f then
                      ()
                    else
                      begin
                        changed := true;
                        new_f := xy :: !new_f
                      end;
                  )

              (* Andernfalls muss das naechste Symbol in der Regel zur
               * Hilfe genommen werden:
               *)
              else
                let xy = x @ y
                in let (changed', new_f') =
                  test_fi f_table xy (List.tl rhs_rest) !new_f
                in
                  ( if changed' then
                      begin
                        changed := true;
                        new_f := new_f'
                      end
                  );
            )
            all_y
      );

    (!changed, !new_f)

  
(** [compute_fi]: Hilfsfunktion zur Berechnung der F'_i-Mengen in
  * [compute_firstk_nt]:
  * [compute_fi ig f_table nt_key old_f] bildet fuer das
  * Nichtterminalsymbol [nt_key] mittels der F_{i-1}-Mengen in [f_table]
  * aus der Menge F_{i-1}(A), repraesentiert durch [old_f], die neue Menge
  * F_i(A), die den Rueckgabewert der Funktion darstellt.
  * @param ig Grammatik vom Typ [BogenGrammar.t]
  * @param f_table Hashtabelle aller bisher berechneten F_i-Mengen.
  * @param nt_key Name des Nichtterminalsymbols, dessen F_i-Menge berechnet
  *   werden soll, als String.
  * @param old_f Liste, die F_{i-1}(A) darstellt.
  * @return Paar aus changeflag und einer Liste, die F_i(A) repraesentiert.
  *)
let compute_fi ig f_table nt_key old_f =
  let changed = ref false
  and new_f = ref old_f

  (* [rule_contrib] ist eine interne Hilfsfunktion, die den Beitrag der
   * uebergebenen Regel zur Resultatliste bestimmt. Dabei kommen nur solche
   * Regeln in Betracht, die das aktuelle Symbol [nt_key] auf der linken
   * Seite haben.
   *)
  in let rule_contrib (num, lhs, rhs) = 
    match lhs with
      lhs' when lhs' = nt_key ->

        (* Falls die betrachtete Regel die Form "A -> Epsilon" hat
         * und sich der Epsilon-Eintrag [] noch nicht in F_i(A)
         * befindet: Fuege ihn hinzu!
         *)
        if rhs = [] then
          ( if List.mem [] !new_f then
              ()
            else
              begin
                changed := true;
                new_f := [] :: !new_f
              end;
          )

        (* Andernfalls hat die Regel die Form A -> B_1...B_s.
         * Betrachte dann zunaechst alle Terminalstrings x \in
         * F_{i-1}(B_1):
         *)
        else
          let (_, all_x) = Hashtbl.find f_table (List.hd rhs)
          in
            List.iter
              ( fun x ->
                (* Falls |x| = k, so ist der String x schon fertig: *)
                if (List.length x = !lookahead_length) then
                  (* Result := Result \cup \{x\} *)
                  ( if List.mem x !new_f then
                      ()
                    else
                      begin
                        changed := true;
                        new_f := x :: !new_f
                      end;
                  )

                (* Ansonsten muss das naechste Symbol der Regel zur
                 * Hilfe genommen werden, was in der Hilfsfunktion
                 * [test_fi] geschieht:
                 *)
                else
                  (* Result := Result \cup TestF_{i-1}(x, B_2...B_s) *)
                  let (changed', new_f') =
                    test_fi f_table x (List.tl rhs) !new_f
                  in
                    ( if changed' then
                        begin
                          changed := true;
                          new_f := new_f'
                        end;
                    )
              )
              all_x;
    | _ ->
      ()
    
  in let _ =
    (* Fuer jede Regel [rule] in [ig.BogenGrammar.productions], die das
     * Symbol [nt_key] auf der linken Seite hat, berechne den Beitrag von
     * [rule] zur Resultatliste.
     *)
    List.iter
      rule_contrib
      ig.BogenGrammar.productions;

  in
    (!changed, !new_f)


(** [compute_firstk_nt ig] dient der Vorausberechnung der FIRST_k-Mengen
 *  fuer alle Nichtterminalsymbole der Grammatik [ig] mittels
 *  Fixpunktiteration.
 *  @param ig Grammatik vom Typ [BogenGrammar.t]
 *  @return [()], da die Ergebnisse der Berechnung direkt in die Hashtabelle
 *    [ig.nonterminals] eingetragen werden.
 *)
let compute_firstk_nt ig =
  (* Temporaerer Speicher fuer die jeweils aktuell benoetigten F_i- und
   * F_{i-1}-Mengen:
   *)
  let f_table =
    Hashtbl.create ( (Hashtbl.length ig.BogenGrammar.terminals) +
                     (Hashtbl.length ig.BogenGrammar.nonterminals) )

  (* Flag, ob sich im aktuellen Durchgang etwas an f_table geandert hat.
   * Falls ja, ist eine erneute Iteration des Verfahrens noetig, falls
   * nein, ist der Fixpunkt erreicht und somit sind alle FIRST_k-Menge
   * fertig bestimmt.
   *)
  and changed = ref true

  in begin
    (* Initialisierung: Berechne fuer alle Symbole die F_0-Werte.
     * Die berechneten Werte werden in den Datenfeldern der Tabelle
     * [f_table] nach folgendem Muster abgelegt: (F_i, F_{i-1}).
     * Jede Menge F_i wiederum ist durch eine Liste von Eintraegen
     * repraesentiert, von denen jeder fuer einen Terminalstring
     * steht.
     * Jeder dieser Eintraege ist als Liste von Strings implementiert, wobei
     * jeder String ein Terminalsymbol darstellt.
     *
     * Fuer a \in \Sigma: F_i(a) = \{a\} \forall i.
     * Fuer A \in N:
     *   F_0(A) = \{ x \in \Sigma^* | A \to x \beta in P, |x| = k
     *               oder |x| < k, \beta = \eps \}.
     *)
    Hashtbl.iter
      (fun key data ->
        Hashtbl.add f_table (BogenGrammar.Terminal key) ([[key]], [[key]])
      )
      ig.BogenGrammar.terminals;
    Hashtbl.iter
      (compute_f0 ig f_table)
      ig.BogenGrammar.nonterminals;

    (* Berechne via Fixpunktiteration basierend auf den soeben bestimmten
     * F_0-Mengen die F_i-Mengen der Nichtterminalsymbole, bis sich fuer
     * ein i keine Menge mehr veraendert.
     *)
    while !changed do
      changed := false;
      
      (* Kopiere fuer jedes Nichtterminalsymbol in [f_table] im Datenfeld
       * [(a, b)] den Inhalt von [a] nach [b] (dies ist F_{i-1}(A)).
       * Fuer Terminalsymbole muss nichts unternommen werden, da sich ihre
       * F_i-Mengen ja nicht aendern koennen.
       *)
      Hashtbl.iter
        (fun key (data1, _) ->
          match key with
            BogenGrammar.Nonterminal nt_key ->
              Hashtbl.replace f_table key (data1, data1)
          | _ ->
            ()
        )
        f_table;

      (* Fuer jedes Nichtterminalsymbol A, berechne die Menge
       * F'_i(A) = \{ x \in \Sigma^* | A \to B_1 B_2 \ldots B_s \in P und
       *              x \in F_{i-1}(B_1) \oplus_k F_{i-1}(B_2} \oplus_k
       *                \dots \oplus_k F_{i-1}(B_s) \}.
       * F_i(A) ergibt sich dann aus der Vereinigung von
       * F_{i-1}(A) mit F'_i(A).
       *)
      Hashtbl.iter
        (fun key (data1, data2) ->
          match key with

            (* Nur fuer Nichtterminalsymbole muss etwas berechnet werden: *)
            BogenGrammar.Nonterminal nt_key ->
              (* Im Feld data2 ist F_{i-1}(A) abgelegt.
               * Berechne nun F_i(A) in der Hilfsfunktion [compute_fi]:
               *)
              let (changeflag, data_new) =
                compute_fi ig f_table nt_key data2
              in
                (* Teste, ob sich F_i(A) im Vergleich zu F_{i-1}(A)
                 * veraendert hat. Setze dann changed auf true und speichere
                 * die modifizierte Menge F_i(A) in f_table:
                 *)
                if changeflag then begin
                  Hashtbl.replace f_table key (data_new, data2);
                  changed := true
                end

          | _ ->
            ()
        )
        f_table
    done;

    (* Uebertrage die berechneten FIRST_k-Mengen in die Hashtabelle der
     * Nichtterminalsymbole. Fuer Terminalsymbole ist keine Aktion noetig.
     *)
    Hashtbl.iter
      (fun key (data1, data2) ->
        match key with
          BogenGrammar.Nonterminal nt_key ->
            let nt_data = Hashtbl.find ig.BogenGrammar.nonterminals nt_key
            in
              Hashtbl.replace
                ig.BogenGrammar.nonterminals
                nt_key
                { BogenGrammar.number  = nt_data.BogenGrammar.number;
                  BogenGrammar.items   = nt_data.BogenGrammar.items;
                  BogenGrammar.first_k = data1
                }
        | _ ->
          ();
      )
      f_table;

    (* Gib den Speicher fuer die temporaere Hashtabelle f_table wieder frei
     *)
    Hashtbl.clear f_table
  end


(** Der Funktionsaufruf [first_k_string s nonterminals] bestimmt fuer die
 *  Symbolfolge, die durch die Liste [s] repraesentiert wird, die Menge
 *  FIRST_k(s).
 *  @param s Liste von Symbolen vom Typ [BogenGrammar.symbol_t].
 *  @param nonterminals Hashtabelle der Nichtterminalknoten, in der die
 *    FIRST_k-Werte fuer die einzelnen in [s] vorkommenden
 *    Nichtterminalsymbole nachgeschlagen werden.
 *  @return Liste von Listen aus Terminalsymbolen, wobei jede Subliste
 *    einen String aus FIRST_k(s) repraesentiert.
 *)
 let first_k_string s nonterminals =
  let rec first_i_string i s =
    match i with
      0 -> [[]]
    | _ ->
        ( match s with
            (BogenGrammar.Terminal t) :: tail ->
              oplus [[t]] (first_i_string (i-1) tail)
          | (BogenGrammar.Nonterminal nt) :: tail ->
              ( let f_nt =
                  take_uniq i
                    (Hashtbl.find nonterminals nt).BogenGrammar.first_k
                in
                oplus_uniq (i)
                  f_nt
                  (first_i_string (i - (minlength f_nt)) tail)
              )
          | _ -> [[]] (* BogenGrammar.Epsilon :: [] oder [] *)
        )

  in
  match s with
    (BogenGrammar.Terminal t) :: tail ->
      oplus [[t]] (first_i_string (!lookahead_length - 1) tail)
  | (BogenGrammar.Nonterminal nt) :: tail ->
      ( let f_nt = (Hashtbl.find nonterminals nt).BogenGrammar.first_k
        in
        oplus_uniq !lookahead_length
          f_nt
          (first_i_string (!lookahead_length - (minlength f_nt)) tail)
      )
  | _ -> [[]] (* BogenGrammar.Epsilon :: [] oder [] *)     

      (**************************************************************************
 * Funktionen - Erzeugung der Items:                                      *
 **************************************************************************)

(** [items_of_rule' nonterminals (num, lhs, rhs) pos] erzeugt fuer die
 *  Regel [(num, lhs, rhs)] dasjenige Item, das den Punkt auf der rechten
 *  Seite an Position [pos] hat, sowie rekursiv alle Items mit einer
 *  frueheren Position pos', 0 < pos' < pos des Punktes, bis 0.
 *  Die Teilergebnisse werden zu einer Resultatliste zusammengefasst.
 *  Dabei bedeutet [pos = 0], dass sich der Punkt vor dem ersten Symbol
 *  befindet, und [pos = (List.length rhs)], dass sich der Punkt hinter dem
 *  letzten Symbol befindet.
 *  Im Fall [pos = 0] wird ausserdem in der Nichtterminaltabelle
 *  [nonterminals] der Eintrag fuer das Symbol [lhs] ergaenzt, indem die
 *  Nummer des neu generierten Items dort vermerkt wird.
 *  @param nonterminals Hashtabelle der Nichtterminalsymbole mit Schluesseln
 *    vom Typ [string] und Eintraegen vom Typ [BogenGrammar.nonterminal_t].
 *  @param rule Regel der Form [(num, lhs, rhs)] vom Typ
 *    [Bogengrammar.rule_t], fuer die alle Items generiert werden sollen.
 *  @param pos Position, an der im aktuell zu generierenden Item der Punkt
 *    stehen soll.
 *  @return Liste von Items vom Typ [item_t].
 *)
let rec items_of_rule' nonterminals (num, lhs, rhs) pos =
  (* Die Hilfsfunktion [right rhs] bestimmt fuer ein Item (lhs, rhs)
   * die "rechte Seite". Der Fall rhs.hd = Epsilon sollte nicht eintreten,
   * da dies in der uebergeordneten Funktion abgefangen wird.
   *)
  let right rhs = 
    match rhs with
      (BogenGrammar.Nonterminal sym) :: tail ->
        tail
    | _ ->
        rhs

  in match pos with

    (* Im Fall [pos = 0] wird fuer eine Regel der Form "A -> \alpha" gerade
     * das erste Item \[A -> . \alpha\] generiert, das bei Expansion des
     * NT-Symbols A entsteht.
     * Also muss nicht nur das Item selbst erzeugt werden, sondern auch die
     * Nummer dieses Items in der Itemliste von A eingetragen werden.
     *)
    0 ->
      begin
        item_count := !item_count + 1;

        let nt_key = lhs
        in let nt_data = Hashtbl.find nonterminals nt_key
        in
          (* Als erstes der Eintrag in die Itemliste des NT-Symbols: *)
          Hashtbl.replace nonterminals nt_key
            {
              BogenGrammar.number = nt_data.BogenGrammar.number;
              BogenGrammar.items = (!item_count :: nt_data.BogenGrammar.items);
              BogenGrammar.first_k = nt_data.BogenGrammar.first_k
            };
          
          (* Dies ist das erste Item, das fuer diese Regel erzeugt wird.
           * Daher ist der Rueckgabewert eine Liste von Items, die nur
           * dieses eine Element enthaelt.
           *)
          if rhs = [] then
            [{
              number      = !item_count;
              dot_symbol  = BogenGrammar.Epsilon;
              dot_pos     = 0;
              rule_number = num;
              rule        = (BogenGrammar.string_of_rule (num, lhs, rhs));
              (* right([B->.]) = eps => FIRST_k(right([B->.])) = {eps} *)
              first_k     = [[]]
            }]
          else
            [{
              number      = !item_count;
              dot_symbol  = (List.hd rhs);
              dot_pos     = 0;
              rule_number = num;
              rule        = (BogenGrammar.string_of_rule (num, lhs, rhs));
              (* right(item) haengt davon ab, ob rhs mit einem Terminal-
               * oder Nichtterminalsymbol beginnt. Daher wird zunaechst
               * mit dem Funktionsaufruf [right rhs] die rechte Seite
               * von [item] bestimmt und dann auf das Ergebnis die Funktion
               * [first_k_string] angewandt:
               *)
              first_k     = first_k_string (right rhs) nonterminals
            }];
      end;

    (* Wenn pos = Laenge(rhs), so wird das letzte Item dem Form
     * \[A -> \alpha .\] angefordert.
     * Im rekursiven Aufruf werden zuerst alle vorherigen Items erzeugt und
     * dann dieses letzte (mit Symbol \eps hinter dem Punkt) hinten an die
     * Liste angehaengt.
     *)
  | n' when n' = List.length rhs ->
      let items_so_far = items_of_rule' nonterminals (num, lhs, rhs) (pos - 1)
      in begin
        item_count := !item_count + 1;
        items_so_far @
          [{
            number      = !item_count;
            dot_symbol  = BogenGrammar.Epsilon;
            dot_pos     = pos;
            rule_number = num;
            rule        = (BogenGrammar.string_of_rule (num, lhs, rhs));
            (* right([B->\alpha.]) = eps =>
             * FIRST_k(right([B->\alpha.])) = {eps}
             *)
            first_k     = [[]]
          }]
      end
      
    (* Fuer Werte von [pos] zwischen 0 und Laenge(rhs) wird analog zum Fall
     * [pos] = Laenge(rhs) verfahren, nur dass hinter dem Punkt nicht \eps
     * steht, sondern das Symbol, das in der Regel an Position [pos] steht.
     *)
  | _ ->
      let items_so_far = items_of_rule' nonterminals (num, lhs, rhs) (pos - 1)
      in begin
        item_count := !item_count + 1;
        items_so_far @
          [{
            number      = !item_count;
            dot_symbol  = (List.nth rhs pos);
            dot_pos     = pos;
            rule_number = num;
            rule        = (BogenGrammar.string_of_rule (num, lhs, rhs));
            (* right(item) haengt davon ab, ob hinter dem Punkt ein
             * Terminal- oder Nichtterminalsymbol steht.
             * Daher wird zunaechst mit dem Funktionsaufruf
             * [right (remove pos rhs)] die rechte Seite von [item]
             * bestimmt und dann auf das Ergebnis die Funktion
             * [first_k_string] angewandt:
             *)
            first_k     = first_k_string (right (remove pos rhs)) nonterminals
          }]
      end
      

(** [items_of_rule nonterminals items_so_far (num, lhs, rhs)] erzeugt eine
 *  Liste aller Items, die sich aus der aktuellen Regel [lhs, rhs] ergeben,
 *  indem die rekursive  Hilfsfunktion [items_of_rule'] mit dieser Regel
 *  aufgerufen wird.
 *  Die Hilfsfunktion erzeugt alle Items, die sich durch die moeglichen
 *  Positionen des Bearbeitungspunktes von 0 bis [List.length rhs] ergeben
 *  koennen.
 *  Das Ergebnis dieses Aufrufs wird mit der bisherigen Ergebnisliste
 *  zusammengefuegt.
 *  @param nonterminals Hashtabelle der Nichtterminalsymbole der Grammarik.
 *    Wird von [items_of_rule'] benoetigt, um Informationen zu den im Item
 *    enthaltenen Symbolen nachzuschlagen.
 *  @param items_so_far Liste der bisher generierten Items vom Typ [item_t].
 *  @param rule Regel der Form [(num, lhs, rhs)], fuer die jetzt alle Items
 *    generiert werden sollen.
 *  @return Liste [items_so_far], ergaenzt um alle Items, die aus der Regel
 *    [(num, lhs, rhs)] entstehen.
 *)
let items_of_rule nonterminals items_so_far (num, lhs, rhs) =
  items_so_far @ (items_of_rule' nonterminals (num, lhs, rhs) (List.length rhs))


(** [items ig] erzeugt eine Liste aller Items, die aus den Produktionen der
 *  Grammatik [ig] entstehen koennen.
 *  Besondere Items sind \[S' -> .S\] und \[S' -> S.\], die Beginn und Ende
 *  des Parse-Vorgangs markieren. Fuer sie sind die Nummern 0 und 1
 *  reserviert.
 *  @param ig Grammatik vom Typ [BogenGrammar.t].
 *  @return Liste von Items vom Typ [item_t], bestehend aus:
 *    - einer eindeutigen Nummer fuer das Item,
 *    - dem Symbol direkt hinter dem Punkt (Typ [BogenGrammar.symbol_t]),
 *    - Nummer und Stringrepraesentation der Regel, aus der das Item
 *      entstanden ist, und
 *    - der FIRST_k-Menge fuer die rechte Seite des Items.
 *)
let items ig =
  { 
    number      = 0;
    dot_symbol  = BogenGrammar.Nonterminal ig.BogenGrammar.startsymbol;
    dot_pos     = 0;
    rule_number = 0;
    rule        = "0: S' -> " ^ ig.BogenGrammar.startsymbol;
    first_k     = [[]];
  }
  ::
  { 
    number      = 1;
    dot_symbol  = BogenGrammar.Epsilon;
    dot_pos     = 1;
    rule_number = 0;
    rule        = "0: S' -> " ^ ig.BogenGrammar.startsymbol;
    first_k     = [[]];
  }
  ::
  (* Generiere fuer jede Regel in [ig.productions] mittels der Funktion
   * [items_of_rule] eine Liste der Items fuer diese Regel und haenge die
   * Ergebnislisten aneinander:
   *)
  List.fold_left (items_of_rule ig.BogenGrammar.nonterminals)
    [] ig.BogenGrammar.productions


(** [create_c_items ig item_list] erzeugt aus einer Liste [item_list] von
 *  Items den entsprechenden C-Code als String.
 *  @param ig Grammatik vom Typ [BogenGrammar.t].
 *  @param item_list Liste der Items.
 *  @return String mit C-Code.
 *)
let create_c_items ig item_list =

   (* Die Hilfsfunktion [first_k_to_c item] erzeugt fuer die FIRST_k-Menge
   * der rechten Seite des Items [item] C-Code fuer ein Array, das diese
   * Menge repraesentiert. Das Ergebnis wird an den String [code_so_far]
   * gehaengt, der den bisher generierten C-Code enthaelt.
   * @param code_so_far Bisher erzeugter C-Code fuer FIRST_k-Arrays.
   * @param item Das Item, fuer dessen rechte Seite das FIRST_k-Array
   *  erzeugt wird.
   * @return String mit C-Code.
   *)
  let first_k_to_c code_so_far item =
    code_so_far ^ "\n"
    ^ "unsigned long first_k_item_" ^ (string_of_int item.number)
    ^ "[][LOOKAHEAD_LENGTH] = {"

    (* Fuer jeden Eintrag der Liste item.first_k: erzeuge eine Codezeile *)
    ^ ( List.fold_left
        ( fun so_far1 elem ->
            
            so_far1

            (* Als C-Kommentar: Textuelle Darstellung der FIRST_k-Menge zu
             * Diagnosezwecken.
             *)
            ^ "\t// "
            ^ ( List.fold_left
                  ( fun so_far2 s ->
                      so_far2
                      ^ s
                      ^ ", "
                  )
                  ""
                  elem
              )
            ^ "\n"

            ^ "\t{ "
            ^ ( List.fold_left
                  ( fun so_far2 s ->
                      so_far2
                      ^ (string_of_int
                          (Hashtbl.find ig.BogenGrammar.terminals s)
                        )
                      ^ ", "
                  )
                  ""
                  elem
              )
            ^ " },\n"
        )
        "\n"
        item.first_k
      )
    ^ "};\n"

  (* Die Hilfsfunktion [array_entry_for_item item] erzeugt den eigentlichen
   * Eintrag fuer das aktuelle Item. Er besteht aus einer C-Struktur mit den
   * Feldern [dot_symbol_type] und [dot_symbol], die beschreiben, welches
   * Symbol sich im Item direkt hinter dem Punkt befindet, sowie einem
   * Zeiger auf ein zweidimensionales C-Array [first_k_item], das die
   * FIRST_k-Menge der rechten Seite des Items aufnimmt, und einem
   * Integereintrag fuer die Laenge dieses Arrays.
   * @param item Das Item, fuer das aktuell der Arrayeintrag erzeugt wird.
   * @return String mit C-Code fuer das aktuelle Item.
   *)
  in let array_entry_for_item item =
    let rnum = (string_of_int item.rule_number)
    and sym = item.dot_symbol
    in let first_k_name = "first_k_item_" ^ (string_of_int item.number)

		in match sym with
			BogenGrammar.Epsilon ->
				"{" ^ rnum ^ ", Epsilon, 0, "
          ^ (string_of_int (List.length item.first_k)) ^ ", "
					^	first_k_name ^ "},"
		| BogenGrammar.Nonterminal n ->
				let nt = Hashtbl.find ig.BogenGrammar.nonterminals n
				in let num = (string_of_int nt.BogenGrammar.number)
        in
          "{" ^ rnum ^ ", Nonterminal, " ^ num ^ ", "
						^ (string_of_int (List.length item.first_k))
						^ ", " ^ first_k_name ^ "},"
    | BogenGrammar.Terminal t ->
        let num = string_of_int (Hashtbl.find ig.BogenGrammar.terminals t)
        in 
          "{" ^ rnum ^ ", Terminal, " ^ num ^ ", "
						^ (string_of_int (List.length item.first_k))
						^ ", " ^ first_k_name ^ "},"
    
  (* Die Hilfsfunktion [item_to_c code_so_far item] erzeugt den Eintrag
   * fuer das Item [item] im C-Array und haengt das Ergebnis an den
   * String [code_so_far], der den bisher generierten C-Code enthaelt.
   * Die eigentliche Arbeit wird dabei in der Hilfsfunktion
   * [array_entry_for_item] verrichtet.
   * @param code_so_far Bisher erzeugter C-Code fuer Items.
   * @param item Das Item, fuer das aktuell C-Code erzeugt werden soll.
   * @return String aus C-Code.
   *)
  in let item_to_c code_so_far item =
    code_so_far ^ "\t"
    ^ "/* "
      ^ "Item " ^ (string_of_int item.number) ^ ": "
      ^ "\n\t   rule " ^ item.rule
      ^ ", \n\t   dot position: "
      ^ (string_of_int item.dot_pos)
      ^ "; dot symbol: "
      ^ (BogenGrammar.string_of_symbol item.dot_symbol)
    ^ " */\n\t" 
    ^ (array_entry_for_item item)
    ^ "\n\n"

  (* Zusammensetzen des Ausgabestrings von C-Code fuer die Items aus
   * den Resultaten der Hilfsfunktionen:
   *)
  in let out_string =
    (* Fuer jedes Item: das Array mit der FIRST_k-Menge seiner rechten
     * Seite:
     *)
    (List.fold_left first_k_to_c "" item_list)
    ^ "\n"

    (* Deklaration der noetigen Datenstrukturen: *)
    ^ "\nenum symbol_type {\n\tEpsilon, Nonterminal, Terminal\n};\n\n\n"
    ^ "struct item_t {\n"
    ^ "\tint rule_number;\n"
    ^ "\tenum symbol_type dot_symbol_type;\n"
    ^ "\tunsigned long dot_symbol;\n"
		^ "\tint first_k_count;\n"
    ^ "\tunsigned long (*first_k_item)[LOOKAHEAD_LENGTH];\n"
    ^ "};\n\n\n"

    (* Initialisierung des Arrays fuer die Items: *)
    ^ "struct item_t items[] = {\n"

    (* Fuer jedes Element der Item-Liste, liefere C-Code, der das Item 
     * repraesentiert!
     *)
    ^ (List.fold_left item_to_c "\n " item_list)

    (* Schliesse das Item-Array: *)
    ^ "};\n\n"

  in out_string


(**************************************************************************
 * Funktionen - Erzeugung der C-Strukturen fuer die Grammatikregeln:      *
 **************************************************************************)

(** [create_c_rules ig] erzeugt C-Code, der die Regelmenge der Grammatik
 *  [ig] repraesentiert.
 *  @param ig Die Grammatik vom Typ [BogenGrammar.t], deren Regelmenge
 *  dargestellt werden soll.
 *  @return String mit dem entsprechenden C-Code.
 *)
let create_c_rules ig =
	let string_of_rhs_sym code_so_far sym =
		match sym with
			BogenGrammar.Epsilon -> "0, " ^ code_so_far
		| BogenGrammar.Nonterminal n ->
				(* Nichtterminalsymbole werden negiert und vom Ergebnis noch 1
				 * subtrahiert, um mit dem Ergebnis des C-Operators [~]
				 * kompatibel zu sein!
				 *)
				code_so_far
				^ (string_of_int
					(-(Hashtbl.find ig.BogenGrammar.nonterminals n).BogenGrammar.number
						- 1)
				)
				^ ", "
		| BogenGrammar.Terminal t ->
				code_so_far
				^ (string_of_int (Hashtbl.find ig.BogenGrammar.terminals t))
				^ ", "
		
	(* Fuer jede Regel, erzeuge ein Array, das die Symbole der rechten
	 * Regelseite enthaelt:
	 *)
	in let string_of_rhs code_so_far (num, lhs, rhs) =
		code_so_far
		^ "\nsigned long rule_rhs" ^ (string_of_int num) ^ "[] = {\n\t"
		^ (List.fold_left string_of_rhs_sym "" rhs)
		^ "\n};\n"

	(* Fuer jede Regel, erzeuge einen Eintrag im Array [rules], bestehend
	 * aus:
	 * - numerischer Codierung des Symbols auf der linken Regelseite,
	 * - Anzahl der Symbole auf der rechten Regelseite und
	 * - Zeiger auf das Array, das die rechte Regelseite enthaelt.
	 *)
	in let rule_entry code_so_far (num, lhs, rhs) =
		code_so_far ^ "\t{"
		^ (string_of_int (Hashtbl.find ig.BogenGrammar.nonterminals lhs).BogenGrammar.number)
		^ ", " ^ (string_of_int (List.length rhs))
		^ ", rule_rhs" ^ (string_of_int num)
		^ "},\n\n"

	in
		"\n#ifdef PARSE_TREE"
		^ "\nstruct rule_t {"
		^ "\n\t// Nummer des Nichtterminalsymbols auf der linken Regelseite:"
		^ "\n\tsigned long lhs;"
		^ "\n\t// Anzahl der Symbole auf der rechten Regelseite:"
		^ "\n\tint rhs_count;"
		^ "\n\t// Zeiger auf ein Array mit den Symbolen der rechten Regelseite."
		^ "\n\t// (Terminalsymbole > 0, Nichtterminalsymbole < 0."
		^ "\n\tsigned long *rhs;"
		^ "\n};\n\n"
		^ List.fold_left string_of_rhs "" ig.BogenGrammar.productions
		^ "\n\nstruct rule_t rules[] = {\n"
		^ List.fold_left rule_entry "" ig.BogenGrammar.productions
		^ "};\n"
		^ "\n#endif\n\n\n"

(**************************************************************************
 * Funktionen - Ausgabe des Resultatparsers:                              *
 **************************************************************************) 

(** [create_headerfile token_string] erzeugt eine C-Quellcodedatei
 *  <token_filename> mit den Tokentyp-Deklarationen, die im String
 *  [token_string] als C-Code abgelegt sind.
 *  Benoetigt wird dies bspw., um dem Scannergenerator (f)lex die
 *  Tokentypen bekannt zu machen, die er an den Parser liefern soll.
 *  @param token_string String mit C-Code fuer die Tokentyp-
 *    Deklarationen.
 *  @return [()]
 *)
let create_headerfile token_string = 
  let oc2 = 
  try
    if Sys.file_exists !token_filename
    then Sys.remove !token_filename;
    open_out !token_filename;
  with
    Sys_error e -> begin
      print_endline "FEHLER:";
      print_endline
        ("Ausgabedatei " ^ !token_filename ^
          " konnte nicht geschrieben werden:");
      print_endline e;
      exit 1
    end
  in begin
    output_string oc2 token_string;
    flush oc2;
    close_out oc2;
    print_endline ("bogenlr: C-Code für Tokentypen wurde in Datei "
      ^ !token_filename ^ " geschrieben.");
  end


(** [copy_bogenfix oc] kopiert den fixen Anteil des Parsers, der sich in
 *  der Datei "bogenfix.c" befindet, in die Ausgabedatei [oc].
 *	@param oc Ausgabekanal
 *	@return [()]
 *)
let copy_bogenfix oc =
	let ic =
	try
		open_in "bogenfix.c"
	with _ ->
		begin
      print_endline "bogenlr: FEHLER:";
      print_endline ("bogenfix.c konnte nicht geöffnet werden!");
      exit 1
    end
	in let eof = ref false
  in begin
		output_string oc "\n/* ===== Begin of fixed part (bogenfix.c) ===== */\n";
		try
			while not !eof do
				let c = input_char ic in
					output_char oc c
			done
		with End_of_file ->
			output_string oc "\n/* ===== End of fixed part (bogenfix.c) ===== */\n";
			eof := true;
			flush oc
	end


(** [create_parser ig oc] erzeugt aus der internen Darstellung [ig] einer
 *  LR(k)-Grammatik den C-Code fuer den Resultatparser und schreibt ihn in
 *  den Ausgabekanal [oc].
 *	@param ig Eingabegrammatik von Typ [BogenGrammar.t].
 *	@param oc Ausgabekanal fuer den erzeugten Parser.
 *	@return [()]
 *)
let create_parser ig oc =

  (* Erzeuge C-Code fuer einen Aufzaehlungstyp fuer die Tokentypen in der
   * Grammatik ig:
   *)
  let token_string = create_c_tokentypes ig

  (* Bestimme die FIRST_k-Mengen fuer alle Nichtterminalsymbole: *)
  in let _ = compute_firstk_nt ig

  (* Erzeuge C-Code fuer alle Items, die auftreten koennen: *)
  in let item_string = create_c_items ig (items ig)

  (* Erzeuge C-Code fuer die Nichtterminalsymbole in der Grammatik ig.
   * Dies kann erst jetzt, also nach der Erzeugung aller Items geschehen,
   * da erst dann in der Hashtabelle der Nichtterminalsymbole auch die
   * Information eingetragen wurde, welche Items bei einer Expansion anhand
   * eines bestimmten Nichtterminalsymbols entstehen koennen.
   *)
  in let nt_string = create_c_nonterminals ig

	(* Erzeuge C-Code fuer die Regeln der Grammatik ig. Diese Informationen
	 * werden benoetigt, falls per Compilerschalter [-DPARSE_TREE] beim
   * uebersetzen des Resultatparsers die Erzeugung von Ableitungsbaeumen
	 * angefordert wird.
	 *)
	in let rules_string = create_c_rules ig

  in begin
    output_string oc ("\n\n/* ===== Begin of generated code ===== */\n");

    (* Kopiere die Konstantendeklaration fuer die Laenge des Lookahead in
     * die Zieldatei (aktuell auf 10 festgelegt):
     *)
    output_string oc ("\n/* Lookahead-Laenge: */\n");
    output_string oc ("\n#define LOOKAHEAD_LENGTH " ^
      (string_of_int !lookahead_length)^ "\n\n");

    (* Kopiere die Deklaration der Tokentypen in die Zieldatei.
     * Falls die Kommandozeilenoption "-d" angegeben wurde: Kopiere die 
     * Tokentypen auch in eine zusaetzliche Headerdatei namens
     * <token_filename>,
     * bpsw. zum Gebrauch mit dem Scannergenerator (f)lex.
     *)
    output_string oc ("\n/* Tokentypen: */\n");
    output_string oc ("\n" ^ token_string);
    if !option_d then
      create_headerfile token_string;

    (* Kopiere den Code fuer die Nichtterminalsymbole und die Items in die
     * Zieldatei:
     *)
    output_string oc ("\n\n/* Nichtterminale: */\n");
    output_string oc nt_string;
    output_string oc ("\n\n/* Items: */\n");
    output_string oc item_string;
 
		(* Kopiere den Code fuer die Regelmenge in die Zieldatei: *)
		output_string oc ("\n\n/* Regeln: */\n");
		output_string oc rules_string;

    output_string oc ("/* ===== End of generated code ===== */\n\n\n");

    (* Kopiere das fixe Parserskelett in die Zieldatei: *)
	  copy_bogenfix oc

  end
