(**************************************************************************
 * module BogenGrammar                                                    *
 **************************************************************************)

(** Datenstrukturen fuer die interne Darstellung von kontextfreien
 *  Grammatiken und einige Operationen darauf.
 *)


(**************************************************************************
 * Exceptions:                                                            *
 **************************************************************************)

(** Ein Terminalsymbol darf innerhalb einer Grammatikbeschreibung nicht
 *  mehrfach deklariert sein.
 *)
exception Token_already_exists of string

(** Die Deklaration mehrerer Startsymbole ist nicht vorgesehen. *)
exception Startsymbol_already_exists

(** Ein Terminalsymbol darf nicht auf der linken Seite einer Grammatikregel
 *  auftreten.
 *)
exception Token_as_lhs of string

(** Wenn ein Startsymbol explizit deklariert wird, muss es auch eine Regel
 *  dafuer geben.
 *)
exception No_rule_for_startsymbol of string

(** Jedes Nichtterminal, das durch eine Regel deklariert ist, sollte auch
 *  erreichbar sein, d.h. auf der rechten Seite einer Produktion auftreten.
 *)
exception Unreachable_nonterminal

(** Jedes in der Grammatik benutzte Nichtterminal muss durch eine Regel
 *  definiert werden.
 *)
exception Undefined_nonterminal


(**************************************************************************
 * Datentypdeklarationen:                                                 *
 **************************************************************************)

(** Datentyp fuer die Eintraege in der Hashtabelle der
 *  Nichtterminalsymbole.
 *)
type nonterminal_t = {
  number  : int; (** Laufende Nummer des Symbols *)
  items   : int list;
    (** Liste von Nummern derjenigen Items \[A -> . \alpha_j\], die bei
     *  Expansion anhand A die Folgeknoten im Graphen werden.
     *)
  first_k : string list list
    (** Die FIRST_k-Menge des Symbols, repraesentiert als Liste von Listen
     *  von Strings. Dabei enthaelt jeder String die Interdarstellung fuer
     *  ein Terminalsymbol.
     *)
}


(** Der Tabelleneintrag fuer ein Terminalsymbol besteht aus seiner
 *  Nummer in einer bei 258 beginnenden fortlaufenden Nummerierung.
 *)
type terminal_t = int


(** Nichtterminal- und Terminalsymbole der Grammatik werden in der internen
 *  Darstellung fuer Grammatikregeln durch Strings repraesentiert, mit
 *  vorangestelltem Kontruktor [Nonterminal] oder [Terminal], der den
 *  Symboltyp kennzeichnet.
 *  Das leere Wort wird in Items durch den Konstruktor [Epsilon]
 *  dargestellt.
 *)
type symbol_t = Nonterminal of string | Terminal of string | Epsilon


(** Eine einzelne Grammatikregel wird dargestellt als Tripel aus:
 *  - Nummer der Regel
 *  - linke Seite = Nichtterminalsymbol als String
 *  - rechte Seite = Folge von Terminal- und Nichtterminalsymbolen
 *  Produktionen mit mehreren Alternativen werden intern als einzelne
 *  Regeln gespeichert, d.h. es kann mehrere Regeln mit identischer linker
 *  Seite geben.
 *)
type rule_t = int * string * symbol_t list


(** Ein Element vom Typ [BogenGrammar.t] enthaelt eine interne
 *  Repraesentation der Eingabegrammatik, bestehend aus:
 *  - Menge der Nichtterminalsymbole als Hashtabelle,
 *  - Menge der Terminalsymbole als Hashtabelle,
 *  - Liste der Produktionen,
 *  - Startsymbol als String,
 *  - Positionen von Anfang und Ende der Headers und Anfang des
 *    Trailers in der Eingabedatei, jeweils als Integer.
 *)
type t = {
  nonterminals  : (string, nonterminal_t) Hashtbl.t;
  terminals     : (string, terminal_t) Hashtbl.t;
  productions   : rule_t list;
  startsymbol   : string;
  header_start  : int;
  header_end    : int;
  trailer_start : int
}


(**************************************************************************
 * Ausgabe-Funktionen:                                                    *
 **************************************************************************)

(** Hilfsfunktion, um aus einem Grammatiksymbol vom Typ [symbol_t] einen
 *  String zu erzeugen.
 *
 *  [string_of_symbol s] liefert die Stringrepraesentation von [s].
 *  @param s Symbol vom Typ [symbol_t]
 *  @return Stringrepraesentation des Symbols
 *)
let string_of_symbol s =
  match s with
    Nonterminal n -> n
  | Terminal t -> t
  | Epsilon -> "EPSILON"


(** Hilfsfunktion zur einfachen textuellen Ausgabe der Hashtabelle aller
 *  Nichtterminalsymbole.
 *
 *  [string_of_nt_tbl tbl sep] fuegt fuer jeden Eintrag in [tbl] im
 *  Ergebnisstring einen Teilstring der Form
 *  "<Schluessel> = <Stringdarstellung des Datenfeldes>" an.
 *  @param tbl Eine Hashtabelle, deren Schluesselfelder vom Typ [string]
 *    und die Datenfelder vom Typ [nonterminal_t] sind.
 *  @param sep Separator, der im Ergebnisstring zwischen den einzelnen
 *    Schluesselwerten eingefuegt wird.
 *  @return String, der Stringrepraesentationen aller Schluessel- und der
 *    zugehoerigen Datenfelder der Hashtabelle enthaelt, getrennt durch
 *    [sep].
 *)
let string_of_nt_tbl tbl sep =
  let f a b rest =
    match rest with
      "" ->
        a ^ " = (" ^ (string_of_int b.number) ^ " [" ^
          (List.fold_left
            (fun x y -> (string_of_int y) ^ "; " ^ x)
            "])"
            b.items)
    | _ ->
        a ^ " = (" ^ (string_of_int b.number) ^ " [" ^
          (List.fold_left
            (fun x y -> (string_of_int y) ^ "; " ^ x)
            "])"
            b.items)
          ^ sep ^ rest
  in Hashtbl.fold f tbl ""


(** Hilfsfunktion zur einfachen textuellen Ausgabe der Hashtabelle aller
 *  Tokens.
 *
 *  [string_of_token_tbl tbl sep] verwirft fuer jeden Eintrag in [tbl] das
 *  Datenfeld und gibt nur den Schluessel aus.
 *  @param tbl Eine Hashtabelle, deren Schluesselfelder vom Typ [string]
 *    sind.
 *  @param sep Separator, der im Ergebnisstring zwischen den einzelnen
 *    Schluesselwerten eingefuegt wird.
 *  @return String, der alle Schluesselwerte der Hashtabelle enthaelt,
 *    getrennt durch [sep].
 *)
let string_of_token_tbl tbl sep =
  let f a b rest =
    match rest with
      "" -> a ^ " = " ^ (string_of_int b)
    | _ -> a ^ " = " ^ (string_of_int b) ^ sep ^ rest
  in Hashtbl.fold f tbl ""


(** Hilfsfunktion zur textuellen Ausgabe einer Grammatikregel.
 *
 *  [string_of_rule (num, lhs, rhs)] liefert einen String, der die
 *  uebergebene Regel repraesentiert.
 *  @param num Laufende Nummer der Regel.
 *  @param lhs Linke Seite der Grammatikregel, gespeichert als String.
 *  @param rhs Rechte Seite der Grammatikregel, gespeichert als Liste von
 *    Symbolen.
 *  @return String der Form
 *    "<num>: <lhs> -> <rhs als leerzeichen-separierter String>".
 *)
let string_of_rule (num, lhs, rhs) =
  let f a b = " " ^ (string_of_symbol a) ^ b
  in (string_of_int num) ^ ": " ^ lhs ^ " ->" ^ (List.fold_right f rhs "" )


(** [string_of_grammar g] erzeugt eine einfache Stringrepraesentation
 *  der internen Grammatikrepraesentation [g].
 *  @param g Grammatik vom Typ [BogenGrammar.t]
 *  @return Stringdarstellung der Grammatik.
 *)
let string_of_grammar g =
  let nts = string_of_nt_tbl g.nonterminals "\n "
  and ntc = string_of_int (Hashtbl.length g.nonterminals)
  and ts = string_of_token_tbl g.terminals "\n "
  and tc = string_of_int (Hashtbl.length g.terminals)
  and ps = List.fold_left
    (fun x y -> x ^ (string_of_rule y) ^ "\n")
    ""
    g.productions
  and pc = string_of_int (List.length g.productions)

  in let result =
    "NICHTTERMINALE (" ^ ntc ^ "):\n " ^ nts ^ "\n"
    ^ "\nTERMINALE (" ^ tc ^ "):\n " ^ ts ^ "\n"
    ^ "\nPRODUKTIONEN (" ^ pc ^ "):\n" ^ ps
    ^ "\nSTARTSYMBOL: " ^ g.startsymbol ^ "\n"
    ^ "\nHEADER: "
      ^ (string_of_int g.header_start)
      ^ " - "
      ^ (string_of_int g.header_end)
      ^ "\n"
    ^ "\nTRAILER: "
      ^ (string_of_int g.trailer_start)
  in
    result


(**************************************************************************
 * Test-Funktionen:                                                       *
 **************************************************************************)

(** [exists_rule_for_symbol g s] testet, ob die Grammatikspezifikation [g]
 *  eine Regel fuer das Nichtterminalsymbol enthaelt, das durch den String
 *  [s] repraesentiert wird.
 *  @param g Grammatik vom Typ [BogenGrammar.t]
 *  @param s Terminal- oder Nichtterminalsymbol als String
 *  @return [true], falls es in der Grammatik eine Regel fuer [s] gibt;
 *    sonst [false].
 *)
let exists_rule_for_symbol g s =
  let lhs (num, a, b) = a
  in List.exists (fun x -> s = (lhs x)) g.productions


(** [all_nonterminals_reachable g] testet, ob alle in der
 *  Grammatikspezifikation [g] deklarierten Nichtterminale erreichbar sind,
 *  d.h. ob sie auch auf der rechten Seite von Regeln	vorkommen.
 *	Eine Ausnahme stellt das Startsymbol der Grammatik dar, dieses muss
 *	nicht auf einer rechten Regelseite benutzt werden.
 *  @param g Grammatik vom Typ [BogenGrammar.t]
 *  @return [true], falls alle Nichtterminale erreichbar sind;
 *    sonst [false].
 *)
let all_nonterminals_reachable g =
  let rhs (num, a, b) = b

  (* Das Praedikat [symbol_in_rhs rule symbol] ist wahr, g.d.w. das Symbol
   * [sym] auf der rechten Seite der Regel [rule] vorkommt:
   *)
  in let symbol_in_rhs rule sym =
    List.exists (fun x -> (string_of_symbol x) = sym) (rhs rule)

  (* Die Funktion [f start key data init] verwirft fuer jeden ihr
	 * uebergebenen Eintrag der Hashtabelle das Datenfeld [data].
	 * Sie prueft, ob es fuer [key] mindestens eine Regel in der Liste
	 * [g.productions] gibt, die es auf seiner rechten Seite enthaelt.
   * Sollte es ein Nichtterminal ungleich dem Startsymbol geben, das nicht
	 * benutzt wird, ist der Rueckgabewert [false].
	 * Falls [key] das Startsymbol [start] der Grammatik ist, lautet der
	 * Rueckgabewert [true].
   *)
	in let f start key data init =
		if (key = start) then
			true
		else
			(List.exists (fun x -> symbol_in_rhs x key) g.productions) && init

  in Hashtbl.fold (f g.startsymbol) g.nonterminals true


(** [all_nonterminals_defined g] testet, ob alle in der Grammatik [g]
 *  auftretenden Nichtterminale durch eine Regel definiert sind.
 *  @param g Grammatik vom Typ [BogenGrammar.t]
 *  @return [true], falls alle Nichtterminalsymbole, die auf rechten Seiten
 *    von Regeln auftreten, durch Regeln definiert sind; sonst [false].
 *)
let all_nonterminals_defined g =
  let lhs (num, a, b) = a

  (* Die Funktion [f key data init] verwirft fuer jeden Eintrag der
   * Hashtabelle das Datenfeld [data]. Sie prueft, ob es fuer [key]
   * mindestens eine Regel in der Liste [g.productions] gibt, die es auf
   * seiner linken Seite enthaelt.
   * Sollte es ein Nichtterminal geben, das nicht durch eine Regel
   * deklariert ist, ist der Rueckgabewert [false].
   *)
  in let f key data init =
    (List.exists (fun x -> (lhs x) = key) g.productions) && init

  in Hashtbl.fold f g.nonterminals true


(** [test g] testet einige einfache Kriterien, um festzustellen, ob die
 *  Grammatik [g] eine gueltige Grammatikspezifikation sein kann:
 *  - Gibt es eine Regel fuer das im Deklarationsteil festgelegte
 *    Startsymbol?
 *  - Werden alle Nichtterminale erreicht, oder gibt es solche, fuer die
 *    zwar Regeln vorhanden sind, die aber auf keiner rechten Seite
 *    auftreten? Dies ist nur fuer das Startsymbol erlaubt.
 *  - Sind alle Nichtterminale, die auf rechten Seiten von Regeln
 *    auftreten, durch eigene Regeln definiert?
 *  @param g Grammatik vom Typ [BogenGrammar.t]
 *  @raise No_rule_for_startsymbol
 *    Im Deklarationsteil wurde ein Startsymbol festgelegt, aber es gibt
 *    keine Regel fuer dieses Symbol.
 *  @raise Unreachable_nonterminal
 *    Es gibt ausser dem Startsymbol mindestens ein Nichtterminalsymbol,
 *		das durch eine Regel definiert wird, aber niemals auf der rechten
 *		Seite einer Regel auftritt.
 *  @raise Undefined_nonterminal
 *    Es gibt mindestens ein Nichtterminalsymbol, das auf der rechten
 *    Seite einer Regel benutzt wird, das aber durch keine eigene Regel
 *    definiert wird.
 *)
let test g =
  begin
    if not (exists_rule_for_symbol g g.startsymbol) then
      raise (No_rule_for_startsymbol g.startsymbol);
    if not (all_nonterminals_reachable g) then
      raise Unreachable_nonterminal;
    if not (all_nonterminals_defined g) then
      raise Undefined_nonterminal;
  end
