/**************************************************************************/
/** \file bogenfix.c
 *
 *	\brief Parserskelett für BoGenLR (Bonn Generator for LR(k) Parsers).
 *
 *	In dieser Datei ist der C-Code für den feststehenden Teil des
 *	Resultatparsers festgelegt, d.h. die Datenstrukturen und Funktionen,
 *	die von einer konkreten Grammatik unabhängig sind.
 *	Sämtlicher weiterer Code wird anhand der Grammatikdatei erstellt,
 *	die dem Programm \c bogenlr als Eingabe übergeben wird.
 *
 *	Einige Makros, die in der Grammatikdatei oder beim Compileraufruf
 *	definiert werden können, und ihre Effekte:
 *	- USE_DEBUG_PRINTF: Erzeugt ausführliche Ausgaben zum Parsevorgang
 *		auf stderr.
 *	- DEBUG_ALL: Aktiviert erweiterte DEBUG-Ausgaben und fügt einigen
 *		Funktionen Integritätstests der verwendeten Datenstrukturen hinzu.
 *	- PARSE_TREE: Der Parser gibt nach erfolgreichem Parsevorgang den
 *  	Ableitungsbaum für die Eingabe auf stdout aus. Ausgabeformat ist die
 *  	dot-Sprache, eine Beschreibungssprache für Graphen zu Benutzung mit
 *  	dem Programm \c dot aus dem graphviz-Paket.
 *	- PARSE_TREE_PHASES: Gibt nicht nur den einen Baum aus, sondern nach
 *  	jeder Phase den Wald aus bisher erzeugten partiellen Ableitungsbäumen.
 *	- DOXYGEN_ONLY: Compiliert Codeteile mit, die eigentlich an anderer
 *  	Stelle erzeugt werden, bspw. im Parsergenerator bogenlr. Dient dazu,
 *  	dass auch diese Codeabschnitte in der automatisch erzeugten
 *  	Dokumentation erläutert werden.
 *  - HASHLENGTH: Überschreibt den Defaultwert der Startlänge für
 *  	Hashtabellen, der in dieser Skelettdatei auf 7 festgelegt ist.
 *  - DNDEBUG: Abschalten der \c assert() -Überprüfungen.
 *  - SUPPRESS_OUTPUT: Ausschalten der Erfolgsmeldung nach stderr.
 *
 **************************************************************************/


/**************************************************************************
 * Präprozessor-Anweisungen:                                              *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h> // NULL, malloc
#include <stdarg.h> // Ermöglicht Argumentlisten variabler Länge
#include <stdint.h> // Bspw. für uintptr_t bei Hashtabelle
#include <string.h> // Für memset().
#include <assert.h> // Für assert().


#ifdef DEBUG_ALL
/** \brief Abfangen von fehlgeschlagenen Speicheranforderungen.
 *
 *  Wird nur übersetzt, wenn das Makro \c DEBUG_ALL definiert ist.
 *
 * 	\param size Größe des anzufordernden Speicherbereichs.
 * 	\return Zeiger auf den alloziierten Speicherbereich.
 */
static
void*
xmalloc (size_t size)
{
	void *ptr = malloc(size);

	if (ptr == NULL)
		abort();

	return ptr;
} // xmalloc (size_t size)

/** \brief Abfangen von fehlgeschlagenen Speicheranforderungen.
 *  
 *  Falls das Makro \c DEBUG_ALL definiert ist, werden Aufrufe von
 *  \c malloc vom Makro \c xmalloc abgefangen.
 */
#   define malloc(x) (xmalloc((x)))
#endif


#ifdef USE_DEBUG_PRINTF
/** Diagnose-Informationen nach \c stderr werden nur dann ausgegeben, wenn
 *  beim Übersetzen das Makro \c USE_DEBUG_PRINTF definiert ist.
 *  Ansonsten wird das Makro \c DEBUG_PRINTF zu <tt>(void(0))</tt>
 *  expandiert.
 */
#   define DEBUG_PRINTF(...)  (fprintf(stderr, __VA_ARGS__))
#else
#   define DEBUG_PRINTF(...)  ((void)0)
#endif


#ifndef HASHLENGTH
/** Startwert für die Größe der Hashtabellen, in denen die Knoten des
 *  Parsegraphen verwaltet werden.
 */
#   define HASHLENGTH 7
#endif


/**************************************************************************
 * Datentyp-Deklarationen:                                                *
 **************************************************************************/

/* Die Deklarationen im folgenden Abschnitt dienen nur der Dokumentation.
 * Sie werden im Normalfall von bogenlr generiert.
 */
#ifdef DOXYGEN_ONLY

/** \brief Datentyp zur Repräsentation der Nichtterminalsymbole
 *	(= Variablen) der Grammatik.
 *	
 *	Der Generator \c bogenlr erzeugt ein Array \c nonterminals[] von
 *	diesem Typ.
 *
 *	Nur zur Dokumentation hier aufgenommen, wird normalerweise vom
 *	Generator eingefügt.
 */
struct nonterminal_t {
	/// Anzahl Elemente des Arrays, auf das der Zeiger \a items zeigt.
	int item_count;
	/** \brief Zeiger auf ein Array mit den Nummern derjenigen Items, für
	 * 		die	bei einer Expansion anhand des aktuell betrachteten Symbols
	 * 		neue Nachfolgerknoten in den Graphen eingefügt werden müssen.
	 */
	int *items;
	/** \brief Zeiger auf einen String mit dem Klartextnamen des Symbols.
	 *
	 * 	Wird bei definiertem Makro \c PARSE_TREE benutzt, um beim Erzeugen
	 * 	eines Ableitungsbaumes die inneren Knoten mit den Namen der
	 * 	korrespondierenden Nichtterminalsymbolen zu beschriften.
	 */
	char *name;
};


/** \brief Legt fest, welche Arten von Symbolen direkt nach dem
 *  	Punkt in einem Item auftreten können.
 *
 *  Möglich sind: Nichtterminalsymbole, Terminalsymbole und das
 *  leere Wort \f$ \varepsilon \f$.
 *
 * 	Nur zur Dokumentation hier aufgenommen, wird normalerweise vom
 *	Generator eingefügt.
 */
enum symbol_type
{
	Epsilon, Nonterminal, Terminal
};


/** \brief Datenstruktur zur Repräsentation von Parser-Items.
 *
 * 	Der Parsergenerator BoGenLR erzeugt aus der Menge der Produktionen
 * 	der Grammatik	ein Array \c items[] mit Einträgen von diesem Typ.
 * 	Jedem Item wird eine eindeutige Nummer zugeordnet, die zugleich sein  
 * 	Arrayindex ist.
 *
 * 	Beim Generieren der Items wurde die Reihenfolge so gewählt, dass
 * 	für ein Item \f$ [A \rightarrow \alpha \cdot x \beta]\f$ bzw.
 * 	\f$ [A \rightarrow \alpha \cdot B \beta]\f$ mit Nummer \f$ l\f$ das
 * 	Nachfolgeitem die Nummer \f$l+1\f$ erhält. Dies ist das Item, das
 * 	nach Lesen von \f$ x\f$ bzw. nach Reduktion eines Nachfolgerknotens
 * 	von \f$B\f$ entsteht.
 *
 *	Nur zur Dokumentation hier aufgenommen, wird normalerweise vom
 *	Generator eingefügt.
 */
struct item_t {
	/// Nummer der Grammatikregel, aus der das Item entstanden ist.
	int rule_number;
	/** \brief Art des Symbols, das direkt hinter dem Punkt folgt
	 *  (Nichtterminalsymbol, Terminalsymbol oder \f$\varepsilon\f$).
	 */
	enum symbol_type dot_symbol_type;
	/** \brief Nummer des Symbols hinter dem Punkt. Undefiniert bei
	 *		\f$\varepsilon\f$.
	 */
	unsigned long dot_symbol;
	/** \brief Anzahl der Elemente in der \f$\mathrm{FIRST}_k\f$-Menge der
	 * 		rechten Seite des Items.
	 */
	int first_k_count;
	/** \brief Zeiger auf das Array, dass die \f$ \mathrm{FIRST}_k\f$-Menge
	 *  	der rechten Seite des Items enthält.
	 *
	 *  Die rechte Seite eines Items
	 *  \f$[A \rightarrow \alpha \cdot a \beta], a \in \Sigma\f$, ist
	 *  \f$a\beta\f$; die rechte Seite von
	 *  \f$[A \rightarrow \alpha \cdot B \beta], B \in N\f$, ist
	 *  \f$\beta\f$.
	 */
	unsigned long (*first_k_item)[LOOKAHEAD_LENGTH];
};


/** \brief Datenstruktur zur Repräsentation der Produktionen der
 * 		Grammatik.
 *
 * 	Wird nur compiliert, wenn das Makro \c PARSE_TREE definiert ist.
 *
 * 	Der Generator erzeugt ein Array \c rules[] von diesem Typ.
 * 	Die Nummer jeder Regel ist zugleich ihr Arrayindex und wird
 * 	durch die Reihenfolge ihres Auftretens in der Grammatikdatei
 * 	bestimmt.
 *
 *	Nur zur Dokumentation hier aufgenommen, wird normalerweise vom
 *	Generator eingefügt.
 */
struct rule_t {
	/// Nummer des Nichtterminalsymbols auf der linken Regelseite.
	signed long lhs;
	/// Anzahl der Symbole auf der rechten Regelseite.
	int rhs_count;
	/** \brief Zeiger auf ein Array mit den Symbolen der rechten Regelseite.
	 *  
	 *  Dabei stehen positive Zahlen für Terminalsymbole und negative für
	 *  die mit dem Operator \c ~ negierte Nummer eines Nichtterminalsymbols.
	 */
	signed long *rhs;
};

#endif


/** \brief Aufzählungstyp für die gewünschte Suchmenge bei der
 * 		rückwärts-topologischen Suche.
 *
 * 	Durch übergabe des entsprechenden Parameters an die Funktion
 * 	\c rtsearch wird festgelegt, ob nur lesbare, nur reduzierbare,
 * 	oder ob beide Endknotentypen durchsucht werden sollen.
 */
enum searchflag_t
{
	Read, Reduce, Both
};


// Forward-Deklarationen:
struct node_list_t;
struct node_table_t;


/** \brief Datentyp für die Knoten des Graphen.
 *
 *  Der erweiterte LR(k)-Parser verwaltet einen Graphen, dessen Knoten
 *  Items oder Nichtterminalsymbole repräsentieren können.
 */
struct node_t {
	/** \brief Die Nummer des repräsentierten Items oder Symbols.
	 *
	 *	Wenn der Knoten ein Item repräsentiert, enthält \c content die
	 *	Nummer dieses Items als positive Zahl; wenn der Knoten für ein
	 *	Nichtterminal steht, enthält \c content die mit dem Operator \c ~
	 *	negierte Nummer des Symbols.
	 */
	signed long content;
	/// Hashtafel der direkten Vorgängerknoten im Graphen
	struct node_table_t *predecessors;
	/// Hashtafel der direkten Nachfolgerknoten im Graphen
	struct node_table_t *successors;
	/** \brief Laufende Nummer des Suchlaufs, in dem der Knoten zuletzt
	 * 		besucht wurde.
	 *
	 *  Um während der rückwärts-topologischen Suche festzustellen, ob der
	 *  Präfixvektor des betrachteten Knotens aktuell ist, wird während jedes
	 *  Aufrufs der Suchfunktion ein globaler Zähler \c search_id 
	 *  inkrementiert und ggf. mit <tt>node->last_visited</tt>
	 *  abgeglichen.
	 */
	unsigned long last_visited;
	/** \brief Für die rückwärts-topologische Suche:
	 *  	Array aus Zeigern auf die Endknoten, die zu den bisher hergeleiteten
	 *  	Präfixen des Lookahead-Strings korrespondieren.
	 *
	 *	Dabei bedeutet für einen Knoten v ein Endknoten-Eintrag
	 *	endnodes[i] = w != \c NULL:
	 *	Zwischen v und w konnte ein Präfix der Länge i des Lookaheads
	 *	hergeleitet werden, und der Endknoten w ist mit diesem Präfix
	 *	assoziiert.
	 */
	struct node_t *endnodes[LOOKAHEAD_LENGTH];
	/** \brief Für die rückwärts-topologische Suche: Zusätzliche Integer-
	 *		Darstellung eines Präfixvektors.
	 *
	 *	Wird benutzt, um Endlosschleifen in Zyklen des Graphen zu vermeiden,
	 *	indem beim Besuchen eines Itemknotens v geprüft wird, ob sich der
	 *	Präfixvektor des nachfolgenden Nichtterminalknotens v seit dem
	 *	letzten	Besuch verändert hat. Nur in diesem Fall lohnt sich ein
	 *	erneuter Besuch von v.
	 *	Für einen Nichtterminalknoten wird als Bitmaske gespeichert, welche
	 *	Präfixe des Lookahead zum aktuellen Zeitpunkt bis zu diesem Knoten
	 *	hergeleitet werden können.
	 *	Für Itemknoten wird gespeichert, wie der Präfixvektor des
	 *	unmittelbar nachfolgenden Nichtterminalknotens zu dem Zeitpunkt
	 *	aussah, als der	Itemknoten zuletzt untersucht wurde.
	 *	Bedeutung der Werte:
	 *	- 0 - Kein Element des Vektors ist gesetzt.
	 *	- 1 - Nur das Präfix Epsilon (1 = 2^0)
	 *	- 2 - Nur das Präfix der Länge 1 (2 = 2^1)
	 *	- 3 - Das Präfix Epsilon und das der Länge 1 (3 = 2^0 + 2^1)
	 *	- 4 - Nur das Präfix der Länge 2 (4 = 2^2)
	 *	- ...
	 */
	unsigned long prefixes;
};


/** \brief Datentyp für Listen von Knoten.
 *
 * Eine Knotenliste besteht aus einem Knotenzeiger \a head, gefolgt von
 * Zeigern \a left und \a right auf die linken und rechten Nachbarlisten.
 */
struct node_list_t {
	/// Zeiger auf den gespeicherten Knoten
	struct node_t *head;
	/// Zeiger auf das linke Nachbarelement.
	struct node_list_t *left;
	/// Zeiger auf das rechte Nachbarelement.
	struct node_list_t *right;
};


/** \brief Datentyp für eine Hashtabelle von Knoten.
 *
 *	Aus Effizienzgründen werden die verschiedenen Knotenmengen, die
 *	während eines Parserlaufes verwaltet werden müssen, durch Hashtabellen
 *	repräsentiert.
 */
struct node_table_t {
	/// Maximallänge der Tabelle.
	size_t length;
	/** \brief Anzahl Elemente in der Tabelle.
	 *
	 *  Entspricht der Anzahl Felder im Array \c items. Wenn die 
	 *  Elementanzahl \c item_count diesen Wert erreicht, wird die
	 *  Tabelle mittels \c node_table_double vergrößert.
	 */
	size_t item_count;
	/// Zeiger für die aktuelle Position in einer Iteration durch die Tabelle.
	struct node_list_t *cursor;
	/** \brief Zeiger auf das "erste" Element der Tabelle
	 *
	 *  Dies ist das Element, mit dem eine Iteration per
	 *  \c node_table_iter startet, nachdem der Cursor mit 
	 *  \c node_table_reset_cursor initialisiert wurde. Dies ist auch das
	 *  Element, das bei \c node_table_pop geliefert und aus der Tabelle
	 *  entfernt wird.
	 */
	struct node_list_t *first;
	/** \brief  Dynamisches Array aus Knotenlisten.
	 *	
	 *	Ein Eintrag in der Hashtabelle ist ein Zeiger auf den ersten Knoten
	 *	der internen Liste, dessen Hashwert der Tabellenposition entspricht.
	 */
	struct node_list_t **items; 
};


/**************************************************************************
 * Globale Konstanten:                                                    *
 **************************************************************************/

/** \brief Reservierte Item-Nummer.
 *  
 *  Der Item-Index 0 ist reserviert für das Item [S' -> .S], das den
 *  Beginn des Parse-Vorgangs markiert.
 */
static signed long const startitem_number = 0;


/** \brief Reservierte Item-Nummer.
 *  
 *  Der Item-Index 1 ist reserviert für das Item [S' -> S.], das das
 *  Ende des Parse-Vorgangs markiert.
 */
static signed long const acceptitem_number = 1;


/**************************************************************************
 * Globale Variablen:                                                     *
 **************************************************************************/

/** \brief Flag, ob das Ende der Eingabe erreicht ist.
 *
 *  <tt>eof_reached == 0</tt>, solange \c yylex noch Eingabetoken > 0
 *  liefert.
 *  <tt>eof_reached == 1</tt>, sonst.
 */
static int eof_reached = 0;


/** \brief Zähler für die aktuelle Phase.
 *
 *	Entspricht der Anzahl bisher gelesener Token, da in jeder Phase genau
 *	ein Zeichen der Eingabe verbraucht wird.
 */
static unsigned long phase_number = 0;


/** \brief Laufende Nummer für rückwärts-topologische Suchen.
 *
 *  Jeder Suchlauf bekommt eine eindeutige Nummer zugewiesen. In jedem
 *  Knoten des Graphen befindet sich ein Feld \c last_visited, das angibt,
 *  während welches Suchlaufs der Knoten zuletzt besucht wurde.
 */
static unsigned long search_id = 0;


/** \brief Puffer für die nächsten k ungelesenen Eingabesymbole.
 *  
 *  Dabei ist k die Länge des Lookahead, wie sie durch das Makro
 *  \c LOOKAHEAD_LENGTH festgelegt ist. Um nach jedem Leseschritt das
 *  explizite Verschieben aller restlichen neun Array-Elemente zu
 *  vermeiden, werden die Array-Elemente immer zyklisch über den Index
 *  <tt>(phase_number % k)</tt> angesprochen.
 */
static unsigned long lookahead[LOOKAHEAD_LENGTH];


#ifndef DEBUG_ALL
/** \brief Pool für aktuell nicht benötigte Listenelemente.
 *
 *  Listenelemente, die aktuell nicht benötigt werden, werden nicht
 *  mittels \c free() freigegeben, sondern in dieser Liste gesammelt,
 *  um bei einer neuen Speicheranforderung ein aufwändiges \c malloc
 *  einzusparen.
 */
static struct node_list_t *pool = NULL;
#endif


/** \brief Der Parser verwaltet einen Graph aus Item- und
 * 		Nichtterminalknoten.
 *
 *  Da der Graph aus mehreren Zusammenhangskomponenten bestehen kann,
 *  ist die Datenstruktur \a graph ein Zeiger auf eine \b Liste von
 *  Anfangsknoten.
 */
static struct node_list_t *graph = NULL;
	

/** \brief Hashtafel für expandierbare Endknoten */
static struct node_table_t *expandable_nodes = NULL;


/** \brief Hashtafel für lesbare Endknoten */
static struct node_table_t *readable_nodes = NULL;


/** \brief Temporäre Liste für lesbare Endknoten, die nach einem
 *  	Reduktions-/Leseschritt neu entstanden sind:
 */
static struct node_list_t *new_readable_nodes = NULL;


/** \brief Hashtafel für reduzierbare Endknoten */
static struct node_table_t *reducible_nodes = NULL;


/** \brief Hashtafel bereits expandierter Knoten.
 *
 *  Um den Aufwand bei den Expansionsschritten zur verringern, wird in
 *  jeder Phase Buch darüber geführt, welche Itemknoten in der aktuellen
 *  Phase schon expandiert wurden.
 */
static struct node_table_t *expanded_nodes = NULL;


/** \brief Hashtafel bereits expandierter Nichtterminalknoten.
 *
 *  Um den Aufwand bei den Expansionsschritten zur verringern, wird in
 *  jeder Phase Buch darüber geführt, welche Nichtterminalknoten
 *  in der aktuellen Phase schon bei Expansionen erschaffen wurden.
 */
static struct node_table_t *expanded_nonterminals = NULL;


/**************************************************************************
 * Funktionen - Verwaltung von Knoten und Knotenlisten:                   *
 **************************************************************************/

// forward
static struct node_table_t* node_table_init (void);


/** \brief Erzeugen eines neuen Knotens.
 *
 *  \param content Die Nummer des Items oder die (negierte) Nummer des
 *  	Nichtterminalsymbols, für das der Knoten steht.
 *  \return Einen Zeiger auf einen neuen Knoten vom Typ
 *  	<tt>struct node_t</tt>, dessen weitere Felder mit \c NULL bzw. \c 0
 *  	initialisiert wurden.
 */
static
struct node_t*
create_node (signed long content)
{
	struct node_t *node = malloc(sizeof(struct node_t));
	node->content = content;
	node->predecessors = node_table_init();
	node->successors = node_table_init();
	node->last_visited = 0;
	memset(node->endnodes, 0, LOOKAHEAD_LENGTH * sizeof(struct node_t*));
	node->prefixes = 0;

	return node;	
} // struct node_t *create_node


/** \brief Anfordern eines neuen Listenelementes aus dem Pool.
 *
 *  Anstatt bei jeder Eintragung eines Knotens in eine Liste Speicher für
 *  ein neues Listenelement per \c malloc anzufordern, wird zunächst im
 *  Pool nachgesehen, ob es noch ein ungenutztes Listenelement gibt.
 *  Dieses kann allerdings noch ungültige Daten enthalten. Im aufrufenden
 *  Code muss daher für eine sinnvolle Belegung gesorgt werden.
 *
 *  \return Zeiger auf ein Listenelement vom Typ
 *  	<tt>struct node_list_t</tt>.
 */
static
struct node_list_t*
pool_get (void)
{
#ifdef DEBUG_ALL
	return malloc(sizeof(struct node_list_t));
#else
	/* Falls der Pool kein Element mehr enthält, muss Speicher für ein
	 * neues angefordert werden:
	 */
	if (pool == NULL) {
		return malloc(sizeof(struct node_list_t));
	}
	else
	{
		struct node_list_t *elem = pool;
		pool = elem->right;
		return elem;
	}
#endif
} // static struct node_list_t* pool_get


/** \brief Zurückgeben eines aktuell nicht benötigten Listenelementes
 *  	an den Pool.
 *
 *  Anstatt den Speicher für ein aktuell nicht benötigtes Listenelement
 *  per \c free sofort wieder an das System zurückzugeben, werden die
 *  Listenelemente in einem Pool gesammelt, aus dem bei Bedarf schnell
 *  wieder Speicher zugewiesen werden kann. Dabei wird nicht darauf
 *  geachtet, ob das Listenelement noch mit linken und rechten
 *  Nachbarn verbunden ist; darauf muss im aufrufenden Code geachtet
 *  werden.
 *
 *  \param elem Zeiger auf das Listenelement, das wieder in den Pool
 *  	aufgenommen wird.
 */
static
void
pool_release(struct node_list_t *elem)
{
#ifdef DEBUG_ALL
	free(elem);
#else
	if (elem != NULL)
	{
		/* Setze elem an den Anfang des Pools: */
		elem->left = NULL;
		elem->right = pool;
		if (pool != NULL)
			pool->left = elem;
		pool = elem;
	}
#endif
} // static void pool_release


/** \brief Zurückgeben einer kompletten Liste an den Pool.
 *
 * 	Vergleichbar mit \c pool_release werden auch hier die Elemente
 * 	der zu löschenden Liste nicht per \c free an das System
 * 	zurückgegeben, sondern im Pool gesammelt. Im Unterschied zu
 * 	\c pool_release werden hier alle Listenelemente bis zum
 * 	Listenende an den Pool zurückgegeben.
 *
 * 	\param first Zeiger auf den Anfang der zurückzugebenden Liste.
 */
static
void
pool_release_list(struct node_list_t *first)
{
	if (first == NULL)
		return;
#ifdef DEBUG_ALL
	struct node_list_t *next = first->right;
	free(first);
	while (next != NULL) {
		first = next;
		next = first->right;
		free(first);
	}
#else
	/* Suche zunächst das Ende der zu löschenden Liste: */
	struct node_list_t *last = first;
	while(last->right != NULL) {
		last = last->right;
	}

	/* Nun zeigt last auf das letzte Element der zu löschenden Liste. */
	if (pool != NULL)
		pool->left = last;
	last->right = pool;
	pool = first;
#endif
} // static void pool_release_list


/** \brief Anhängen eines neuen Elementes an den Anfang einer Knotenliste.
 *
 *	Dabei wird nicht darauf geachtet, dass die Liste duplikatfrei bleibt.
 *	Sollte der Zeiger *list, über den die Liste betreten wird, aus der Mitte
 *	einer größeren Liste stammen, wird das neue Element auch mit dem
 *	bisherigen linken Nachbarn von *list verbunden.
 *
 *	\param list Zeiger auf einen Zeiger auf die Liste, in die der neue
 *		Knoten eingefügt werden soll (Call by Reference).
 *	\param node Zeiger auf den Graphknoten, der neu in die Liste
 *		eingefügt werden soll.
 */
static
void
list_add (struct node_list_t **list, struct node_t *node)
{
	/* Falls node auf keinen Knoten zeigt, liegt ein Fehler vor: */
	assert(node != NULL);

	struct node_list_t *lst_new = pool_get();

	lst_new->head = node;

	/* Verbinde das neue Listenelement mit seinen neuen Nachbarn: */
	if (*list != NULL) {
		lst_new->left = (*list)->left;
		(*list)->left = lst_new;
		if (lst_new->left != NULL)
			lst_new->left->right = lst_new;
	}
	else
		lst_new->left = NULL;

	lst_new->right = *list;

	*list = lst_new;
} // void list_add


/** \brief Duplikatfreies Anhängen eines neuen Elementes an das Ende
 *		einer	Knotenliste.
 *
 *	Sollte sich das Element bereits in der Liste befinden, wird nichts
 *	unternommen.
 *
 *	\param list Zeiger auf einen Zeiger auf die Liste, in die der neue
 *		Knoten eingefügt werden soll (Call by Reference).
 *	\param node Zeiger auf den Graphknoten, der neu in die Liste
 *		eingefügt werden soll.
 *	\return \c 0, falls der Knoten eingefügt werden konnte, \c 1,
 * 		falls sich \c node schon vorher in der Liste befindet.
 */
static
int
list_enqueue (struct node_list_t **list, struct node_t *node)
{
	/* Falls node auf keinen Knoten zeigt, liegt ein Fehler vor: */
	assert(node != NULL);

	/* Falls die Liste bisher leer ist, bildet das neue Element den
	 * Listenanfang.
	 */
	if (*list == NULL) {
		struct node_list_t *lst_new = pool_get();
		lst_new->head = node;
		lst_new->left = NULL;
		lst_new->right = NULL;
		*list = lst_new;
		return 0;
	}

	/* Falls die Liste nicht leer ist: Suche das letzte Element und füge
	 * das neue dahinter an.
	 */
	else {

		// Zeiger auf das aktuell betrachtete Listenelement:
		struct node_list_t *lst_tmp = *list;

		/* Sollte sich das neue Element bereits am Listenanfang befinden, wird
		 * die Funktion ohne Aktion abgebrochen.
		 */
		if (lst_tmp->head == node)
			return 1;

		while (lst_tmp->right != NULL) {
			lst_tmp = lst_tmp->right;

			/* Falls das aktuelle Listenelement gleich dem neu einzufügenden
			 * ist, endet die Funktion ohne Aktion.
			 */
			if (lst_tmp->head == node)
				return 1;
		}

		/* Wenn die Ausführung der Funktion an diese Stelle gelangt, ist das
		 * Ende der Liste gefunden (lst_tmp zeigt auf das letzte Listenelement,
		 * lst_tmp->right ist NULL), und das neue Element ist noch nicht in der
		 * Liste enthalten. 
		 */
		struct node_list_t *lst_new = pool_get();
		lst_new->head = node;
		lst_new->left = lst_tmp;
		lst_new->right = NULL;
		lst_tmp->right = lst_new;
		return 0;
	
	} // if (*list == NULL) ... else

} // void list_enqueue


/** \brief Überprüft, ob in einer Knotenliste ein Knoten mit einer
 *  bestimmten Nummer enthalten ist.
 *
 *	Der Aufruf <tt>list_contains_node_num (lst, node_number)</tt> prüft,
 *  ob sich in der Liste, auf die \a lst zeigt, ein Eintrag für den
 *  Knoten mit Index \a node_number befindet.
 *
 *	\param list Zeiger auf die zu durchsuchende Knotenliste.
 *	\param node_number Index des Knotens, der gesucht wird.
 *		Wenn \a node_number eine	positive Zahl ist, wird ein Itemknoten
 *		gesucht, sonst ein Nichtterminalknoten.
 *	\return \c 1, falls der gesuchte Knoten sich in \a lst befindet;
 *		\c 0, sonst.
 */
static
int
list_contains_node_num (struct node_list_t *list, signed long node_number)
{
	while (list != NULL) {
		if (list->head->content == node_number)
			return 1;
		else
			list = list->right;
	}

	/* An diesen Punkt gelangt die Funktion nur, wenn die Liste erfolglos
	 * durchsucht wurde:
	 */
	return 0;
} // int list_contains_node_num


/** \brief Entnimmt den ersten Knoten einer Liste.
 *
 *	Die Funktion \c list_pop liefert einen Zeiger auf den	ersten Knoten
 *	einer Knotenliste und entfernt seinen Eintrag aus dieser Liste.
 *
 *	\param list Zeiger auf einen Zeiger auf die behandelte Liste.
 *		(Call by Reference).
 *	\return Zeiger auf das erste Listenelement, falls es existiert;
 *		Sonst: \c NULL.
 */
static
struct node_t*
list_pop (struct node_list_t **list)
{
	if (*list == NULL)
		return NULL;
	else {
		struct node_list_t *old_first = *list;
		struct node_t *node = old_first->head;
		*list = old_first->right;
		/* Falls sich das Listenelement, über den die Liste betreten wird, in
		 * ihrer Mitte befindet, muss auch der left-Zeiger angepasst werden!
		 */
		if ((*list) != NULL)
			(*list)->left = old_first->left;

		pool_release(old_first);

		return node;
	}
} // struct node_t *list_pop


/** \brief Sucht einen Knoten anhand seines Zeigers und entfernt ihn aus
 *	einer Knotenliste.
 *
 *	Die Funktion \c list_remove erhält als Argumente die Adresse eines
 *	Zeigers auf eine Knotenliste und einen Zeiger auf einen	Knoten.
 *	Wenn sich der Knoten in der Liste befindet, wird sein erstes Vorkommen
 *	aus der Liste entfernt.
 *
 *	\param list Zeiger auf einen Zeiger auf die Knotenliste, aus der
 *		der Knoten erntfernt werden soll (Call by Reference).
 *	\param node Zeiger auf den Knoten, der aus der Liste
 *		ausgetragen werden soll.
 *	\return Integerwert für Erfolg oder Misserfolg der Aktion.
 *		Dabei wird \c 0 zurückgegeben, wenn der Knotenzeiger erfolgreich
 *		aus der Liste entfernt wurde, \c 1, falls die Liste leer ist,
 *		und 2, falls der Knotenzeiger nicht enthalten ist.
 */
static
int
list_remove (struct node_list_t **list, struct node_t *node)
{
	/* Falls node auf keinen Knoten zeigt, liegt ein Fehler vor: */
	assert(node != NULL);

	/* Falls die Liste leer ist, gibt es nichts zu entfernen: */
	if ( (*list) == NULL ) {
		return 1;
	}

	// Zeiger auf das aktuell betrachtete Listenelement:
	struct node_list_t *tmp = *list;

	/* Falls schon das erste Listenelement das zu Löschende ist, wird es
	 * entfernt und der Listenanfang auf \c right umgebogen. Für den
	 * Fall, dass der übergebene Listenzeiger aus der Mitte einer größeren
	 * Liste stammt, wird auch der left-Zeiger angepasst.
	 */
	if (tmp->head == node) {
		if (tmp->left != NULL)
			tmp->left->right = tmp->right;
		if (tmp->right != NULL)
			tmp->right->left = tmp->left;

		*list = tmp->right;
		pool_release(tmp);

		return 0;
	}

	/* Falls es nicht das erste Element ist: Ueberprüfe iterativ den
	 * Nachfolger des aktuellen Elementes:
	 */
	// Zeiger auf den Nachfolger des aktuellen Listenelementes:
	struct node_list_t *tmp2 = NULL;
	while (tmp->right != NULL) {
		if (tmp->right->head == node) {
			/* Merke Dir die Restliste tmp->right als tmp2. Verbinde dann tmp
			 * und tmp->right->right, so dass tmp2 "übergangen" wird.
			 * Nun kann tmp2 entfernt werden.
			 */
			tmp2 = tmp->right;
			tmp->right = tmp2->right;
			if (tmp->right != NULL)
				tmp->right->left = tmp;

			pool_release(tmp2);

			return 0;
		}
		else
			tmp = tmp->right;
	}

	/* An diesen Punkt gelangt die Funktion nur, wenn die Liste vollständig
	 * erfolglos durchsucht wurde:
	 */
	return 2;
} // int remove_node (...)


#ifdef USE_DEBUG_PRINTF
/** \brief Ausgabe einer Knotenliste auf der Standardausgabe.
 *
 *	Der Aufruf <tt>list_print (list)</tt> gibt zu Diagnosezwecken eine
 *	einfache textuelle Darstellung der Knotenliste \a list aus.
 *
 *	Die Funktion wird hinter dem Makro \c LIST_PRINT versteckt und nur
 *	ausgeführt, falls das Makro \c USE_DEBUG_PRINTF definiert ist.
 *	Ansonsten wird \c LIST_PRINT zu <tt>(void)0</tt> expandiert.
 *
 *	\param list Zeiger auf die Knotenliste, die ausgegeben werden soll.
 */
static
void
list_print (struct node_list_t *list)
{
	DEBUG_PRINTF("[");
	while (list != NULL) {
		if (list->head->content >= 0)
			DEBUG_PRINTF(" %ld", list->head->content);
		else
			DEBUG_PRINTF(" n%ld", ~(list->head->content));
		list = list->right;
	}
	DEBUG_PRINTF("]\n");
} // void list_print


/** Falls das Makro \c USE_DEBUG_PRINTF nicht definiert ist, expandiert
 *  \c LIST_PRINT zu <tt>(void)0</tt>.
 */
#   define LIST_PRINT(x) (list_print((x)))
#else
#   define LIST_PRINT(x) ((void)0)
#endif


/** \brief Vollständiges Leeren einer Knotenliste.
 *
 *	Der Aufruf <tt>list_clear (list)</tt> entfernt alle Listeneinträge,
 *	so dass \c *list anschließend ein Nullzeiger ist.
 *
 *	\param list Zeiger auf einen Zeiger auf die zu leerende Knotenliste
 *		(Call by Reference).
 */
static
void
list_clear (struct node_list_t **list)
{
	/* Falls die Liste schon leer ist, gibt es nichts zu tun: */
	if (*list == NULL)
		return;

	/* Falls das erste Element der zu löschenden Liste aus der Mitte einer
	 * größeren Liste stammt, löse erst diese Verbindung:
	 */
	if ((*list)->left != NULL) {
		(*list)->left->right = NULL;
		(*list)->left = NULL;
	}

	/* Nun gib die Liste an den Pool zurück: */
	pool_release_list(*list);
	*list = NULL;
} // void list_clear


/**************************************************************************
 * Funktionen - Verwaltung von Hashtafeln von Knoten:                     *
 **************************************************************************/

/** \brief Berechnen des Hashwertes für einen Knoten.
 *
 * 	\param node Zeiger auf den Knoten, dessen Hashwert bestimmt werden
 * 		soll
 * 	\param length Länge der Tabelle, in der der Knoten eingetragen
 * 		ist/werden soll. Wird benutzt als Modulo-Faktor für den
 * 		Hashwert.
 * 	\return Hashwert als unsigned int
 */
static inline
unsigned int
compute_hash (struct node_t *node, size_t length)
{
	return ( ((uintptr_t)node >> 4) % length);
} // static unsigned int compute_hash


#ifdef DEBUG_ALL
/** \brief Test, ob eine Liste in einer anderen enthalten ist.
 *
 * 	Bei einem Aufruf <tt>list_contains_sublist(lst, sublst)</tt> wird
 * 	geprüft, ob die Liste \c sublst Teil von \c lst ist.
 *
 * 	Diese Funktion wird nur compiliert, wenn das Makro \c DEBUG_ALL
 * 	definiert ist und dient als Hilfsfunktion für
 * 	\c node_table_check.
 *
 * 	\param lst Liste, in der nach \a sublst gesucht werden soll.
 * 	\param sublst Liste, auf deren Enthaltensein in \c lst getestet wird.
 *	\return 0, falls \a sublst in \a lst enthalten ist; 1, sonst,
 */
static
int
list_contains_sublist (struct node_list_t *lst, struct node_list_t *sublst)
{
	/* Wenn schon das erste Element von lst identisch mit sublst ist,
	 * sind wird schon fertig:
	 */
	if (lst == sublst)
		return 0;

	/* Ansonsten gehe die Liste lst durch und teste für jedes Element, ob
	 * sein Nachfolger NULL oder sublst entspricht.
	 */
	while (lst != NULL) {
		if (lst->right == sublst)
			return 0;

		lst = lst->right;
	} // while (lst != NULL)

	/* Wenn die Funktionsausführung an diese Stelle gelangt, wurde lst
	 * komplett durchsucht, ohne sublst zu finden.
	 */
	return 1;
} // static int list_contains_sublist


/** \brief Integritätstest für eine Knoten-Hashtafel.
 *
 * 	Für eine Hashtafel \a tbl wird getestet, ob die Einträge in
 * 	\c tbl->items und in der Liste \c tbl->first zueinander passen.
 * 	Dadurch fällt bspw. auf, falls in \c tbl->items Listenelemente
 * 	eingtragen sind, die eigentlich nicht Teil der Tabelle sein
 * 	dürfen.
 *
 * 	Diese Funktion wird nur compiliert, falls das Makro
 * 	\c DEBUG_ALL definiert ist.
 *
 * 	\param tbl Zeiger auf die zu überprüfende Tabelle.
 * 	\param flag Zahl, die als Flag benutzt wird.
 * 		<tt>flag == 1</tt> bewirkt, dass eine Ausgabezeile generiert wird
 * 		mit den Angaben:
 * 		- <tt>tbl->length</tt>
 * 		- \c count (im Normalfall <tt>== tbl->item_count</tt>, also Anzahl 
 * 			gefundener Tabellenelemente
 * 		- \c line_count (Anzahl belegter Tabellenzeilen)
 * 		- Durchschnittliche Anzahl von Elementen je belegter 
 * 			Tabellenzeile <tt>count / line_count</tt>.
 */
static
void
node_table_check (struct node_table_t *tbl, int flag)
{
	/* Für jeden Listenzeiger in tbl->items wird geprüft, ob er auch von
	 * tbl->first aus erreichbar ist. Wenn nicht, so liegt ein Fehleintrag
	 * in der Tabelle vor!
	 */
	unsigned int i = 0;
	int found_first = 0;
	size_t line_count = 0, count = 0;
	struct node_list_t *lst = NULL;
	for (i=0; i<tbl->length; i++) {
		lst = tbl->items[i];
		if (lst != NULL) {
			count++;
			line_count++;
			if (compute_hash(lst->head, tbl->length) != i)
				abort();
			if (lst == tbl->first)
				found_first = 1;
			if (list_contains_sublist(tbl->first, lst)	!= 0)
				abort();

			while ( lst->right != NULL
					&& (compute_hash(lst->right->head, tbl->length) == i) )
			{
				count++;
				lst = lst->right;
			}
		}
	} // for (i=0; i<tbl->length; i++)

	if (flag == 1)
		printf("stat: length: %zu item_count: %zu line_count: %zu relativ: %f\n",
			tbl->length, tbl->item_count, line_count,
			((double)count / (double)line_count));

	if (count != tbl->item_count)
		abort();

	if (found_first == 0 && count > 0)
		abort();

	count = 0;
	lst = tbl->first;
	while (lst != NULL) {
		count++;
		lst = lst->right;
	}
	if (count != tbl->item_count)
		abort();

} // static void node_table_check

/** Falls das Makro \c DEBUG_ALL nicht definiert ist, expandiert das
 *  Makro \c NODE_TABLE_CHECK zu <tt>(void)0</tt>.
 */
#   define NODE_TABLE_CHECK(tbl, flag) (node_table_check((tbl), (flag)))
#else
#   define NODE_TABLE_CHECK(tbl, flag) ((void)0)
#endif


/** \brief Erzeugen einer frischen Hashtabelle der Größe 0.
 *
 * 	Erzeugt wird eine leere Hashtabelle für Knoten. Die Erzeugung erfolgt
 * 	lazy, d.h. alle Elemente werden mit Nullwerten initialisiert.
 * 	Das Array	\c items aus Listenzeigern wird erst beim ersten Zugriff
 * 	mittels	\c node_table_add tatsächlich mit der Größe initialisiert,
 * 	die durch das Makro \c HASHLENGTH festgelegt ist. Dann erst wird auch
 * 	das Feld \c length auf \c HASHLENGTH gesetzt.
 *
 * 	\return Frische Hashtabelle vom Typ node_table_t.
 */
static
struct node_table_t*
node_table_init (void)
{
	struct node_table_t *tbl = malloc(sizeof(struct node_table_t));
	tbl->length = 0;
	tbl->item_count = 0;
	tbl->first = NULL;
	tbl->cursor = NULL;
	tbl->items = NULL;

	NODE_TABLE_CHECK(tbl, 0);
	return tbl;
} // static void node_table_init


/** \brief Leeren einer Hashtabelle von Knoten.
 *
 * 	Ein Aufruf von <tt>node_table_clear(tbl)</tt> entfernt die Tabelle
 * 	\a tbl nicht vollständig aus dem Speicher, sondern leert nur das
 * 	enthaltene Array der Listenzeiger.
 *
 * 	\param tbl Zeiger auf die zu bereinigenden Knotentabelle.
 */
static
void
node_table_clear(struct node_table_t *tbl)
{
	/* Falls die Tabelle noch nicht initialisiert war oder keine Einträge
	 * enthält, gibt es nichts zu bereinigen:
	 */
	if (tbl == NULL)
		return;
	if (tbl->item_count == 0)
		return;

	NODE_TABLE_CHECK(tbl, 0);

	list_clear(&(tbl->first));
	memset(tbl->items, 0, tbl->length * sizeof(struct node_list_t*));
	tbl->item_count = 0;

	NODE_TABLE_CHECK(tbl, 0);
} // static void node_table_clear


/** \brief Vollständiges Löschen einer Knotentabelle.
 *
 * 	Nach einem Aufruf von <tt>node_table_free(tbl)</tt> sollte im
 * 	aufrufenden Code der Zeiger \a tbl sofort wieder neu belegt werden,
 * 	ggf. mit <tt>NULL</tt>, da der Zeiger sonst nicht mehr zwingend auf
 * 	ein gültiges Datum zeigt.
 *
 * 	\param tbl Zeiger auf die zu löschende Tabelle.
 */
static
void
node_table_free(struct node_table_t *tbl)
{
	if (tbl == NULL)
		return;

	NODE_TABLE_CHECK(tbl, 0);

	/* Falls die Tabelle noch Einträge enthält, müssen diese zuerst
	 * entfernt werden:
	 */
	if (tbl->item_count > 0) {
		list_clear(&(tbl->first));
		/* Ein Nullsetzen von tbl->items per memset() wie bei node_table_clear
		 * ist nicht nötig, da der Speicher für die Hashtabelle als nächstes
		 * vollständig freigegeben wird:
		 */
	}

	free(tbl->items);
	free(tbl);
} // static void node_table_free


// forward:
static
void
node_table_add(struct node_table_t *tbl, struct node_t *node);


/** \brief Verdoppeln einer vollen Hashtabelle.
 *
 * 	Wenn der Belegungsfaktor der Hashtabelle den Wert 1 erreicht, wird
 * 	node_table_double aufgerufen, um sie in eine neue Tabelle mit
 * 	doppelter Größe umzukopieren.
 *
 * 	\param tbl Zeiger auf die Hashtabelle, die verdoppelt werden soll.
 */
static
void
node_table_double(struct node_table_t *tbl)
{
	assert(tbl != NULL);

	if (tbl->length == 0)
		return;

	NODE_TABLE_CHECK(tbl, 0);

	unsigned int old_length = tbl->length;
	/* Zeiger auf den alten Inhalt der Hashtabelle: */
	struct node_list_t *lst_old = tbl->first;

	/* Erschaffe zunächst eine neue Tabelle mit doppelter Größe wie
	 * tbl->length und trage sie als neue tbl->items:
	 */
	tbl->length = 2 * old_length;
	tbl->cursor = NULL;
	tbl->first = NULL;
	free(tbl->items);
	tbl->items =
		calloc(tbl->length, sizeof(struct node_list_t*));

	/* Nun werden alle Elemente von lst_old in die neue Tabelle tbl->items
	 * übertragen.
	 * Dabei muss im Gegensatz zu node_table_add keine Duplikatprüfung
	 * stattfinden, da hier ja nur eine bereits als duplikatfrei bekannte
	 * Tabelle umkopiert wird. Auch wird tbl->item_count nicht angefasst,
	 * da sich die Belegung der Tabelle ja nicht ändert.
	 */
	struct node_t *node = NULL;
	unsigned int node_hash = 0;
	struct node_list_t *lst_new = NULL;

	while (lst_old != NULL) {
		node = list_pop(&(lst_old));
		node_hash = compute_hash(node, tbl->length);

		/* Falls noch kein Knoten mit dem Hashwert node_hash existiert, wird
		 * das Feld tbl->items[node_hash] neu bestückt und der neue Eintrag
		 * auch mit dem Listenzeiger tbl->first verbunden:
		 */
		if (tbl->items[node_hash] == NULL) {
			lst_new = pool_get();
			lst_new->head = node;
			lst_new->left = NULL;
			lst_new->right = tbl->first;

			if (tbl->first != NULL)
				tbl->first->left = lst_new;
			tbl->first = lst_new;

			tbl->items[node_hash] = lst_new;
		} // if (tbl->items[node_hash] == NULL)

		/* Falls es schon Einträge für andere Knoten mit dem Hashwert
		 * node_hash in tbl gibt, wird der Eintrag für node davor gesetzt.
		 * Dabei muss node auch mit einem ggf. vorhanden linken Nachbarn
		 * verbunden werden. Falls der neue rechte Nachbar von node das
		 * bisherige first-Element von tbl war, muss auch dieser Zeiger
		 * aktualisiert werden:
		 */
		else {
			lst_new = pool_get();
			lst_new->head = node;
			lst_new->left = tbl->items[node_hash]->left;
			lst_new->right = tbl->items[node_hash];

			tbl->items[node_hash]->left = lst_new;
			
			if (lst_new->left != NULL)
				lst_new->left->right = lst_new;

			if (tbl->first == lst_new->right)
				tbl->first = lst_new;

			tbl->items[node_hash] = lst_new;
		} // if (tbl->items[node_hash] == NULL) ... else

	} // while (lst_old != NULL)

	NODE_TABLE_CHECK(tbl, 1);

} // static void node_table_double


/** \brief Hinzufügen eines neuen Knotens zu einer Hashtabelle.
 *
 * 	Ein Aufruf <tt>node_table_add(tbl, node)</tt> fügt den Knoten, auf
 * 	den \a node zeigt, in die Hashtabelle ein, auf die \a tbl zeigt.
 * 	Dabei wird davon ausgegangen, dass es sich um eine gültig
 * 	vorinitialisierte Hashtabelle handelt, ansonsten wird das Programm
 * 	abgebrochen.
 * 	Sollte \a node sich bereits in der Tabelle befinden, wird die Funktion
 * 	ohne Aktion beendet.
 *
 * 	\param tbl Zeiger auf die Hashtabelle, der ein neuer Knoten
 * 		hinzugefügt werden soll.
 * 	\param node Zeiger auf den Knoten, der neu in die Tabelle aufgenommen
 * 		werden soll.
 */
static
void
node_table_add (struct node_table_t *tbl, struct node_t *node)
{
	assert(node != NULL && tbl != NULL);

	NODE_TABLE_CHECK(tbl, 0);

	/* Falls die Tabelle bisher nur mit Nullwerten initialisiert, aber noch
	 * nicht benutzt wurde, wird sie nun mit der Standardgröße HASHLENGTH
	 * initialisiert:
	 */
	if (tbl->length == 0) {
		/* Alle Einträge werden zunächst auf NULL gesetzt. Die Länge des Arrays
		 * tbl->items wird durch das Makro HASHLENGTH festgelegt.
		 */
		tbl->length = HASHLENGTH;
		tbl->item_count = 0;
		tbl->cursor = NULL;
		tbl->first = NULL;
		tbl->items = 
			calloc(HASHLENGTH, sizeof(struct node_list_t*));
	} // if (tbl->length == 0);

	unsigned int node_hash = compute_hash(node, tbl->length);

	/* Versuche, den neuen Knoten in das Tabellenfeld mit der Nummer
	 * node_hash einzutragen. Falls dort noch kein anderer Knoten mit dem
	 * Hashwert node_hash eingetragen war oder unter den dort eingetragenen
	 * Knoten der neue Knoten node noch nicht enthalten war, wird
	 * add_success auf 1 gesetzt, sonst auf 0.
	 */
	int add_success = 0;

	/* Falls noch kein Knoten mit demselben Hashwert existiert, wird das
	 * Feld tbl->items[node_hash] neu bestückt und der neue Eintrag auch mit
	 * dem Zeiger first verbunden:
	 */
	if (tbl->items[node_hash] == NULL) {
		struct node_list_t *lst_new = pool_get();
		lst_new->head = node;
		lst_new->left = NULL;
		lst_new->right = tbl->first;

		if (tbl->first != NULL)
			tbl->first->left = lst_new;

		tbl->first = lst_new;
		tbl->items[node_hash] = lst_new;
		add_success = 1;
	} // if (tbl->items[node_hash] == NULL)

	/* Wenn es schon Knoten mit demselben Hashwert in tbl gibt, muss geprüft
	 * werden, ob vielleicht schon node selbst eingetragen ist. Dann gibt es
	 * nichts zu tun. Ansonsten muss ein neues Listenelement für node an
	 * geeigneter Stelle eingefügt werden.
	 */
	else {
		/* Ist es schon der erste Knoten? */
		if (tbl->items[node_hash]->head == node)
			add_success = 0;

		/* Falls es nicht der erste Knoten ist, untersuche alle Nachfolger mit
		 * demselben Hashwert.
		 */
		else {
			struct node_list_t *lst_ptr = tbl->items[node_hash];
			int found = 0;

			/* Solange es noch Nachfolgeknoten mit demselben Hashwert gibt, müssen
			 * diese daraufhin untersucht werden, ob sie identisch mit node sind.
			 * Falls einer gefunden wird, findet keine Einfügung statt,
			 * falls nicht, wird der neue Knoten an passender Stelle eingefügt.
			 */
			while ( (lst_ptr->right != NULL)
					&& (compute_hash(lst_ptr->right->head, tbl->length) == node_hash)
					&& (found != 1) )
			{
				/* Das Schleifeninnere wird also nur betreten, wenn der Knoten im
				 * Listenfeld lst_ptr->right existiert und auch den Hashwert 
				 * node_hash hat. Falls dieser Knoten identisch mit node ist, 
				 * findet keine Einfügung statt:
				 */
				if (lst_ptr->right->head == node)
					found = 1;

				/* Ansonsten wird die Untersuchung beim Nachfolgerknoten wiederholt:
				 */
				else
					lst_ptr = lst_ptr->right;
			} // while ( ... )

			/* Falls nun found == 0, so verweist der Zeiger lst_ptr auf den 
			 * letzten Knoten in der Knotenliste von tbl, der den Hashwert
			 * node_hash hat.
			 * Der rechte Nachbar lst_ptr->right ist entweder NULL oder zeigt
			 * auf einen Knoten mit anderem Hashwert. Es muss ein neues
			 * Listenelement für node angelegt werden, das zwischen lst_ptr
			 * und lst_ptr->right eingetragen wird.
			 */
			if (found == 0) {
				struct node_list_t *lst_new = pool_get();
				lst_new->head = node;
				lst_new->left = lst_ptr;
				lst_new->right = lst_ptr->right;

				lst_ptr->right = lst_new;

				if (lst_new->right != NULL)
					lst_new->right->left = lst_new;

				add_success = 1;
			} // if (found == 0)

			/* Falls aber found == 1, so befindet sich ein Eintrag für node
			 * bereits in tbl, und es findet keine Einfügung statt.
			 */
			else {
				add_success = 0;
			}

		} // if (tbl->items[node_hash]->head == node) ... else

	} // if (tbl->items[node_hash] == NULL) ... else

	/* Falls tatsächlich eine Einfügung stattgefunden hat, muss getestet
	 * werden, ob die Tabelle vergrößert werden muss:
	 */
	if (add_success == 1) {
		tbl->item_count++;
		NODE_TABLE_CHECK(tbl, 1);
		if (tbl->item_count >= tbl->length)
			node_table_double(tbl);
	} // if (add_success == 1)

} // static void node_table_add


/** \brief Entfernen eines Knotens aus einer Hashtabelle.
 *
 * 	Der Aufruf <tt>node_table_remove(tbl, node)</tt> entfernt den
 * 	Knotenzeiger \a node aus der Tabelle, auf die \a tbl zeigt.
 * 	Dabei wird davon ausgegangen, dass \a tbl auf eine gültig
 * 	initialisierte Hashtabelle zeigt, ansonsten wird das Programm
 * 	abgebrochen.
 * 	Wenn der Knoten nicht in der Tabelle enthalten war, endet die Funktion
 * 	ohne Aktion.
 *
 *	\todo Testen, ob bei Unterschreitung eines Belegungsfaktors von 1/4
 *		ein node_table_shrink Vorteile bei Performance und/oder
 *		Speicherverbrauch bewirkt.
 *
 *	\param tbl Zeiger auf die Tabelle, aus der der Knoten entfernt werden
 *		soll.
 *	\param node Knotenzeiger, der aus \a tbl entfernt werden soll.
 */
static
void
node_table_remove(struct node_table_t *tbl, struct node_t *node)
{
	assert(node != NULL && tbl != NULL);

	NODE_TABLE_CHECK(tbl, 0);

	if (tbl->length == 0 || tbl->item_count == 0) {
		return;
	}

	unsigned int node_hash = compute_hash(node, tbl->length);

	/* Versuche, den Knoten aus der Liste, die an das Tabellenfeld
	 * mit der Nummer node_hash angehängt ist, zu entfernen. Wenn
	 * die Entfernung erfolgreich ist, gibt list_remove den Wert 0
	 * zurück, sonst 1 oder 2.
	 */
	/* Aufgrund der angepassten Implementierung von list_remove muss an dieser
	 * Stelle auch nach Änderung der Implementierung auf eine doppelt 
	 * verkettete Liste nichts geändert werden. Es muss einzig beachtet
	 * werden, dass nun ggf. das first-Element gelöscht wird, sodass dieser
	 * Zeiger noch korrigiert werden muss.
	 */
	if (tbl->first->head == node)
		tbl->first = tbl->first->right;
	/* TODO: Evtl. ist es besser, den Knoten hier "per Hand" zu löschen und
	 * nicht mittels list_remove, sonst laufe ich in manchen Fällen bis ans
	 * Listenende, also über den letzten Knoten mit Hashwert node_hash hinaus!
	 * Das führt zwar auch zu einem korrekten Ergebnis, verbraucht aber
	 * unnötig Zeit!
	 */
	int remove_success = list_remove(&(tbl->items[node_hash]), node);
	if ( tbl->items[node_hash] != NULL &&
			(compute_hash(tbl->items[node_hash]->head, tbl->length) != node_hash)
		)
		tbl->items[node_hash] = NULL;

	/* Wenn die Entfernung erfolgreich war, muss der Belegzähler der Tabelle
	 * dekrementiert werden. Ausserdem findet ein Test statt, ob die Tabelle
	 * verkleinert werden kann:
	 */
	if (remove_success == 0) {
		tbl->item_count--;
/*
		if (tbl->item_count < (tbl->length / 4))
			node_table_shrink(tbl);
*/
	}

	NODE_TABLE_CHECK(tbl, 1);

} // static void node_table_remove


/** \brief Test, ob eine Hashtafel leer ist.
 *
 * 	\param tbl Zeiger auf eine Hashtafel, die auf Leersein getestet
 * 		werden soll.
 * 	\return \c 0, falls die Tabelle Elemente enthält,
 * 		\c 1, falls die Tafel initialisiert ist, aber 0 Elemente enthält,
 * 		oder \c 2, falls die Tafel noch nicht initialisiert ist.
 */
static inline
int
node_table_empty(struct node_table_t *tbl)
{
	if (tbl == NULL)
		return 2;
	else if (tbl->item_count == 0)
		return 1;
	else
		return 0;
} // static int node_table_empty


/** \brief Suchen eines Knotens in einer Knotentabelle anhand seines
 *	content-Feldes.
 *	
 *	\param tbl Zeiger auf die Knotentabelle, in der gesucht wird.
 *	\param content Inhalt des zu findenden Knotens.
 *	\return Zeiger auf den gesuchten Knoten, falls er sich in der
 *		Tabelle befindet; \c NULL, sonst.
 */
static
struct node_t*
node_table_find(struct node_table_t *tbl, signed int content)
{
	if (tbl == NULL)
		return NULL;
	if (tbl->item_count == 0)
		return NULL;
	else {
		struct node_list_t *lst = tbl->first;
		while (lst != NULL) {
			if (lst->head->content == content)
				return lst->head;
			else
				lst = lst->right;
		}
		/* Absicherung, falls die Schleife komplett durchlaufen wird, ohne
		 * einen Knoten zu finden; sollte eigentlich nie vorkommen:
		 */
		return NULL;
	}
} // static struct node_t* node_table_find


/** \brief Entnimmt einen beliebigen Knoten aus einer Hashtabelle.
 *
 * 	Achtung: Da nicht auszuschließen ist, dass das gerade angeforderte
 * 	Element auch dasjenige ist, auf das gerade \c cursor zeigt, kann eine
 * 	verschränkte Benutzung von \c node_table_iter und \c node_table_pop
 * 	zu unvorhersehbaren und unerwünschten Ergebnissen führen!
 *  
 *	\param tbl Zeiger auf die Tabelle, aus der ein Knoten entnommen
 *		werden soll.
 *	\return Zeiger auf den ersten gefundenen Knoten.
 */
static
struct node_t*
node_table_pop (struct node_table_t *tbl)
{
	if (tbl == NULL) {
		return NULL;
	}
	else if (tbl->item_count == 0) {
		return NULL;
	}
	else {
		NODE_TABLE_CHECK(tbl, 0);

		/* Es wird das Element aus der Tabelle entfernt, auf das der Zeiger
		 * tbl->first verweist.
		 */
		struct node_t *node = tbl->first->head;

		/* Falls tbl->first keinen Nachfolgerknoten hat, wird soeben der
		 * einzige Knoten der ganzen Tabelle entnommen. Es gibt also auch
		 * keinen weiteren Knoten mit dem Hashwert node_hash, somit muss
		 * das entsprechende Feld auf NULL gesetzt werden.
		 */
		unsigned int node_hash = compute_hash(node, tbl->length);
		if (tbl->first->right == NULL) {
			tbl->items[node_hash] = NULL;
			tbl->item_count = 0;
		} // if (tbl->first->right == NULL)

		/* Falls es noch Knoten in der Tabelle gibt, aber keinen mehr mit
		 * dem Hashwert node_hash von node, muss tbl->items[node_hash]
		 * ebenfalls auf NULL gesetzt werden, und der first-Zeiger wird um ein
		 * Element weiter geschoben. Ansonsten kann auch der Zeiger auf
		 * tbl->items[i] um ein Element weitergeschoben werden.
		 */
		else {
			unsigned int next_hash =
				compute_hash(tbl->first->right->head, tbl->length);
			if ( next_hash != node_hash )
				tbl->items[node_hash] = NULL;
			else
				tbl->items[node_hash] = tbl->first->right;
			tbl->item_count--;
		} // if (tbl->first->right == NULL) ... else

		/* Nun wird das erste Element der Liste first an den Pool 
		 * zurückgegeben. */
		struct node_list_t *old_first = tbl->first;
		tbl->first = old_first->right;
		if (tbl->first != NULL)
			tbl->first->left = NULL;
		pool_release(old_first);

		/* Somit ist der Knoteneintrag jetzt sauber aus der Hashtabelle 
		 * entfernt, und der gewünschte Knoten kann als Rückgabewert
		 * geliefert werden.
		 */
		NODE_TABLE_CHECK(tbl, 1);
		return node;
	}

} // static struct node_t* node_table_pop


/** \brief Zurücksetzen des Cursors in einer Hashtabelle.
 *
 *  Die Funktion setzt den internen Cursor zurück, mit dessen Hilfe die
 *  Funktion \c node_table_iter eine Hashtabelle iteriert.
 *  Nach Aufruf von \c node_table_reset_cursor zeigt <tt>tbl->cursor</tt>
 *  auf das erste Element der Tabelle. Dabei bedeutet "erstes" Element
 *  nicht, dass der Knoten den niedrigsten Hashwert hat o.ä., sondern nur,
 *  dass der Zeiger <tt>tbl->first</tt> darauf verweist.
 *
 *  \param tbl Zeiger auf die Tabelle, deren Cursor zurückgesetzt werden
 *  	soll.
 */
static
void
node_table_reset_cursor(struct node_table_t *tbl)
{
	if (tbl == NULL)
		return;

	tbl->cursor = tbl->first;
} // static void node_table_reset_cursor


/** \brief Iterator über einer Hashtabelle.
 *
 * 	Die Funktion iteriert über einer Hashtafel. Bei Aufruf von
 * 	<tt>node_table_iter(tbl)</tt> wird ein Zeiger auf denjenigen Knoten
 * 	in \a tbl zurückgegeben, auf den der interne Zeiger \c cursor
 * 	verweist. Das Element wird dabei nicht aus der Tabelle entfernt, der
 * 	Zeiger wird um ein Element weitergeschoben. Ist das Ende der Tabelle
 * 	erreicht, zeigt der Cursor auf NULL, und dies ist auch der Rückgabewert.
 * 	Zurückgesetzt wird der Cursor mit der Funktion
 * 	\c node_table_reset_cursor.
 *
 * 	\param tbl Zeiger auf die Hashtafel, in der iteriert wird.
 * 	\return Zeiger auf den Knoten, auf den der Cursor zeigt.
 */
static
struct node_t*
node_table_iter(struct node_table_t *tbl)
{
	if (tbl == NULL)
		return NULL;

	if (tbl->item_count == 0)
		return NULL;

	if (tbl->cursor == NULL)
		return NULL;

	/* Wenn keiner der obigen Fälle eintritt, zeigt der Cursor auf ein
	 * gültiges Tabellenelement. Gib dies zurück und setze zuvor den Cursor
	 * auf das nächste Element, sofern es existiert.
	 */
	struct node_t *node = tbl->cursor->head;
	tbl->cursor = tbl->cursor->right;
	return node;
} // static struct node_t* node_table_iter


/** \brief Kopieren einer bestehenden Hashtabelle.
 *
 *	\param tbl Zeiger auf die Tabelle, von der eine Kopie erzeugt
 *  	werden soll.
 *	\return Zeiger auf die Kopie.
 */
static
struct node_table_t*
node_table_copy(struct node_table_t *tbl)
{
	if (tbl == NULL)
		return NULL;
	if (tbl->item_count == 0) {
		struct node_table_t *new_table = node_table_init();
		return new_table;
	}

	NODE_TABLE_CHECK(tbl, 0);

	struct node_table_t *new_tbl = malloc(sizeof(struct node_table_t));
	new_tbl->length = tbl->length;
	new_tbl->item_count = tbl->item_count;
	new_tbl->first = NULL;
	new_tbl->cursor = NULL;
	new_tbl->items =
		calloc(tbl->length, sizeof(struct node_list_t*));

	/* Nun klone alle Elemente aus tbl und trage die neu entstehenden
	 * Listeneinträge in new_tbl->items ein!
	 */
	tbl->cursor = tbl->first;
	struct node_t *node = NULL;
	struct node_list_t *lst_new = NULL;
	unsigned int node_hash = 0;
	while (tbl->cursor != NULL) {
		node = tbl->cursor->head;
		node_hash = compute_hash(node, tbl->length);

		/* Falls in new_tbl noch kein Knoten mit dem Hashwert node_hash
		 * existiert, wird das Feld new_tbl->items[node_hash] neu bestückt
		 * und der neue Eintrag auch mit dem Listenzeiger new_tbl->first
		 * verbunden:
		 */
		if (new_tbl->items[node_hash] == NULL) {
			lst_new = pool_get();
			lst_new->head = node;
			lst_new->left = NULL;
			lst_new->right = new_tbl->first;

			if (new_tbl->first != NULL)
				new_tbl->first->left = lst_new;
			new_tbl->first = lst_new;

			new_tbl->items[node_hash] = lst_new;
		}

		/* Falls es in new_tbl schon Einträge für andere Knoten mit dem
		 * Hashwert node_hash gibt, wird der Eintrag für node davor gesetzt.
		 * Dabei muss node auch mit einem ggf. vorhandenen linken Nachbarn
		 * verbunden werden. Falls der neue Nachfolger von node das bisherige
		 * first-Element von new_tbl war, muss auch dieser Zeiger angepasst
		 * werden.
		 */
		else {
			lst_new = pool_get();
			lst_new->head = node;
			lst_new->left = new_tbl->items[node_hash]->left;
			lst_new->right = new_tbl->items[node_hash];

			if (lst_new->left != NULL)
				lst_new->left->right = lst_new;
			lst_new->right->left = lst_new;

			if (new_tbl->first == lst_new->right)
				new_tbl->first = lst_new;

			new_tbl->items[node_hash] = lst_new;
		}

		tbl->cursor = tbl->cursor->right;
	} // while (tbl->cursor != NULL)

	NODE_TABLE_CHECK(new_tbl, 0);
	return new_tbl;
} // static struct node_table_t* node_table_copy


#ifdef USE_DEBUG_PRINTF
/** \brief Ausgabe eine Hashtafel von Knoten auf stderr.
 *
 *  Der Aufruf <tt>node_table_print (tbl)</tt> gibt zu Diagnosezwecken
 *  eine einfache textuelle Darstellung der Knotentabelle \a tbl aus.
 *
 *  Die Funktion wird hinter dem Makro \c NODE_TABLE_PRINT versteckt und
 *  nur ausgeführt, falls das Makro \c USE_DEBUG_PRINTF definiert ist.
 *  Ansonsten wird \c NODE_TABLE_PRINT zu <tt>(void(0))</tt> expandiert.
 *
 *  \param tbl Zeiger auf die Knotentabelle, die ausgegeben werden soll.
 */
static
void
node_table_print (struct node_table_t *tbl)
{
	if (tbl == NULL) {
		DEBUG_PRINTF("[]");
		return;
	}

	LIST_PRINT(tbl->first);
} // static void node_table_print


/** Falls das Makro \c USE_DEBUG_PRINTF nicht definiert ist, expandiert
 *  \c NODE_TABLE_PRINT zu <tt>(void)0</tt>.
 */
#   define NODE_TABLE_PRINT(x) (node_table_print((x)))
#else
#   define NODE_TABLE_PRINT(x) ((void)0)
#endif


/**************************************************************************
 * Datenstrukturen und Funktionen für die Verwaltung und Ausgabe         *
 * von Ableitungsbäumen:                                                 *
 **************************************************************************/

#ifdef PARSE_TREE

struct parse_forest_t; // forward

/** \brief Datenstruktur für Ableitungsbäume.
 *
 *  Ein (partieller) Ableitungsbaum besteht aus einer Wurzel,
 *  die die Integerrepräsentation eines Symbols enthält, und einer
 *  Liste der Nachkommen.
 *  Das Symbol \f$ \varepsilon \f$ wird durch den Wert 0 codiert, 
 *  Terminalsymbole durch positive Ganzzahlen, Nichtterminalsymbole durch
 *  negative, und zwar als "~<symbolnummer>".
 *  Ausserdem könnte man anhand der Position im Baum bestimmen, ob der
 *  aktuelle Knoten ein Terminal- oder Nichtterminalsymbol enthält:
 *  Blattknoten können nur Terminalsymbole inkl. \f$ \varepsilon \f$ 
 *  enthalten, innere Knoten nur Nichtterminalsymbole.
 *
 *  Wird nur compiliert, falls das Makro \c PARSE_TREE definiert ist.
 */
struct parse_tree_t {
	/// Wurzelsymbol
	signed long root;
	/// Liste der Nachfolgerknoten
	struct parse_forest_t *children;
};


/** \brief Datenstruktur für Wälder aus partiellen Ableitungsbäumen.
 *
 *  Ein Wald von (partiellen) Ableitungsbäumen besteht aus einem
 *  am weitesten rechts gelegenen Baum als Kopfelement und einem Zeiger
 *  auf den Restwald links.
 *
 *  Wird nur compiliert, falls das Makro \c PARSE_TREE definiert ist.
 */
struct parse_forest_t {
	/// Der am weitesten rechts stehende Baum des Waldes
	struct parse_tree_t *head;
	/// Zeiger auf den Restwald.
	struct parse_forest_t *right;
};


/** \brief Der Wald von partiellen Ableitungsbäumen wird über den
 * 		globalen Zeiger \c parse_forest angesprochen.
 *
 *  Wird nur compiliert, falls das Makro \c PARSE_TREE definiert ist.
 */
static struct parse_forest_t *parse_forest = NULL;


/** \brief Für die Ausgabe des Parsewaldes: globale Zählvariable für
 * 		die Knoten.
 *
 *  Wird nur compiliert, falls das Makro \c PARSE_TREE definiert ist.
 */
static unsigned long node_number = 0;


/** \brief Bei Leseschritt Erschaffung eines neuen partiellen
 * 		Ableitungsbaums.
 *
 *  Bei einem Leseschritt wird dem bisherigen Parsewald ein neuer Baum
 *  ganz rechts hinzugefügt. Dieser besteht zunächst nur aus einem
 *	Wurzelknoten, in dem das aktuell gelesene Symbol enthalten ist.
 *	\param forest Zeiger auf einen Zeiger (Call by Reference) auf den
 *		Parsewald, an den	der neue Baum rechts angefügt werden soll.
 *  \param symbol Integerdarstellung des gerade gelesenen Symbols.
 *
 *  Wird nur compiliert, falls das Makro \c PARSE_TREE definiert ist.
 */
static
void
parse_forest_read (struct parse_forest_t **forest, unsigned long symbol)
{
	DEBUG_PRINTF("  parse_forest_read %ld\n", symbol);

	// Kreiere einen neuen Baum für symbol:
	signed long sym = (signed long) symbol;
	struct parse_tree_t *new_tree = malloc(sizeof(struct parse_tree_t));
	new_tree->root = sym;
	new_tree->children = NULL;

	// Kreiere für tree einen neuen Wald-Eintrag:
	struct parse_forest_t *new_forest =
		malloc(sizeof(struct parse_forest_t));
		new_forest->head = new_tree;
		new_forest->right = NULL;

	// Falls der Wald bisher leer ist, wird der neue Baum sein erster!
	if (*forest == NULL) {
		*forest = new_forest;
	}

	/* Ansonsten hänge den neuen Baum direkt ans rechte Ende des Waldes
   * und mache den alten Wald zu seinem right-Element.
	 */
	else {
		new_forest->right = *forest;
		*forest = new_forest;
	} // if (*forest == NULL) ... else

} // static void parse_forest_read


/** \brief Bei Reduktion Zusammenfassen mehrerer partieller
 *  	Ableitungsbäume zu einem neuen.
 *
 *  Wenn ein Endknoten w anhand der Regel \f$ A \rightarrow \alpha \f$
 *  reduziert wird, muss das am weitesten rechts stehende Waldstück
 *  gefunden werden, dessen Wurzelsymbole konkateniert die Symbolfolge
 *  \f$ \alpha \f$ ergeben.
 *  An dieser Stelle wird ein neuer Wurzelknoten für \f$ A \f$ eingefügt
 *  und die Bäume für \f$ \alpha \f$ zu seinen Nachkommen gemacht.
 *  \param forest Zeiger auf einen Zeiger (Call by Reference) auf den
 *  	Parsewald, in dem	die zusammenzufassenden Bäume liegen.
 *	\param item_number Nummer des Items, anhand dessen die Reduktion
 *		durchgeführt wurde.
 *
 *  Wird nur compiliert, falls das Makro \c PARSE_TREE definiert ist.
 */
static
void
parse_forest_reduce (struct parse_forest_t **forest, int item_number)
{
	DEBUG_PRINTF("  parse_forest_reduce: Item %d, Regel %d\n",
			item_number, items[item_number].rule_number);

	/* Als erstes muss die passende Regel im Array rules aller Grammatikregeln
	 * gefunden werden. Beachte dabei: Regel Nr. 1 hat im Array den Index 0,
	 * usw. Das liegt daran, dass für die eigentliche Regel 0, S' -> S,
	 * kein Eintrag existiert, denn anhand ihrer darf niemals reduziert
	 * werden!
	 */
	struct rule_t rule = rules[items[item_number].rule_number - 1];

	/* Falls eine Reduktion von Epsilon durchgeführt wurde, muss ganz
	 * rechts ein neuer Baum angefügt werden:
	 */
	if (rule.rhs_count == 0) {
		DEBUG_PRINTF("    Kreiere Epsilon-Baum\n");
		
		// Erzeuge einen neuen Teilbaum für die Epsilon-Reduktion:
		struct parse_tree_t *eps = malloc(sizeof(struct parse_tree_t));
			eps->root = 0;
			eps->children = NULL;
		struct parse_tree_t *eps_root = malloc(sizeof(struct parse_tree_t));
			eps_root->root = ~(rule.lhs);
			eps_root->children = malloc(sizeof(struct parse_forest_t));
				eps_root->children->head = eps;
				eps_root->children->right = NULL;

		/* Hänge den neuen Baum mit Wurzel eps_root ganz rechts an den
     * globalen Parsewald:
		 */
		struct parse_forest_t *new_forest =
			malloc(sizeof(struct parse_forest_t));
			new_forest->head = eps_root;
			new_forest->right = *forest;
		*forest = new_forest;

	} // if (rule.rhs_count == 0)

	/* Falls die rechte Seite der reduzierten Regel alpha != Epsilon ist,
	 * so muss im Wald das am weitesten rechts liegende Waldstück gefunden
   * werden, dessen Wurzeln die Symbole aus alpha enthalten. Diese
	 * Bäume werden dann unter einer neuen Wurzel zusammengefasst.
	 * Da der Wald von rechts nach links durchmustert wird, müssen
	 * die Symbole auf der rechten Seite von rule auch von rechts nach links
	 * betrachtet werden!
	 */
	else {
		DEBUG_PRINTF("    Suche zusammenzufügende Bäume\n");
		// Universelle Zählvariable:
		int i = 0;

		// Flag, ob ein Baum gefunden wurde, dess Wurzel zum rechten Symbol
		// der rechten Regelseite passt:
		int found_right = 0;

		// Flag, ob die Bäume gefunden wurden, die zur rechten Regelseite passen
		int found_string = 0;

		// Temporärer Zeiger auf den aktuellen betrachteten Waldabschnitt:
		struct parse_forest_t *tmp_forest = *forest;

		/* Zuerst muss der Fall betrachtet werden, dass das gesuchte Wald-
		 * stück am rechten Rand des Gesamtwaldes liegt. In diesem Fall gibt
		 * es nämlich keinen weiter rechts gelegenen Baum, dessen right-Zeiger
		 * nach der Reduktion angepasst werden muss!
		 */
		DEBUG_PRINTF("      Rechter Baum:\n");
		DEBUG_PRINTF("        Symbol %d: %ld ... ",
				rule.rhs_count, rule.rhs[rule.rhs_count - 1]);
		if (tmp_forest->head->root == rule.rhs[rule.rhs_count - 1]) {
			DEBUG_PRINTF("passt\n");

			/* Nun wird getestet, ob die weiteren Baumwurzeln zu den restlichen
			 * Symbolen in rule.rhs passen.
			 */
			for (i=rule.rhs_count - 2; i>=0; i--) {
				DEBUG_PRINTF("        Symbol %d: %ld ... ",
						i+1, rule.rhs[i]);

				tmp_forest = tmp_forest->right;

				/* Falls es weniger Baumwurzeln als Symbole in der reduzierten
				 * Regel gibt, liegt ein Programmierfehler vor:
				 */
				assert(tmp_forest != NULL);

				/* Falls das aktuelle Paar aus Symbol und Baumwurzel zusammen
				 * passt, wird die Schleife fortgesetzt, ansonsten abgebrochen:
				 */
				if (tmp_forest->head->root == rule.rhs[i]) {
					DEBUG_PRINTF("passt zu Wurzel %ld!\n",
							tmp_forest->head->root);
				}
				else {
					DEBUG_PRINTF("passt nicht zu Wurzel %ld!\n",
							tmp_forest->head->root);
					break;
				}

			} // for (i=rule.rhs_count - 2; i>=0; i--)

			/* Im Erfolgsfall ist die FOR-Schleife bis i = -1 durchgelaufen, und
			 * tmp_forest zeigt auf den an weitesten links stehenden Eintrag des
			 * gesuchten Waldstückes. Danach kann noch ein Restwald folgen,
			 * der später noch zum right-Element des neu erschaffenen Baumes
			 * gemacht werden muss.
			 */
	
			/* Falls der passende Waldabschnitt gefunden wurde, fasse die Bäume
			 * unter einer neuen Wurzel zusammen und pflanze den Resultatbaum an
			 * der entsprechenden Stelle wieder im Wald ein.
			 */
			if (i == -1) {

				DEBUG_PRINTF("      Passendes Waldstück gefunden!\n");
				found_string = 1;

				/* Das gefundene Waldstück, beginnend mit dem rechten Baum des
				 * Gesamtwaldes, stellt die Kindknoten der neu zu erschaffenden
				 * Baumwurzel dar.
				 */
				struct parse_forest_t *first_child = *forest;

				/* Erschaffe nun eine neue Baumwurzel für das Symbol rule.lhs: */
				struct parse_tree_t *new_tree =
					malloc(sizeof(struct parse_tree_t));
					new_tree->root = ~(rule.lhs);
					new_tree->children = first_child;

				/* Erstelle einen neuen Waldeintrag für den neuen Baum und
				 * mache ihn zum ersten Eintrag des Gesamtwaldes.
				 */
				struct parse_forest_t *new_forest =
					malloc(sizeof(struct parse_forest_t));
					new_forest->head = new_tree;
					new_forest->right = tmp_forest->right;
				*forest = new_forest;

				/* Der letzte Kindeintrag von new_tree, auf den der Zeiger
				 * tmp_forest noch zeigt, muss noch von seinem bisherigen
				 * Nachfolgereintrag im Gesamtwald getrennt werden.
				 * Anschließend kann der Zeiger tmp_forest verworfen werden!
				 */
				tmp_forest->right = NULL;
				tmp_forest = NULL;

			} // if (i == -1)

			/* Falls das passende Waldstück noch nicht vom Waldanfang aus
			 * gefunden wurde, muss tmp_forest zurückgesetzt werden, damit
			 * die Weitersuche an der korrekten Stelle startet!
			 */
			else {
				DEBUG_PRINTF("      Passendes Waldstück NICHT gefunden!\n");
				tmp_forest = *forest;
				found_string = 0;
			} // if (i == -1) else

		} // if (rechtes Symbol passt zu rechter Baumwurzel)
	
		else {
			DEBUG_PRINTF("passt nicht zu %ld!\n",
					tmp_forest->head->root);
		} // if (rechtes Symbol passt zu rechter Baumwurzel) else

		/* Sollte das relevante Waldstück nicht am rechten Rand des Waldes
		 * beginnen, wird weitergesucht, bis entweder des Waldstück gefunden
		 * ist oder der Wald komplett erfolglos durchmustert wurde.
		 * Achtung: Im Erfolgsfall muss man sich noch merken, welcher
		 * Waldeintrag der direkte Vorgänger des rechtesten passenden Baumes
		 * ist, um dessen right-Zeiger am Ende korrekt umbiegen zu können!
		 */
		while (found_string == 0) {

			/* Suche das rechte Symbol von rule.rhs in tmp_forest: */
			found_right = 0;
			while ( (tmp_forest->right != NULL) && (found_right == 0) ) {
				DEBUG_PRINTF("      Nächster Baum:\n");
				DEBUG_PRINTF("        Symbol %d: %ld ... ",
						rule.rhs_count, rule.rhs[rule.rhs_count - 1]);

				/* Sobald ein Baum gefunden wurde, der zum rechten Symbol in
				 * rule.rhs passt, kann die Schleife beendet werden:
				 */
				if (tmp_forest->right->head->root == rule.rhs[rule.rhs_count - 1])
				{
					DEBUG_PRINTF("passt zu Wurzel %ld!\n",
							tmp_forest->right->head->root);
					found_right = 1;
				}

				/* Ansonsten wird im nächsten Schleifendurchlauf der
				 * nächste Baum des Waldes getestet:
				 */
				else {
					DEBUG_PRINTF("passt nicht zu Wurzel %ld!\n",
							tmp_forest->right->head->root);
					tmp_forest = tmp_forest->right;
				}

			} // while ( (tmp_forest->right != NULL) && (found_right == 0) )

			/* Falls tmp_forest ganz durchmustert wurde, ohne das rechte Symbol
			 * der reduzierten Regel zu finden, liegt ein Programmierfehler vor!
			 */
			assert(tmp_forest->right != NULL);

			/* Falls die Suche nach einem passenden ersten Waldeintrag
			 * erfolgreich war, zeigt tmp_forest nun auf den direkten Vorgänger
			 * des rechtesten Baumes im Wald, der zu rule passen könnte. Diese
			 * spezielle Position im Wald muss aus zwei Gründen in einem
			 * eigenen Zeiger candidate_before festgehalten werden:
			 * 1.) Im Erfolgsfall muss ich den right-Zeiger von candidate_before
			 *     auf die Wurzel des neuen Baumes umbiegen.
			 * 2.) Im Falle eines Misserfolges muss tmp_forest auf diese Stelle
			 *     gesetzt werden, damit die Suche von dort aus fortgesetzt
			 *     werden kann.
			 */ 
			struct parse_forest_t *candidate_before = tmp_forest;

			/* Nun soll tmp_forest auf den ersten Baum des aktuelle
			 * betrachteteten Waldstückes zeigen, damit die weitere Suche
			 * analog zum obigen Fall stattfinden kann:
			 */
			tmp_forest = tmp_forest->right;

			/* Nun muss getestet werden, ob auch die Wurzeln der Nachfolgerbäume
			 * zu den restlichen Symbole in rule.rhs passen:
			 */
			for (i=rule.rhs_count - 2; i>=0; i--) {
				DEBUG_PRINTF("        Symbol %d: %ld ... ",
						i+1, rule.rhs[i]);

				tmp_forest = tmp_forest->right;

				/* Falls es weniger Baumwurzeln gibt als Symbole in der reduzierten
				 * Regel, liegt ein Programmierfehler vor:
				 */
				assert(tmp_forest != NULL);

				/* Falls das aktuelle Paar aus Symbol und Baumwurzel zusammen
				 * passt, wird die Schleife fortgesetzt, ansonsten abgebrochen:
				 */
				if (tmp_forest->head->root == rule.rhs[i]) {
					DEBUG_PRINTF("passt zu Wurzel %ld!\n",
							tmp_forest->head->root);
				}
				else {
					DEBUG_PRINTF("passt nicht zu Wurzel %ld!\n",
							tmp_forest->head->root);
					break;
				}

			} // for (i=rule.rhs_count - 2; i>=0; i--)

			/* Im Erfolgsfall ist die FOR-Schleife bis i = -1 durchgelaufen, und
			 * tmp_forest zeigt auf den am weitesten links stehenden Eintrag des
			 * gesuchten Waldstückes. Danach dann noch ein Restwald folgen,
			 * der später noch zum right-Element des neu erschaffenen Baumes
			 * gemacht werden muss.
			 */

			/* Falls der passende Waldabschnitt gefunden wurde, fasse die Bäume
			 * unter einer neuen Wurzel zusammen und pflanze den Resultatbaum an
			 * der entsprechenden Stelle wieder im Wald ein.
			 */
			if (i == -1) {

				DEBUG_PRINTF("      Passendes Waldstück gefunden!\n");
				found_string = 1;

				/* Das gefundene Waldstück, rechts beginnend mit dem Eintrag
				 * candidate_before->right, stellt die Kindknoten der neu zu
				 * erschaffenden Baumwurzel dar:
				 */
				struct parse_forest_t *first_child = candidate_before->right;

				/* Erschaffe nun eine neue Baumwurzel für das Symbol rule.lhs: */
				struct parse_tree_t *new_tree =
					malloc(sizeof(struct parse_tree_t));
					new_tree->root = ~(rule.lhs);
					new_tree->children = first_child;

				/* Erschaffe einen neuen Waldeintrag für den neuen Baum und
				 * pflanze ihn zwischen candidate_before und tmp_forest->right ein:
				 */
				struct parse_forest_t *new_forest =
					malloc(sizeof(struct parse_forest_t));
					new_forest->head = new_tree;
					new_forest->right = tmp_forest->right;
				candidate_before->right = new_forest;

				/* Der letzte Kindeintrag von new_tree, auf den der Zeiger
				 * tmp_forest noch zeigt, muss noch von seinem bisherigen
				 * Nachfolgereintrag im Gesamtwald getrennt werden.
				 * Anschließend kann der Zeiger tmp_forest verworfen werden!
				 */
				tmp_forest->right = NULL;
				tmp_forest = NULL;

			} // if (i == -1)

			/* Falls das passende Waldstück noch nicht gefunden wurden, muss
			 * tmp_forest zurückgesetzt werden, damit die weitere Suche
			 * an der richtigen Stelle startet:
			 */
			else {
				DEBUG_PRINTF("      Passendes Waldstück NICHT gefunden!\n");
				tmp_forest = candidate_before->right;
				found_string = 0;
			} // if (i == -1) else

		} // while (found_string == 0)

	} // if (rule.rhs_count == 0) else

} // static void parse_forest_reduce


/** \brief Ausgabe eines partiellen Ableitungsbaums auf \c stdout.
 *
 *  Der Funktionsaufruf <tt>parse_tree_print(tree)</tt> gibt die Wurzel des
 *  Baumes \a tree aus und rekursiv alle seine Kindknoten.
 *  Das Ausgabeformat ist konform zum Pogramm \c dot aus dem
 *  Paket \c graphviz.
 *  \param tree Zeiger auf den Baum, der ausgegeben werden soll. 
 *  \param father Nummer des Vaterknotens.
 *
 *  Wird nur compiliert, falls das Makro \c PARSE_TREE definiert ist.
 */
static
void
parse_tree_print (struct parse_tree_t *tree, unsigned long father)
{
	if (tree != NULL) {

		node_number++;
		unsigned long my_number = node_number;

		/* Für ein Terminalsymbol, gib seine Nummer aus; für ein
		 * Nichtterminalsymbol, gib seinen Namen aus der
		 * Grammatikdatei aus:
		 */
		if (tree->root >= 0)
			printf("\t\t%ld [label=\"Token %ld\", shape=box];\n",
					node_number, tree->root);
		else
			printf("\t\t%ld [label=\"%s\"];\n",
				 node_number, nonterminals[~(tree->root)-1].name);

		/* Erschaffe eine Kante zwischen Vater- und Kindknoten: */
		printf("\t\t\t%ld -> %ld;\n", father, node_number);

		/* Drucke die (von rechts nach links gespeicherten) Kinder von
		 * links nach rechts aus). Dazu müssen sie allerdins zuerst
		 * umgeordnet werden.
		 */
		struct parse_forest_t *children = tree->children;
		struct parse_forest_t *print_children = NULL;
		struct parse_forest_t *tmp = NULL;
		while (children != NULL) {
			tmp = malloc(sizeof(struct parse_forest_t));
			tmp->head = children->head;
			tmp->right = print_children;
			print_children = tmp;
			children = children->right;
		} // while (children != NULL)

		while (print_children != NULL) {
			parse_tree_print(print_children->head, my_number);
			print_children = print_children->right;
		}

	} // if (tree != NULL)
} // static void parse_tree_print


/** \brief Rekursive Ausgabe eines Baumes im Parsewald und zuvor aller
 *  	linker Nachbarn.
 *
 *  Der Aufruf <tt>parse_forest_print_rec forest</tt> gibt zuerst
 *  rekursiv alle weiter links im Wald gelegenen Bäume aus, danach
 *  \c forest->head.
 *  \param forest Zeiger auf den rechtesten Baum des auszugebenden
 *  	Waldstückes.
 *
 *  Wird nur compiliert, falls das Makro \c PARSE_TREE definiert ist.
 */
static
void
parse_forest_print_rec (struct parse_forest_t *forest)
{
	if (forest == NULL) {
		return;
	}
	else {
		/* Sofern vorhanden, gib erst rekursiv alle weiter links im Wald
		 * gelegenen Bäume aus. Danach, gib den rechten Baum forest->head
		 * aus.
		 */
		parse_forest_print_rec(forest->right);

		node_number++;
		printf("\n\t{\n");

		/* Unsichtbarer Startknoten für den Baum forest->right: */
		printf("\t\t%ld [shape=none, label=\"\"];\n", node_number);
		parse_tree_print(forest->head, node_number);

		printf("\t}\n");
	}
} // stativ void parse_forest_print_rec


/** \brief Ausgabe eine Parsewaldes.
 *
 *  Der Aufruf <tt>parse_forest_print(forest)</tt> gibt die Bäume des
 *  Waldes \a forest von links nach rechts aus.
 *  Bei Aufruf am Ende eines erfolgreichen Parsevorganges besteht dieser
 *  Wald aus einem einzigen Baum, der die eindeutige Rechtsableitung der
 *  Eingabe repräsentiert.
 *  Ausgabeformat ist ein dot-konforme Eingabe.
 *  \param forest Zeiger auf den auszugebenden Wald.
 *
 *  Wird nur compiliert, falls das Makro \c PARSE_TREE definiert ist.
 */
static
void
parse_forest_print (struct parse_forest_t *forest)
{
	if (forest == NULL) {
		return;
	}
	else {
		node_number++;
		printf("{");
		
		node_number++;
		/* Sofern vorhanden, gib erst rekursiv alle weiter links gelegenen
		 * Bäume aus. Danach, gib den rechten Baum forest->head aus.
		 */
		parse_forest_print_rec(forest);

		printf("}\n");
	}
} // static void parse_forest_print

/** Falls das Makro \c PARSE_TREE nicht definiert ist, expandiert
 *  \c PARSE_FOREST_READ zu <tt>(void)0</tt>.
 */
#   define PARSE_FOREST_READ(x, y) (parse_forest_read((x), (y)))
/** Falls das Makro \c PARSE_TREE nicht definiert ist, expandiert
 *  \c PARSE_FOREST_REDUCE zu <tt>(void)0</tt>.
 */
#   define PARSE_FOREST_REDUCE(x, y) (parse_forest_reduce((x), (y)))
/** Falls das Makro \c PARSE_TREE nicht definiert ist, expandiert
 *  \c PARSE_FOREST_PRINT zu <tt>(void)0</tt>.
 */
#   define PARSE_FOREST_PRINT(x) (parse_forest_print((x)))


/** \brief Sauberer Abschluss der Ableitungsbaum-Ausgabe.
 *
 *  Falls das Makro \c PARSE_TREE gesetzt ist, erzeugt der Parser
 *  bei erfolgreichem Parsevorgang einen Ableitungsbaum für die
 *  Eingabe im \c dot- Format. Falls \c PARSE_TREE_PHASES gesetzt
 *  ist, wird sogar für jede Phase der aktuelle Wald auf partiellen
 *  Ableitungsbäumen ausgegeben. Damit die Ausgabe auf jeden Fall
 *  eine gültige Eingabe für \c dot darstellt, muss sie mit einer
 *  geschweiften Klammer abgeschlossen werden.
 */
static
void
End (void)
{
	printf("}");
}


#else

#   define PARSE_FOREST_READ(x, y) ((void)0)
#   define PARSE_FOREST_REDUCE(x, y) ((void)0)
#   define PARSE_FOREST_PRINT(x) ((void)0)

#endif


/**************************************************************************
 * Funktionen - Parser:                                                   *
 **************************************************************************/

/** \brief Speicherbereinigung am Ende des Parsevorgangs.
 *
 *	Die Funktion \c clean wird am Ende der Funktion \c yyparse aufgerufen,
 *	um den Speicher wieder freizugeben, der durch den Graphen und die
 *	verschiedenen anderen Knotenmengen belegt wird.
 */
static
void
clean (void)
{
	// Zeiger auf den aktuell behandelten Knoten:
	struct node_t *node = NULL;
	// Der aktuell betrachtete Nachfolger:
	struct node_t *succ_node = NULL;
	// Der aktuell betrachtete Vorgänger:
	struct node_t *pred_node = NULL;
	// Menge der entfernbaren Nachfolgerknoten:
	struct node_table_t *removable_nodes = node_table_init();

	DEBUG_PRINTF("\nStarte die abschliessende Speicherbereinigung!\n");
	
	/* Zuerst werden alle Knotenmengen geleert, die ausser dem Graphen
	 * verwaltet werden. Da hier nur Knoten eingetragen sein sollen, die sich
	 * auch tatsächlich noch im Graphen befinden, werden hier _nur_ die
	 * Tabelleneinträge entfernt, nicht aber die tatsächlichen Knoten!
	 */
	node_table_free(expandable_nodes);
	expandable_nodes = NULL;
	node_table_free(expanded_nodes);
	expanded_nodes = NULL;
	node_table_free(expanded_nonterminals);
	expanded_nonterminals = NULL;
	node_table_free(readable_nodes);
	readable_nodes = NULL;
	node_table_free(reducible_nodes);
	reducible_nodes = NULL;

	/* Anschliessend wird der Graph durchmustert, um alle darin befindlichen
	 * Knoten zu löschen. Dazu wird als erstes für alle Startknoten, d.h.
	 * für jeden Knoten node in der Liste graph der Anfangsknoten, die Menge
	 * seiner Nachfolger betrachtet.
	 * Für jeden Knoten succ_node in der Nachfolgermenge wird node aus
	 * succ_node->predecessors gelöscht. Anschliessend wird succ_node in
	 * die Menge der löschbaren Knoten aufgenommen.
	 */
	while (graph != NULL) {
		node = list_pop(&graph);

		// Behandlung der Nachfolgerknoten
		while (node_table_empty(node->successors) == 0) {
			succ_node = node_table_pop(node->successors);

			// Entferne node aus der Vorgängermenge von succ_node
			node_table_remove(succ_node->predecessors, node);

			/* Entferne die Kanten zwischen succ_node und all seinen direkten
			 * Vorgängern, d.h. Entferne succ_node aus den Nachfolgermengen
			 * all seiner Vorgänger und entferne alle Vorgänger aus der
			 * Vorgängermenge von succ_node:
			 */
			while (node_table_empty(succ_node->predecessors) == 0) {
				pred_node = node_table_pop(succ_node->predecessors);
				node_table_remove(pred_node->successors, succ_node);
			}

			// Trage succ_node bei den löschbaren Knoten ein:
			node_table_add(removable_nodes, succ_node);
		}
		node_table_free(node->successors);
			
		/* Nachdem er von allen Nachfolgern gekappt wurde, kann der Knoten
		 * node gelöscht werden:
		 */
		free(node);
	} // while (graph != NULL)

	/* Nun werden analog alle Knoten in der Tabelle deleteable_nodes
	 * behandelt. Auf Vorgänger im Graphen muss dabei keine Rücksicht
	 * genommen werden, weil beim Eintragen eines Knotens in die Menge
	 * der Löschkandidaten alle Verbindungen zu Vorgängern entfernt wurden.
	 */
	while (node_table_empty(removable_nodes) == 0) {
		node = node_table_pop(removable_nodes);

		// Behandlung der Nachfolgerknoten:
		while (node_table_empty(node->successors) == 0) {
			succ_node = node_table_pop(node->successors);

			// Entferne node aus der Vorgängerliste von succ_node:
			node_table_remove(succ_node->predecessors, node);

			/* Entferne die Kanten zwischen succ_node und all seinen direkten
			 * Vorgängern, d.h. Entferne succ_node aus den Nachfolgermengen
			 * all seiner Vorgänger und entferne alle Vorgänger aus der
			 * Vorgängermenge von succ_node:
			 */
			while (node_table_empty(succ_node->predecessors) == 0) {
				pred_node = node_table_pop(succ_node->predecessors);
				node_table_remove(pred_node->successors, succ_node);
			}

			// Trage succ_node bei den löschbaren Knoten ein:
			node_table_add(removable_nodes, succ_node);
		}

		/* Nachdem alle Nachfolger behandelt sind, kann der
		 * Knoten node gelöscht werden:
		 */
		node_table_free(node->predecessors);
		node_table_free(node->successors);
		free(node);
		
	} // end while (node_table_empty(removable_nodes) == 0)

#ifndef DEBUG_ALL
	/* Nun muss noch der Pool für Listenelemente bereinigt werden: */
	struct node_list_t *pool_tmp = NULL;
	while (pool != NULL) {
		pool_tmp = pool;
		pool = pool_tmp->right;
		free(pool_tmp);
	} // while (pool != NULL)
#endif

	DEBUG_PRINTF("Speicherbereinigung abgeschlossen!\n");
	
} // void clean


/** \brief Kapselung des yylex-Aufrufs.
 *
 *  Ruft \c yylex auf, sofern das letzte gelesene Token > 0 war,
 *  sonst wird direkt 0 zurückgegeben.
 */
static
int
call_lexer (void)
{
	int result = 0;
	if (eof_reached == 0) {
		result = yylex();
		if (result == 0) {
			eof_reached = 1;
		}
	}
	
	return result;
} // int call_lexer


/** \brief Die Graphjustierung entfernt einen Knoten samt Vorgängern.
 *
 *	Die Funktion \c graph_adjust entfernt einen Endknoten aus dem
 *	Graphen	sowie alle seine Vorgänger, die keine sonstigen Nachfolger
 *	haben.
 *
 *	\param node Zeiger auf einen Zeiger (Call by Reference) auf den zu
 *		entfernenden Knoten.
 */
static
void
graph_adjust (struct node_t **node)
{
	assert((*node) != NULL);

	/* Der Endknoten *node hat zwar keine Nachfolger, wohl aber in der
	 * Regel Vorgänger, die noch auf Löschbarkeit überprüft werden
	 * müssen.
	 * Vorgängerknoten, die als löschbar identifiziert werden, werden in
	 * einer Tabelle gesammelt:
	 */
	struct node_table_t *preds = node_table_init();
	// Zeiger auf den aktuell untersuchten Vorgängerknoten:
	struct node_t *pred_node = NULL;
	// Zeiger auf den aktuell zu löschenden Knoten:
	struct node_t *kill_node = NULL;
	// Flag für Erfolg oder Misserfolg einer Knotenlöschung:
	int remove_success = 0;

	/* Für alle Vorgänger von *node: Entnimm sie aus der predecessor-
	 * Menge von node und entferne node aus ihren successor-Mengen.
	 */
	while (node_table_empty((*node)->predecessors) == 0) {
		/* Nimm den nächsten Vorgängerknoten von *node: */
		pred_node = node_table_pop((*node)->predecessors);

		/* Entferne *node aus der successor-Menge von pred_node: */
		node_table_remove(pred_node->successors, *node);

		/* Falls pred_node keine anderen Nachfolger als *node im Graph hatte,
		 * gehört er selbst zur Menge der entfernbaren Knoten.
		 */
		if (node_table_empty(pred_node->successors) != 0) {
			node_table_add(preds, pred_node);
		}
	}

	/* Bevor der Knoten endgültig entfernt werden kann, muss er ggf. noch aus
	 * einigen Knotentabellen entfernt werden.
	 * Die Funktion graph_adjust wird nur mit Knoten aufgerufen, die nicht
	 * für das Startitem [S' -> .S] oder das Accept-Item [S' -> S.] stehen.
	 * Jeder andere Itemknoten kann in einer der Tabellen für expandierbare,
	 * expandierte, lesbare oder reduzierbare Knoten enthalten sein:
	 */
	if ( (*node)->content >= 0) {
		node_table_remove(expandable_nodes, *node);
		node_table_remove(expanded_nodes, *node);
		node_table_remove(reducible_nodes, *node);
		node_table_remove(readable_nodes, *node);
	}
	/* Ein Nichtterminalknoten kann sich höchstens in der Liste der
	 * expandierten Nichtterminalknoten befinden:
	 */
	else {
		node_table_remove(expanded_nonterminals, *node);
	}

	/* Nachdem alle Verweise von und auf *node entfernt sind, kann node
	 * selbst aus dem Speicher entfernt werden:
	 */
	node_table_free((*node)->predecessors);
	node_table_free((*node)->successors);
	free(*node);
	*node = NULL;


	/* Nun müssen noch die Vorgänger analog bearbeitet werden. */
	while (node_table_empty(preds) == 0) {

		// Entnimmt einen zu löschenden Knoten aus der Menge:
		kill_node = node_table_pop(preds);

		/* Für alle Vorgänger von kill_node: Entnimm sie aus der
		 * predecessor-Menge von kill_node und entferne kill_node aus
		 * ihren successor-Mengen.
		 */
		while (node_table_empty(kill_node->predecessors) == 0) {
			/* Nimm den nächsten Vorgängerknoten von kill_node: */
			pred_node = node_table_pop(kill_node->predecessors);

			/* Entferne den Eintrag für kill_node aus der successor-
			 * Menge von temp_node:
			 */
			node_table_remove(pred_node->successors, kill_node);

			/* Falls pred_node keine anderen Nachfolger als kill_node im Graph
			 * hatte, gehört er jetzt selbst zur Menge der entfernbaren Knoten.
			 */
			if (node_table_empty(pred_node->successors) != 0)
				node_table_add(preds, pred_node);
		} // while (node_table_empty(kill_node->predecessors) == 0)

		/* Bevor der Knoten endgültig entfernt werden kann, muss es ggf. noch 
		 * aus einigen Tabellen entfernt werden.
		 * Ein Itemknoten kann nicht für das Accept-Item [S' -> S.] stehen,
		 * da dieses keine Nachfolger hätte. Steht kill_node für das Startitem
		 * [S' -> .S], so kann der Graph ansonsten nur noch aus Knoten für das
		 * Accept-Item bestehen.
		 */
		if ((kill_node)->content == startitem_number)
			remove_success = list_remove(&graph, kill_node);

		/* Jeder andere Itemknoten kann in einer der Tabellen für
		 * expandierbare, expandierte, lesbare oder reduzierbare Knoten
		 * enthalten sein.
		 */	
		else if ((kill_node)->content >= 0) {
			node_table_remove(expandable_nodes, kill_node);
			node_table_remove(expanded_nodes, kill_node);
			node_table_remove(reducible_nodes, kill_node);
			node_table_remove(readable_nodes, kill_node);
		}

		/* Ein Nichtterminalknoten kann sich höchstens in der Menge der
		 * expandierten Nichtterminalknoten befinden:
		 */
		else {
			node_table_remove(expanded_nonterminals, kill_node);
		}

		/* Nachdem alle Verweise von und auf kill_node entfernt sind, kann
		 * der Knoten selbst nun aus dem Speicher entfernt werden:
		 */
		node_table_free(kill_node->predecessors);
		node_table_free(kill_node->successors);
		free(kill_node);
		kill_node = NULL;

	} // while (NULL != preds)

} // void graph_adjust


/** \brief Durchführung eines Expansionsschrittes.
 *
 *	Die Funktion \c expand führt für alle expandierbaren Endknoten des
 *	aktuellen Graphen den Expansionsschritt durch.
 *	Sollte in der aktuellen Phase bereits eine Expansion für eines der
 *	hier betrachteten Nichtterminalsymbole stattgefunden haben, so wird nur
 *	eine Kante vom expandierbaren Knoten zum Knoten für dieses
 *	Nichtterminalsymbol eingefügt.
 *	Ansonsten wird solch ein Nichtterminalknoten neu erzeugt, und es werden
 *	Items für alle Alternativen hinzugefügt, die zum aktuellen Zeitpunkt
 *	zu einer akzeptierenden Berechnung führen könnten.
 */
static
void
expand (void)
{
	// Universelle Laufvariable:
	int i = 0;
	// Flag, ob die aktuelle Expansion gelungen ist:
	int expand_ok = 0;
	// Der jeweils aktuelle zu expandierende Endknoten:
	struct node_t *node = NULL;
	// Der jeweils aktuelle Nichtterminalknoten:
	struct node_t *current_nt_node = NULL;
	// Der aktuell neu erzeugte Knoten:
	struct node_t *new_node = NULL;
	// Inhalt des Arrayeintrages für das aktuelle Nichtterminalsymbol:
	struct nonterminal_t nt;

	if (node_table_empty(expandable_nodes) != 0) {
		DEBUG_PRINTF("  Keine expandierbaren Knoten gefunden!\n");
		return;
	}

	while (node_table_empty(expandable_nodes) == 0) {

		/* Entnimm einen Knoten aus der Menge der expandierbaren Knoten und
		 * füge ihn in die Menge der Knoten ein, die in der aktuellen Phase
		 * expandiert wurden:
		 */
		DEBUG_PRINTF("  Menge expandierbarer Knoten:\n");
		DEBUG_PRINTF("  ");
		NODE_TABLE_PRINT(expandable_nodes);

		node = node_table_pop(expandable_nodes);
		
		DEBUG_PRINTF("  ");
		NODE_TABLE_PRINT(expandable_nodes);
		DEBUG_PRINTF("  Liste expandierter Knoten:\n");
		DEBUG_PRINTF("  ");
		NODE_TABLE_PRINT(expanded_nodes);

		node_table_add(expanded_nodes, node);

		DEBUG_PRINTF("  ");
		NODE_TABLE_PRINT(expanded_nodes);
		DEBUG_PRINTF("  Expandiere Knoten: %ld\n", node->content);

		expand_ok = 0;

		/* Sei [A -> alpha1 . B alpha2] das aktuell betrachtete expandierbare
		 * Item. Falls eine Expansion anhand B in dieser Phase schon
		 * durchgeführt wurde, müssen keine neuen Knoten in den Graphen
		 * eingefügt werden, sondern nur eine Kante vom aktuellen Item zum 
		 * Nichtterminalknoten für B.
		 */
		unsigned long current_nonterminal = items[node->content].dot_symbol;
		current_nt_node =
			node_table_find(expanded_nonterminals, ~current_nonterminal);
		if (current_nt_node != NULL) {
			DEBUG_PRINTF("  Anhand des NT-Symbols %ld wurde schon expandiert!\n",
					current_nonterminal);
			expand_ok = 1;
		}

		/* Falls die Expansion jedoch zum ersten Mal für diese Phase
		 * stattfindet, müssen ein Nichtterminalknoten für B und einige
		 * Itemknoten für seine Alternativen [B -> beta_j] in den Graphen
		 * eingefügt werden.
		 */
		else {
			/* Erzeuge einen Nichtterminalknoten für current_nonterminal,
			 * verbinde ihn aber noch nicht mit anderen Knoten im Graphen.
			 */
			DEBUG_PRINTF("    Erzeuge Nichtterminalknoten für Symbol %ld\n",
					current_nonterminal);
			current_nt_node = create_node(~current_nonterminal);

			/* Nun müssen alle Alternativen für current_nonterminal untersucht
			 * und ggf. Knoten dafür in den Graphen eingefügt werden.
			 */
			nt = nonterminals[current_nonterminal-1];
			DEBUG_PRINTF("    Starte Test der %d Alternativen\n", nt.item_count);
			for (i=0; i<nt.item_count; i++) {
				
				DEBUG_PRINTF("      Prüfe Alternative %d/%d: Item Nr. %d => ",
					i+1, nt.item_count, nt.items[i]);
			
				// Falls ein reduzierbarer Knoten entsteht:
				if (items[nt.items[i]].dot_symbol_type == Epsilon) {
					DEBUG_PRINTF("neuer reduzierbarer Knoten\n");

					/* Der neue Knoten wird erzeugt und mit dem Nichtterminalknoten
					 * verbunden.
					 */
					new_node = create_node(nt.items[i]);
					node_table_add(new_node->predecessors, current_nt_node);
					node_table_add(current_nt_node->successors, new_node);
						
					/* Im nächsten Reduce-Read-Schritt muss der neue Knoten
					 * als reduzierbar berücksichtigt werden:
					 */
					node_table_add(reducible_nodes, new_node);

					expand_ok = 1;
				} // if (Epsilon)

				// Falls ein lesbarer Endknoten entsteht:
				else if ( items[nt.items[i]].dot_symbol_type == Terminal ) {

					// Entspricht das dot_symbol dem nächsten Symbol der Eingabe?
					if ( items[nt.items[i]].dot_symbol
		           == lookahead[phase_number % LOOKAHEAD_LENGTH] ) {
						DEBUG_PRINTF("neuer lesbarer Knoten (%ld)\n",
								items[nt.items[i]].dot_symbol);
						
						/* Der neue Knoten wird in den Graphen eingefügt und mit dem
						 * Nichtterminalknoten verbunden:
						 */
						new_node = create_node(nt.items[i]);
						node_table_add(new_node->predecessors, current_nt_node);
						node_table_add(current_nt_node->successors, new_node);

						/* Im nächsten Reduce-Read-Schritt muss der neue Knoten
						 * als lesbar berücksichtigt werden:
						 */
						node_table_add(readable_nodes, new_node);

						expand_ok = 1;

					} // if (passendes dot_symbol)

					/* Falls das Terminalsymbol nicht dem nächsten Eingabesymbol
					 * entspricht, kann eine Expansion dieser Alternative nicht zu
					 * einer akzeptierenden Berechnung führen.
					 */
					else {
						DEBUG_PRINTF("UNPASSENDES dot_symbol (%ld)\n",
								items[nt.items[i]].dot_symbol);
					} // if (passendes dot_symbol) {...} else

				} // elseif (Terminal)

				// Falls ein expandierbarer Endknoten ensteht:
				else if (items[nt.items[i]].dot_symbol_type == Nonterminal) {
					DEBUG_PRINTF("neuer expandierbarer Knoten (%ld)\n",
							items[nt.items[i]].dot_symbol);
							
					/* Falls ein entsprechender Knoten schon in dieser Phase
					 * expandiert wurde, muss er nur noch mit seinem neuen
					 * zusätzlichen Vorgängerknoten B_i verbunden werden:
					 */
					new_node = node_table_find(expanded_nodes, nt.items[i]);
					if (new_node != NULL) {
						DEBUG_PRINTF("      ... existiert schon (schon expandiert)\n");
						node_table_add(new_node->predecessors, current_nt_node);
						node_table_add(current_nt_node->successors, new_node);
					}

					/* Ansonsten wurde der Knoten entweder zwar schon erschaffen,
					 * aber noch nicht expandiert, oder aber der Knoten existiert
					 * noch nicht.
					 */
					else {
						/* Falls der Knoten in dieser Phase erschaffen, aber noch
						 * nicht expandiert wurde, so wurde er aber schon zur
						 * Expansion vorgemerkt. Auch dann muss er nur noch
						 * mit seinem neuen zusätzlichen Vorgänger verbunden werden:
						 */
						new_node = node_table_find(expandable_nodes, nt.items[i]);
						if (new_node != NULL) {
							DEBUG_PRINTF
								("      ... existiert schon (noch nicht expandiert)\n");
							node_table_add(new_node->predecessors, current_nt_node);
							node_table_add(current_nt_node->successors, new_node);
						}

						/* Ansonsten wurde solch ein Knoten in der aktuellen Phase noch
						 * nicht erzeugt. Dies wird nun erledigt, und der Knoten wird
						 * als expandierbar vorgemerkt.
						 */
						else {
							new_node = create_node(nt.items[i]);
							node_table_add(new_node->predecessors, current_nt_node);
							node_table_add(current_nt_node->successors, new_node);

							node_table_add(expandable_nodes, new_node);
						}
					} // if (new_node != NULL) {...} else

					expand_ok = 1;

				} // else if (Nonterminal)

				/* Wenn keiner der vorigen Fälle eintritt, liegt ein Programmier-
				 * fehler vor, da es für einen gültigen Knoten nur die drei
				 * Möglichkeiten Epsilon, Terminal und Nonterminal als
				 * dot_symbol_type gibt!
				 */
				else {
					DEBUG_PRINTF
						("FEHLER beim Expansionsschritt: Ungültiger Knoten!\n");
					exit(EXIT_FAILURE);
				} // else

			} // for (i=0; i<nt.item_count; i++)

		} // if (current_nt_node != NULL) ... else

		/* Falls für mindestens eine Alternative des aktuellen
		 * Nichtterminalsymbols ein Knoten in den Graphen eingefügt wurde,
		 * war die Expansion erfolgreich. Der Nichtterminalknoten wird durch
		 * eine Kante mit dem soeben fertig expandierten Knoten verbunden.
		 */
		if (expand_ok) {
			DEBUG_PRINTF("  Expansion war erfolgreich!\n");
			node_table_add(expanded_nonterminals, current_nt_node);
			node_table_add(node->successors, current_nt_node);
			node_table_add(current_nt_node->predecessors, node);
		}

		/* Falls für den aktuellen expandierbaren Endknoten keine Alternative
		 * gefunden wurde, die zu einer akzeptierenden Berechnung führen kann,
		 * müssen der neue Nichtterminalknoten current_nt_node und der
		 * expandierbare Endknoten node aus dem Graphen entfernt werden, ebenso
		 * wie alle Vorgänger, die keine anderen Nachfolger haben.
		 */
		else {
			/* Der Nichtterminalknoten muss wieder aus dem Graph entfernt werden.
			 * Dazu kann sein Speicherbereich einfach freigegeben werden, da er
			 * noch mit keinem anderen Knoten im Graphen verbunden wurde.
			 */
			DEBUG_PRINTF("    Entferne irrelevanten Nichtterminalknoten ... ");
			node_table_free(current_nt_node->predecessors);
			node_table_free(current_nt_node->successors);
			free(current_nt_node);
			current_nt_node = NULL;
			DEBUG_PRINTF("erledigt!\n");

			/* Der Knoten node war zu Beginn des Expansionsversuchs in die Liste
			 * der expandierten Endknoten eingetragen worden. Dies muss nun
			 * zurückgenommen werden.
			 */
			node_table_remove(expanded_nodes, node);
	
			/* Nun können node und evtl. einige seiner Vorgängerknoten aus dem
			 * Graphen entfernt werden:
			 */
			DEBUG_PRINTF("    Entferne Expansionskandidaten %ld ... ",
					node->content);
			graph_adjust(&node);
			DEBUG_PRINTF("erledigt!\n");

		} // if (expand_ok) else

	} // while (node_table_empty(expandable_nodes) == 0)

} // void expand


/** \brief Durchführung der rückwärts-topologischen Suche nach
 *		reduzier- oder lesbaren Endknoten.
 *
 *  Diese Funktion implementiert die Variante der "zyklenignorierenden
 *  Suche". Alle reduzierbaren und lesbaren Endknoten werden daraufhin
 *  untersucht, ob für sie ein passender Pfad gefunden wird.
 *  Sobald der erste Endknoten mit einem solchen Pfad gefunden wurde, ist
 *  die Suche beendet.
 *
 *	\param searchflag Wert vom Typ \c searchflag_t, durch den festgelegt
 *		wird, ob nur reduzierbare Endknoten, nur lesbare Endknoten oder beide
 *		Typen berücksichtigt werden sollen.
 *	\return Ein Zeiger auf einen reduzier- oder lesbaren Endknoten, für den
 *		es einen passenden Pfad gibt, falls solch ein Knoten existiert. Sonst:
 *		\c NULL.
 */
static
struct node_t*
rtsearch (enum searchflag_t searchflag)
{
	search_id++;

	// Universelle Zählvariablen:
	int i, j, k = 0;

	// Aktuell untersuchter Knoten:
	struct node_t *current_node = NULL;

	// Item, dessen rechte Seite aktuell betrachtet wird:
	struct item_t current_item;

	// Menge der noch zu untersuchenden Knoten:
	struct node_table_t *search_nodes = node_table_init();

	/* Temporärer Zeiger zum Iterieren über die Vorgängerknoten des
	 * aktuell untersuchten Knotens:
	 */
	struct node_t *pred_node = NULL;

	/* Flag, ob ein Knoten etwas zur Herleitung des Lookaheads beitragen
	 * konnte:
	 */
	int ok = 0;

	if (node_table_empty(readable_nodes) != 0)
		DEBUG_PRINTF("  rtsearch: Keine lesbaren Knoten!\n");
	if (node_table_empty(reducible_nodes) != 0)
		DEBUG_PRINTF("  rtsearch: Keine reduzierbaren Knoten!\n");
	
	// Betrachte zuerst alle lesbaren Endknoten:
	if (searchflag == Read || searchflag == Both) {

		node_table_reset_cursor(readable_nodes);
		current_node = node_table_iter(readable_nodes);
		while (current_node != NULL) {
	
			current_item = items[current_node->content];
			ok = 0;
			DEBUG_PRINTF("  Betrachte lesbaren Endknoten %ld\n",
				current_node->content);
	
			/* Falls der aktuelle Knoten zuvor innerer Knoten im Graph war, kann
			 * er noch einen alten und somit jetzt ungültigen Präfixvektor
			 * enthalten. Dieser muss erst gelöscht und neu initialisiert werden:
			 */
			memset(current_node->endnodes, 0,
					LOOKAHEAD_LENGTH * sizeof(struct node_t*));
			current_node->prefixes = 0;
			current_node->last_visited = search_id;
	
			// Betrachte alle Strings u' in FIRST_k(right(current_item))
			for (i=0; i < current_item.first_k_count; i++) {
	
				/* Gib Vergleich zwischen aktuellem Lookahead und aktuell
				 * betrachtetem String aus der FIRST_k-Menge des Knotens aus:
				 */
				DEBUG_PRINTF("    Fk: ");
				for (j=0; j < LOOKAHEAD_LENGTH; j++) {
					DEBUG_PRINTF("%3ld, ", current_item.first_k_item[i][j]);
				}
				DEBUG_PRINTF("\n");
				DEBUG_PRINTF("    LA: ");
				for (j=0; j < LOOKAHEAD_LENGTH; j++) {
					DEBUG_PRINTF("%3ld, ",
							lookahead[ (phase_number + j) % LOOKAHEAD_LENGTH ]);
				}
				DEBUG_PRINTF("\n");
	
				// Vergleiche den aktuellen String u' mit dem Lookahead-String u:
				for (j=0; j < LOOKAHEAD_LENGTH; j++) {
					if ( current_item.first_k_item[i][j] !=
							lookahead[ ( (phase_number + j) % LOOKAHEAD_LENGTH) ] )
						break;
				}
	
				/* Falls die innere for-Schleife bis j == LOOKAHEAD_LENGTH
				 * durchläuft, wurde der Lookaheadstring durch die rechte Seite
				 * des aktuell betrachteten lesbaren Endknotens vollständig
				 * hergeleitet. Somit ist die Suche abgeschlossen!
				 */
				if (j == LOOKAHEAD_LENGTH) {
					DEBUG_PRINTF("    Suche erfolgreich beendet!\n");
					return current_node;
				} // if (j == LOOKAHEAD_LENGTH)
	
				/* Ansonsten wurde entweder eine Stelle gefunden, in der u und u'
				 * voneinander abweichen, oder u' endete vorzeitig, d.h.
				 * current_item.first_k_item[i][j] == 0.
				 */
				else {
					/* Falls |u'| < |u| gilt und die beiden Strings in den ersten
					 * j = |u'| Stellen übereinstimmten, wird ein Eintrag für u' in
					 * den Präfixvektor des aktuellen Knotens aufgenommen.
					 * Dies deckt auch den Fall u' = Epsilon ab, in dem die Schleife
					 * schon bei j == 0 terminiert.
					 */
					if (current_item.first_k_item[i][j] == 0) {
						DEBUG_PRINTF(
								"    FIRST_k-Eintrag verbraucht => Muss Vorgänger besuchen!\n");
						current_node->endnodes[j] = current_node;
						current_node->prefixes = current_node->prefixes | (1 << j);
						ok = 1;
					}
		
					/* Ansonsten gibt es an der j+1-sten Position einen Unterschied
					 * zwischen u' und u, so dass dieses Präfix bei der aktuellen
					 * Suche verworfen werden kann!
					 * In diesem Fall muss nichts unternommen werden. Sofern vorhanden,
					 * wird der nächste String aus der FIRST_k-Menge von current_node
					 * betrachtet.
					 */
				} // if (j == LOOKAHEAD_LENGTH) ... else
	
			} // for (i=0; i < current_item.first_k_count; i++)
	
			/* Falls der Endknoten nicht zu einer akzeptierenden Berechnung führen
			 * kann: Entfernen und Graphbereinigung.
			 */
			if (ok == 0) {
				DEBUG_PRINTF("  Graphjustierung für Knoten %ld wird veranlasst ...",
						current_node->content);
				graph_adjust(&current_node);
				DEBUG_PRINTF("  Graphjustierung erledigt\n");
			}
	
			/* Falls der Endknoten allein nicht zur Herleitung des Lookahead genügt,
			 * aber er zu einer akzeptierenden Berechnung führen kann:	
			 * Füge jeden Vorgängerknoten von current_node zur Suchmenge hinzu,
			 * sofern er sich dort noch nicht befindet:
			 */
			else {
				node_table_reset_cursor(current_node->predecessors);
				pred_node = node_table_iter(current_node->predecessors);
				while (pred_node != NULL) {
					node_table_add(search_nodes, pred_node);
					pred_node = node_table_iter(current_node->predecessors);
				}
			}
	
			// Nächster lesbarer Knoten für folgenden Schleifendurchlauf:
			current_node = node_table_iter(readable_nodes);
	
		} // while (current_node != NULL) // readable_nodes
	
	} // if (searchflag == Read || searchflag == Both

	// Nun werden alle reduzierbaren Endknoten behandelt:
	if (searchflag == Reduce || searchflag == Both) {
		
		node_table_reset_cursor(reducible_nodes);
		current_node = node_table_iter(reducible_nodes);
		while (current_node != NULL) {
	
			DEBUG_PRINTF("  Betrachte reduzierbaren Endknoten %ld\n",
					current_node->content);
	
			/* Falls der aktuelle Knoten zuvor innerer Knoten im Graph war, kann
			 * er noch einen alten und somit jetzt ungültigen Präfixvektor
			 * enthalten. Dieser muss erst gelöscht und neu initialisiert werden:
			 */
			memset(current_node->endnodes, 0,
					LOOKAHEAD_LENGTH * sizeof(struct node_t*));
			current_node->prefixes = 0;
			current_node->last_visited = search_id;
	
			/* Assoziiere das Lookahead-Präfix Epsilon mit dem reduzierbaren
			 * Endknoten:
			 */
			current_node->endnodes[0] = current_node;
			current_node->prefixes = 1;
	
			/* Füge jeden Vorgängerknoten von current_node zur Suchmenge hinzu,
			 * sofern er sich dort noch nicht befindet:
			 */
			node_table_reset_cursor(current_node->predecessors);
			pred_node = node_table_iter(current_node->predecessors);
			while (pred_node != NULL) {
				node_table_add(search_nodes, pred_node);
				pred_node = node_table_iter(current_node->predecessors);
			}
	
			// Nächster reduzierbarer Knoten für folgenden Schleifendurchlauf:
			current_node = node_table_iter(reducible_nodes);
	
		} // while (current_node != NULL) // reducible_nodes
	
	} // if (searchflag == Reduce || searchflag == Both)

	/* Falls der Lookahead-String nicht durch die rechte Seite eines 
	 * Endknotens allein hergeleitet werden konnte: Behandle die
	 * Vorgängerknoten.
	 */
	while (node_table_empty(search_nodes) == 0) {
		current_node = node_table_pop(search_nodes);
		current_item = items[current_node->content];
		ok = 0;

		/* Itemknoten:
		 * Falls der aktuell untersuchte Knoten ein Item repräsentiert, so muss
		 * getestet werden, ob die rechte Seite dieses Items etwas zur
		 * Herleitung des Lookaheads beitragen kann.
		 * Das Item kann nur die Form [A -> alpha . B beta] haben, mit B ein
		 * Nichtterminalsymbol.
		 */
		if (current_node->content >= 0) {
			DEBUG_PRINTF("  Betrachte Itemknoten %ld\n",
				current_node->content);

			/* Falls des Knoten in diesem Suchlauf nicht schon einmal besucht
			 * wurde, hat er noch keinen oder einen ungültigen Präfixvektor.
			 * Sollte er in diesem Suchlauf schon besucht worden sein, wird sein
			 * vorhandener Präfixvektor nur ergänzt.
			 */
			if ( current_node->last_visited < search_id )  {
				memset(current_node->endnodes, 0,
						LOOKAHEAD_LENGTH * sizeof(struct node_t*));
				current_node->last_visited = search_id;
			}

			/* Der Itemknoten v = [A -> alpha . B beta] hat genau einen
			 * Nachfolgerknoten. Dieser repräsentiert das Nichtterminalsymbol B.
			 * Für jeden Eintrag u' im Präfixvektor von B wird geprüft, 
			 * ob u' mithilfe der rechten Seite \beta von v zu einem Präfix
			 * mit Länge >= |u'| von u verlängert werden kann.
			 * Der Präfixvektor von B existiert auf jeden Fall und ist auch
			 * garantiert aktuell, da v erst nach Besuch von B in die Suchmenge
			 * aufgenommen wurde!
			 */
			node_table_reset_cursor(current_node->successors);
			struct node_t *nt_node = node_table_iter(current_node->successors);

			/* Wenn current_node als innerer Knoten keinen Nachfolger nt_node
			 * hat, liegt ein Programmierfehler vor:
			 */
			assert(nt_node != NULL);

			current_node->prefixes = nt_node->prefixes;

			/* Prüfe jeden der LOOKAHEAD_LENGTH Einträge im Präfixvektor von
			 * nt_node auf Verlängerbarkeit:
			 */
			for (i=0; i<LOOKAHEAD_LENGTH; i++) {

				/* Ist der Eintrag mit Index i im Präfixvektor von B verlängerbar?
				 */
				if (nt_node->endnodes[i] != NULL) {
					DEBUG_PRINTF("    Prüfe Präfixeintrag Nr. %d\n", i);

					/* Teste jeden String u' in FIRST_k(beta) auf seinen Beitrag: */
					for (j=0; j < current_item.first_k_count; j++) {
							
						DEBUG_PRINTF("    Vergleiche mit aktuellem FIRST_k-String: ");
						for (k=0; (k+i) < LOOKAHEAD_LENGTH; k++) {
							DEBUG_PRINTF("%ld, ", current_item.first_k_item[j][k]);
						}
						DEBUG_PRINTF("\n");

						/* Vergleiche den aktuellen String mit dem Suffix des
						 * Lookahead:
						 */
						for (k=0; (k+i) < LOOKAHEAD_LENGTH; k++) {
							if ( current_item.first_k_item[j][k] !=
									lookahead[( (phase_number + i + k) % LOOKAHEAD_LENGTH)] )
								break;
						}

						/* Falls die for-Schleife über k bis (k+i) == LOOKAHEAD_LENGTH
						 * durchläuft, wurde der Lookaheadstring mithilfe der rechten
						 * Seite des aktuellen Knotens vollständig hergeleitet.
						 * Somit ist die Suche abgeschlossen!
						 */
						if ( (k+i) == LOOKAHEAD_LENGTH ) {
							DEBUG_PRINTF("    Suche erfolgreich beendet!\n");
							return nt_node->endnodes[i];
						}

						/* Ansonsten wurde entweder eine Stelle gefunden, in der u und
						 * u' voneinander abweichen, oder u' endete vorzeitig, d.h.
						 * current_item.first_k_item[j][k] == 0.
						 */
						else {
							/* Falls i + k < |u| und die beiden Strings stimmten in den
							 * ersten i+k Zeichen überein und u' endete vorzeitig, so
							 * kann der bisher hergeleitete Teilstring des Lookaheads
							 * noch forgesetzt werden.
							 * Es wird ein Eintrag in den Präfixvektor des aktuellen
							 * Knotens aufgenommen.
							 * Dies deckt auch den Fall an, dass durch den aktuellen
							 * String aus FIRST_k(beta) nur Epsilon hergeleitet wurde,
							 * die Schleife also schon bei k=0 terminierte.
							 */
							if (current_item.first_k_item[j][k] == 0) {
								DEBUG_PRINTF(
										"    FIRST_k-Eintrag verbraucht => Muss Vorgänger besuchen!\n");
								current_node->endnodes[k+i] = nt_node->endnodes[i];
								ok = 1;
							}

							/* Ansonsten gibt es an der i+k+1-sten Stelle von u einen
							 * Unterschied zwischen u und dem bisher hergeleiteten
							 * Präfix, sodass dieses bei der aktuellen Suche verworfen
							 * werden kann!
							 * In diesem Fall muss nichts unternommen werden. Sofern
							 * vorhanden, wird der nächste String auf FIRST_k(beta)
							 * betrachtet.
							 */
							
						} // if ( (k+1) == LOOKAHEAD_LENGTH ) ... else ...

					} // for (j=0; ...)

				} // if (nt_node->endnodes[i] != NULL) ...

			} // for (i=0; ...
			
			/* Falls der aktuelle Knoten nach aktueller Kenntnis nicht zu einer
			 * akzeptierenden Berechnung führen kann, darf ich ihn dennoch zur
			 * Zeit nicht löschen, da er evtl. später in derselben Suche
			 * aufgrund eines Kreises nochmal besucht werden könnte!
			 */

			/* Falls der aktuelle Knoten zu einer akzeptierenden Berechnung
			 * führen kann: Füge jeden Vorgängerknoten von current_node zur
			 * Suchmenge hinzu, sofern er sich dort noch nicht befindet:
			 */
			if (ok != 0) {
				node_table_reset_cursor(current_node->predecessors);
				pred_node = node_table_iter(current_node->predecessors);
				while (pred_node != NULL) {
					node_table_add(search_nodes, pred_node);
					pred_node = node_table_iter(current_node->predecessors);
				}

				DEBUG_PRINTF("    Suchknoten: ");
				NODE_TABLE_PRINT(search_nodes);

			} // end if (ok != 0)

		} // if (current_node->content >= 0) // Itemknoten

		/* Nichtterminalknoten:
		 * Falls der aktuell untersuchte Knoten ein Nichtterminalsymbol
		 * repräsentiert, so muss sein Präfixvektor um die Einträge aus
		 * den Präfixvektoren all seiner Nachfolger ergänzt werden.
		 */
		else {
			DEBUG_PRINTF("  Betrachte Nichtterminalknoten %ld\n",
				~(current_node->content));

			/* Falls des Knoten in diesem Suchlauf nicht schon einmal besucht
			 * wurde, hat er noch keinen oder einen ungültigen Präfixvektor und
			 * auch evtl. einen ungültigen Eintrag im Feld prefixes.
			 * Sollte er in diesem Suchlauf schon besucht worden sein, wird sein
			 * vorhandener Präfixvektor nicht gelöscht, sondern im
			 * Folgenden nur ergänzt.
			 */
			if ( current_node->last_visited < search_id )  {
				memset(current_node->endnodes, 0,
						LOOKAHEAD_LENGTH * sizeof(struct node_t*));
				current_node->prefixes = 0;
				current_node->last_visited = search_id;
			}

			node_table_reset_cursor(current_node->successors);
			struct node_t *succ_node =
				node_table_iter(current_node->successors);

			/* Für jeden Nachfolger v' von des Nichtterminalknotens
			 * current_node, der während des aktuellen Suchlaufs schon besucht
			 * wurde: Ergänze den Präfixvektor von current_node um die
			 * Einträge im Präfixektor von v'.
			 */
			while (succ_node != NULL) {
				if ( succ_node->last_visited == search_id) {
					for (i=0; i<LOOKAHEAD_LENGTH; i++) {
						if (succ_node->endnodes[i] != NULL) {
							current_node->endnodes[i] = succ_node->endnodes[i];
							current_node->prefixes = current_node->prefixes | (1 << i);
							ok = 1;
						}
					}
				} // if ( succ_node->last_visited == search_id )

				succ_node = node_table_iter(current_node->successors);
			} // end while (succ_node != NULL)

			/** \todo Bei der Betrachtung eines Nichtterminalknotens bietet sich
			 *  eine Möglichkeit zur Fehlererkennung an:
			 *  Wenn für den selben Wert von \c i von zwei verschiedenen
			 *  Nachfolgern aus assoziierte Endknoten gemeldet werden, ist die
			 *  Grammatik nicht LR(k)!
			 *  Bequem zu testen bspw. über einen Bitvergleich in \c prefixes_num!
			 */

			/* Falls es mindestens einen Präfixeintrag zu übertragen gab, kann
			 * der Knoten zu einer akzeptierenden Berechung beitragen. Jeder
			 * Vorgängerknoten kann zur Suchmenge hinzugefügt werden, sofern er
			 * sich dort noch nicht befindet UND sofern der Präfixvektor von
			 * current_node sich seit dem letzten Besuch des Knotens verändert
			 * hat.
			 */
			if (ok != 0) {
				node_table_reset_cursor(current_node->predecessors);
				pred_node = node_table_iter(current_node->predecessors);
				while (pred_node != NULL) {
					DEBUG_PRINTF("    weitersuchen in %ld?", pred_node->content);
					if ( (pred_node->prefixes != current_node->prefixes)
							|| (pred_node->last_visited != search_id) ) {
						pred_node->prefixes = current_node->prefixes;
						node_table_add(search_nodes, pred_node);
						DEBUG_PRINTF(" Ja\n");
					}
					else {
						DEBUG_PRINTF(" Nein\n");
					}
					pred_node = node_table_iter(current_node->predecessors);
				} // while (pred_node != NULL)
				DEBUG_PRINTF("    Suchknoten: ");
				NODE_TABLE_PRINT(search_nodes);
			} // if (ok != 0)

		} // if (current_node->content >= 0) ... else // Nichtterminalknoten

	} // while (node_tabe_empty(search_nodes) == NULL)


	/* Wenn die Funktionsausführung an diese Stelle gelangt, konnte der
	 * Lookahead-String nicht hergeleitet werden:
	 */
	return NULL;
} // struct node_t *rtsearch


// Forward-Deklaration
static
void
test_new_endnode(struct node_t **node);


/** \brief Hilfsfunktion, um die Vorgängerknoten eines reduzierten
 *		Endknotens zu behandeln.
 *
 *	Der Aufruf <tt>treat_predecessors(nt_node)</tt> erfolgt nach der
 *	Reduktion eines reduzierbaren Endknotens w. \a nt_node ist dabei der
 *	Nichtterminalknoten, der im Graph unmittelbarer Vorgänger von w war.
 *	Für jeden Vorgänger von \a nt_node wird nun überprüft, ob er sich
 *	als neuer Endknoten eignet.
 *
 *	\param nt_node Zeiger auf einen Zeiger (Call by Reference) auf den
 *		Nichtterminalknoten, aus dessen Vorgängern die neuen Endknoten des
 *		Graphen entstehen sollen.
 */
static
void
treat_predecessors(struct node_t **nt_node)
{
	// Zeiger auf den aktuell bearbeiteten Vorgängerknoten:
	struct node_t *pred_node = NULL;
	
	/* Es sind Fälle möglich, in denen während der Vorgängerbehandlung
	 * erneut Reduktionen stattfinden, die dann auch eine Wiederholung der
	 * Vorgängerbehandlung implizieren. Daher wird die Bearbeitung von
	 * Vorgängerknoten so oft iteriert, bis *nt_node nicht mehr auf einen
	 * Nichtterminalknoten zeigt.
	 */
	while ( (*nt_node) != NULL ) {

		/* Falls nt_node keine anderen Nachfolger im Graphen als w hat, sind
		 * die Vorgängerknoten von nt_node die neuen Endknoten des Graphen.
		 * Ihr Punkt wird um eine Stelle nach rechts veschoben, und sie werden
		 * darauf überprüft, ob sie Teil einer akzeptierenden Berechnung
		 * sein können:
		 */
		if ( node_table_empty((*nt_node)->successors) != 0 ) {
			
			DEBUG_PRINTF("  nt_node %ld hat keine weiteren Nachfolger!\n",
					(*nt_node)->content);
			while ( node_table_empty((*nt_node)->predecessors) == 0 ) {
			
				/* Betrachte den nächsten Vorgängerknoten v und entferne seine
				 * Kante zu \c nt_node:
				 */
				pred_node = node_table_pop((*nt_node)->predecessors);
				node_table_remove(pred_node->successors, (*nt_node));
		
				/* Es ist möglich, dass v zu einem früheren Zeitpunkt der aktuellen
				 * Phase schon einmal ein expandierbarer Endknoten war, von dem
				 * aus eine Expansion durchgeführt wurde. Dabei wurde v in die
				 * Menge expanded_nodes eingetragen. Da v nach dem Verschieben des
				 * Punktes aber nicht mehr derselbe Knoten wie damals ist, muss
				 * er jetzt aus expanded_nodes ausgetragen werden:
				 */
				node_table_remove(expanded_nodes, pred_node);

				/* Verschiebe den Punkt in v und überprüfe den Knoten: */
				pred_node->content = pred_node->content + 1;
				test_new_endnode(&pred_node);
		
			} // end while ( node_table_empty((*nt_node)->predecessors) == 0)
		
			/* Nachdem alle Vorgängerknoten behandelt wurden, kann auch nt_node
			 * entfernt werden. Falls anhand dieses Symbols in der aktuellen Phase
			 * bereits expandiert wurde, muss der Knoten zuvor auch aus der Liste
			 * expanded_nonterminals ausgetragen werden.
			 */

			DEBUG_PRINTF("  Entferne erledigten Nichtterminalknoten %ld ...",
					(*nt_node)->content);
			node_table_remove(expanded_nonterminals, (*nt_node));
			node_table_free((*nt_node)->predecessors);
			node_table_free((*nt_node)->successors);
			free(*nt_node);
			*nt_node = NULL;
			DEBUG_PRINTF(" erledigt!\n");
		
		} // if ( node_table_empty((*nt_node)->successors) != 0 )

		/* Sollte nt_node noch andere Nachfolger als w haben, müssen Kopien der
		 * Vorgänger von A angefertigt werden, aus denen dann die neuen
		 * Endknoten entstehen:
		 */
		else {

			DEBUG_PRINTF(
					"  NT-Knoten %ld hat Nachfolger => Kopiere Vorgängerknoten!\n",
					(*nt_node)->content);
			/* Erzeuge für jeden Knoten v in der Vorgängermenge von nt_node eine
			 * Kopie v' inklusive aller in v eingehenden Kanten, aber ohne die
			 * ausgehenden Kanten.
			 * Bewege dann in jeder Knotenkopie v' den Punkt um eine Stelle nach
			 * rechts und überprüfe v' mit test_new_endnode.
			 */
			node_table_reset_cursor((*nt_node)->predecessors);
			struct node_t *copy_node = node_table_iter((*nt_node)->predecessors);
			while ( copy_node != NULL ) {
				/* Erzeuge eine Kopie v' des Knotens copy_node mit um eine Stelle
				 * verschobenem Punkt. Der (einzige) direkte Vorgänger von v
				 * wird zugleich zum Vorgänger von v'.
				 * Sollte v und somit auch v' keinen Vorgänger haben, muss v'
				 * in die Liste der Startknoten des Graphen eingetragen werden.
				 */
				DEBUG_PRINTF("  Kopiere Vorgänger %ld\n", copy_node->content);
				pred_node = create_node(copy_node->content + 1);
				if ( node_table_empty(copy_node->predecessors) != 0 ) {
					list_enqueue(&graph, pred_node);
				}
				else {
					DEBUG_PRINTF("  Verbinde neuen Knoten mit Vorgängern ...");
					pred_node->predecessors =
						node_table_copy(copy_node->predecessors);
					/* Der neue Itemknoten pred_node hat jetzt genau einen
					 * Vorgängerknoten, einen Nichtterminalknoten pred_nt.
					 * In dessen Nachfolgerliste muss pred_node noch eintragen
					 * werden:
					 */
					node_table_reset_cursor(copy_node->predecessors);
					struct node_t *pred_nt =
						node_table_iter(copy_node->predecessors);
					node_table_add(pred_nt->successors, pred_node);
					DEBUG_PRINTF(" erledigt!\n");
				}

				// Ueberprüfe den neuen Knoten v':
				DEBUG_PRINTF("  Teste neuen Knoten!\n");
				test_new_endnode(&pred_node);

				copy_node = node_table_iter((*nt_node)->predecessors);
			} // while ( copy_node != NULL)
			
			/* Der Zeiger nt_node muss zurückgesetzt werden, da ansonsten auch
       * dann die äussere Schleife wiederholt wird, wenn im Folgenden
			 * keine erneute Reduktion stattfindet.
			 */
			*nt_node = NULL;
		} // if ( node_table_empty((*nt_node)->successors) != 0 ) ... else

		/* Nun sind alle Kandidaten für neue Endknoten kategorisiert in
		 * lesbare, reduzierbare, expandierbare und irrelevante Knoten.
		 * Neue lesbare Endknoten wurden von der Funktion test_new_endnode
		 * in die Liste new_readable_endnodes eingetragen. Diese Liste
		 * kann jetzt in die Tabelle readable_endnodes überführt werden.
		 */
		struct node_t *read_node = NULL;
		while (new_readable_nodes != NULL) {
			read_node = list_pop(&new_readable_nodes);
			node_table_add(readable_nodes, read_node);			
		}
		read_node = NULL;

		/* Nun muss überprüft werden, ob dem Graph neue reduzierbare Knoten
		 * hinzugefügt wurden. Sollte dies der Fall sein, müssen sie auf
		 * Gültigkeit getestet und ggf. reduziert werden, um vor dem
		 * nächsten Expansionsschritt die Einhaltung der Invariante
		 * sicherzustellen.
		 */
		if ( node_table_empty(reducible_nodes) == 0 ) {

			/* Suche in der Menge der reduzierbaren Endknoten nach einem
			 * Knoten, für die ein passender Pfad existiert.
			 */
			struct node_t *suitable_node = rtsearch(Reduce);

			/* Zeiger für Knoten, die evtl. gelöscht werden müssen: */
			struct node_t *kill_node = NULL;

			/* Sollte kein Knoten gefunden werden, für den ein passender Pfad
			 * existiert, können die neu erzeugten reduzierbaren Knoten nicht
			 * zu einer akzeptierenden Berechung führen. Sie können daher aus
			 * dem Graph entfernt werden.
			 */
			if ( suitable_node == NULL ) {
				DEBUG_PRINTF(
						"  Kein gültiger reduzierbarer Endknoten gefunden!\n");

				// Alle reduzierbaren Endknoten werden entfernt: 
				while (node_table_empty(reducible_nodes) == 0) {
					kill_node = node_table_pop(reducible_nodes);
					graph_adjust(&kill_node);
				}
			} // if ( suitable_node == NULL )

			else {

				/* Sollte es einen reduzierbaren Endknoten w mit gültigem Pfad
				 * geben, so muss die Reduktion dieses Knotens durchgeführt und die
				 * Vorgängerbehandlung wiederholt werden, bevor zum nächsten
				 * Expansionsschritt übergegangen werden kann.
				 */
				if ( items[suitable_node->content].dot_symbol_type == Epsilon ) {
					DEBUG_PRINTF(
							"  Gültiger reduzierbarer Endknoten %ld wurde gefunden!\n",
							suitable_node->content);

					/* Vermerke die Reduktion im Wald der partiellen Ableitungsbäume:
					 */
					PARSE_FOREST_REDUCE(&parse_forest, suitable_node->content);

					// Alle lesbaren Endknoten müssen entfernt werden:
					while (node_table_empty(readable_nodes) == 0) {
						kill_node = node_table_pop(readable_nodes);
						graph_adjust(&kill_node);
					}

					// Alle reduzierbaren Endknoten ausser w werden entfernt: 
					while (node_table_empty(reducible_nodes) == 0) {
						kill_node = node_table_pop(reducible_nodes);
						if (kill_node == suitable_node)
							kill_node = NULL;
						else
							graph_adjust(&kill_node);
					}

					/* Im Unterschied zur Situation bei einer Reduktion innerhalb der
					 * Funktion reduce_read kann der Graph jetzt auch expandierbare
					 * Endknoten enthalten, die noch zu löschen sind:
					 */
					while (node_table_empty(expandable_nodes) == 0) {
						kill_node = node_table_pop(expandable_nodes);
						graph_adjust(&kill_node);
					}

					/* Betrachte den Nichtterminalknoten A, der in Graph einziger
					 * direkter Vorgänger von w ist, und entferne die Kante von
					 * A nach w:
					 */
					*nt_node = 
						node_table_pop(suitable_node->predecessors);
					node_table_remove((*nt_node)->successors, suitable_node);

					/* Falls ein Itemknoten mehr als einen Vorgänger hatte,
					 * liegt ein Programmierfehler vor:
					 */
					assert(node_table_empty(suitable_node->predecessors) != 0);

					// Entferne w aus dem Graphen:
					node_table_free(suitable_node->predecessors);
					node_table_free(suitable_node->successors);
					free(suitable_node);
					suitable_node = NULL;

					/* Nun müssen noch in einem erneuten Durchlauf der äusseren
					 * while-Schleife die bisherigen Vorgängerknoten von A
					 * behandelt werden, aus denen wieder neue Endknoten entstehen.
					 */

				} // if (Epsilon)

				/* Ansonsten ist das Suchergebnis ungültig, denn wenn die
				 * Suche auf lesbare Endknoten eingeschränkt wurde, darf das
				 * Ergebnis nur ein Knoten mit dot_symbol_type == Epsilon sein!
				 */
				else {
					fprintf(stderr, "treat_predecessors: Falscher Knoten wurde gefunden!\n");
					abort();
				} // if ( Epsilon ) ... else

			} // if ( suitable_node == NULL ) ... else

		} // if ( node_tabe_empty(reducible_nodes) == 0 )

	} // while ( (*nt_node) != NULL)

} // void treat_predecessors


/** \brief Hilfsfunktion, um einen neu entstandenen Endknoten nach einem
 *  Reduktions-/Leseschritt zu überprüfen.
 *
 *  Der Aufruf <tt>test_new_endnode(node)</tt> prüft, ob der neue Endknoten
 *  node expandierbar, reduzierbar oder lesbar ist und trägt ihn in die 
 *	entsprechende Menge ein.
 *
 *  \param node Zeiger auf einen Zeiger (Call by Reference) auf den zu
 *  	untersuchenden Knoten.
 */
static
void
test_new_endnode(struct node_t **node)
{

	/* Falls der neue Endknoten das Item [S' -> S.] repräsentiert, darf er
	 * nur im Graph verbleiben, falls die Eingabe verbraucht ist. In diesem
	 * Falle darf er allerdings nicht in die Menge der reduzierbaren Knoten
	 * eingetragen werden!
	 */
	if ( (*node)->content == acceptitem_number ) {
		if (lookahead[ (phase_number) % LOOKAHEAD_LENGTH ] == 0) {
			DEBUG_PRINTF("    Accept-Item gefunden!\n");
			return;
		}
		/* Wurde der Knoten für das Accept-Item erzeugt, bevor die gesamte
		 * Eingabe verbraucht ist, kann er wieder aus dem Graphen entfernt
		 * werden. Da er mit keinem anderen Knoten verbunden ist und in keine
		 * andere Knotenmenge als graph eingetragen wurde, kann die Löschung
		 * in diesem Fall ohne Aufruf von graph_adjust erfolgen.
		 */
		else {
			DEBUG_PRINTF("    Verfrühtes Accept-Item gefunden!\n");
			list_remove(&graph, (*node));
			node_table_free((*node)->predecessors);
			node_table_free((*node)->successors);
			free(*node);
			*node = NULL;
			return;
		}
	} // if (acceptitem_number)

	/* Falls der neue Knoten expandierbar ist, muss er für den
	 * nächsten Expansionsschritt vorgemerkt werden:
	 */
	if ( items[(*node)->content].dot_symbol_type == Nonterminal ) {
		DEBUG_PRINTF(
				"    Modifizierter/neuer Endknoten %ld ist expandierbar.\n",
				(*node)->content);
		node_table_add(expandable_nodes, *node);
	} // end if ( ... == Nonterminal )

	/* Ansonsten ist der Knoten lesbar oder reduzierbar: */
	else {

		/* Falls der neue Knoten reduzierbar ist, muss er für eine
		 * Wiederholung des aktuellen Reduktions-/Leseschrittes
		 * vorgemerkt werden:
		 */
		if ( items[(*node)->content].dot_symbol_type == Epsilon ) {
			DEBUG_PRINTF(
					"    Modifizierter/neuer Endknoten %ld ist reduzierbar.\n",
					(*node)->content);
			node_table_add(reducible_nodes, *node);
		} // end if ( ... == Epsilon )

		/* Ansonsten ist der Knoten lesbar: */
		else {
			
			/* Wenn das nächste Lookahead-Symbol hinter dem Punkt steht,
			 * wird der Knoten beim nächsten Leseschritt mit berücksichtigt.
			 */
			if ( (items[(*node)->content].dot_symbol_type == Terminal)
				&& (items[(*node)->content].dot_symbol == lookahead[
						(phase_number) % LOOKAHEAD_LENGTH])
				 ) {
				DEBUG_PRINTF(
						"    Modifizierter/neuer Endknoten %ld ist lesbar.\n",
						(*node)->content);
				list_add(&new_readable_nodes, *node);
			}
			
			/* Falls der Knoten lesbar ist, aber mit falschen Symbol hinter dem
			 * Punkt, kann er nicht zu einer akzeptierenden Berechnung
			 * korrespondieren und wird aus dem Graphen entfernt.
			 */
			else {
				DEBUG_PRINTF(
						"    Modifizierter/neuer Endknoten %ld ist lesbar mit falschem Symbol.\n",
						(*node)->content);
				graph_adjust(node);
			}

		} // end if ( ... == Epsilon ) else

	} // end if ( ... == Nonterminal ) else

} // void test_new_endnode


/** \brief Durchführung eines Reduktions-/Leseschrittes.
 *
 *	Die Funktion \c reduce_read sorgt für die Durchführung von
 *	Reduktions-/Leseschritten und wird aufgerufen, wenn der Graph
 *	des Parsers nur noch reduzierbare und/oder lesbare Endknoten
 *	enthält. Im Falle eines Leseschrittes wird die aktuelle
 *	Phase beendet und ein neues Eingabesymbol in den Lookahead-Puffer
 *	eingefügt.
 *
 *	\return Integerwert, der Erfolg oder Misserfolg der Aktion
 *		durch die Werte 0 (Erfolg) oder 1 (Misserfolg) kennzeichnet.
 */
static
int
reduce_read (void)
{
	// Zeiger für evtl. zu löschenden Endknoten:
	struct node_t *kill_node = NULL;

	/* Suche in der Menge der reduzier- und lesbaren Knoten nach einem Knoten,
	 * für den ein passender Pfad existiert.
	 */
	struct node_t *suitable_node = rtsearch(Both);

	/* Falls das Suchergebnis leer ist, kann die Eingabe nicht Element der
	 * vom Parser erkannten Sprache sein.
	 */
	if (suitable_node == NULL) {
		DEBUG_PRINTF("  Kein Endknoten mit passendem Pfad gefunden!\n");
		return 1;
	}

	/* Falls das Suchergebnis ein reduzierbarer Endknoten w ist,
	 * wird genau dieser Knoten jetzt reduziert:
	 */
	if ( items[suitable_node->content].dot_symbol_type == Epsilon ) {
		
		DEBUG_PRINTF("  Gültiger reduzierbarer Endknoten %ld wurde gefunden!\n",
				suitable_node->content);

		/* Vermerke die Reduktion im Wald der partiellen Ableitungsbäume:
		 */
		PARSE_FOREST_REDUCE(&parse_forest, suitable_node->content);

		// Alle lesbaren Endknoten müssen entfernt werden:
		while (node_table_empty(readable_nodes) == 0) {
			kill_node = node_table_pop(readable_nodes);
			graph_adjust(&kill_node);
		}

		// Alle reduzierbaren Endknoten ausser w werden entfernt: 
		while (node_table_empty(reducible_nodes) == 0) {
			kill_node = node_table_pop(reducible_nodes);
			if (kill_node == suitable_node)
				kill_node = NULL;
			else
				graph_adjust(&kill_node);
		}

		/* Bemerkung: Aufgrund der Invariante, die der Parser unterhält,
		 * enthält der Graph unmittelbar vor einem Reduktions-/Leseschritt
		 * keine expandierbaren Endknoten. Die Menge expandable_nodes sollte
		 * also leer sein und kann daher beim Entfernen aller verbliebenen
		 * Endknoten übergangen werden.
		 */

		/* Betrachte den Nichtterminalknoten A, der im Graph einziger direkter
		 * Vorgänger von w ist. Entferne die Kante zwischen A und w.
		 */
		struct node_t *nt_node =
			node_table_pop(suitable_node->predecessors);
		node_table_remove(nt_node->successors, suitable_node);

		/* Falls ein Itemknoten mehr als einen Vorgängerknoten hatte, liegt
		 * ein Programmierfehler vor:
		 */
		assert(node_table_empty(suitable_node->predecessors) != 0);

		// Entferne w aus dem Graphen:
		node_table_free(suitable_node->predecessors);
		node_table_free(suitable_node->successors);
		free(suitable_node);
		suitable_node = NULL;

		/* Nun müssen noch die bisherigen Vorgängerknoten von A behandelt
		 * werden, da aus ihnen die neuen Endknoten des Graphen entstehen.
		 */
		treat_predecessors(&nt_node);

		/* Falls die Vorgängerbehandlung fehlerfrei durchgelaufen ist, ist der
		 * Reduktionsschritt abgeschlossen.
		 */
		return 0;

	} // if (Epsilon)

	/* Falls das Suchergebnis ein lesbarer Endknoten ist:
	 * Nun findet ein Leseschritt für alle lesbaren Endknoten statt;
	 * die aktuelle Phase wird beendet.
	 */
	else {

		DEBUG_PRINTF("  Gültiger lesbarer Endknoten %ld wurde gefunden!\n",
				suitable_node->content);

		/* Vermerke den Leseschritt im Wald der partiellen Ableitunsgbäume:
		 */
		PARSE_FOREST_READ(&parse_forest,
				lookahead[phase_number % LOOKAHEAD_LENGTH]);

		/* Alle reduzierbaren Endknoten und ggf. einige ihrer Vorgängerknoten
		 * müssen entfernt werden:
		 */
		while (node_table_empty(reducible_nodes) == 0) {
			kill_node = node_table_pop(reducible_nodes);
			graph_adjust(&kill_node);
		}

		/* Bemerkung:
		 * 1.) Aufgrund der Invariante, die der Parser unterhält, kann
		 * die Menge der expandierbaren Endknoten hier ignoriert werden, da sie
		 * leer sein sollte.
		 * 2.) Wenn ein lesbarer Endknoten gefunden wurde, für den es einen
		 * passenden Pfad gibt, muss das nicht auch für jeden anderen lesbaren
		 * Endknoten gelten. Dennoch wird der Leseschritt an dieser Stelle für
		 * alle lesbaren Endknoten durchgeführt. Knoten, für die dies ein
		 * Fehler war, werden später im Parsevorgang aussortiert.
		 */

		/* Beende die aktuelle Phase durch Lesen des nächsten Eingabesymbols
		 * und Bereinigen der temporären Listen:
		 */
		DEBUG_PRINTF("  Token %ld gelesen!\n",
				lookahead[phase_number % LOOKAHEAD_LENGTH]);
		lookahead[phase_number % LOOKAHEAD_LENGTH] = call_lexer();
		DEBUG_PRINTF("  Lege neues Eingabesymbol %ld in Puffer!\n",
			lookahead[phase_number % LOOKAHEAD_LENGTH]);
		phase_number = phase_number + 1;
		DEBUG_PRINTF("  Bereinige expanded_nodes\n");
		node_table_clear(expanded_nodes);
		DEBUG_PRINTF("  Bereinige expanded_nonterminals\n");
		node_table_clear(expanded_nonterminals);

		/* Bewege in allen lesbaren Endknoten den Punkt um eine Stelle nach
		 * rechts, d.h. ersetze jedes Item [A -> alpha . a_{i+1} beta] durch
		 * sein Nachfolgeitem [A -> alpha a_{i+1} . beta].
		 * Untersuche dabei jeden dieser neuen Endknoten auf Expandier-,
		 * Reduzier- oder Lesbarkeit in der nächsten Phase.
		 */
		while (node_table_empty(readable_nodes) == 0) {
			suitable_node = node_table_pop(readable_nodes);
			DEBUG_PRINTF("    Bearbeite lesbaren Endknoten %ld\n",
					suitable_node->content); 
			suitable_node->content = suitable_node->content + 1;
			test_new_endnode(&suitable_node);
		}
		
		/* Nun sind alle Kandidaten für neue Endknoten kategorisiert in
		 * lesbare, reduzierbare, expandierbare und irrelevante Knoten.
		 * Neue lesbare Endknoten wurden von der Funktion test_new_endnode
		 * in die Liste new_readable_endnodes eingetragen. Diese Liste
		 * kann jetzt in die Tabelle readable_endnodes überführt werden.
		 */
		struct node_t *read_node = NULL;
		while (new_readable_nodes != NULL) {
			read_node = list_pop(&new_readable_nodes);
			node_table_add(readable_nodes, read_node);			
		}
		read_node = NULL;

		DEBUG_PRINTF("    Alle lesbaren Endknoten bearbeitet!\n");

		DEBUG_PRINTF(
				"Abschliessende Behandlung neuer reduzierbarer Knoten:\n");
		/* Falls bei der Bearbeitung der lesbaren Endknoten neue reduzierbare
		 * Endknoten entstanden sind, müssen diese vor Ende der Phase
		 * bearbeitet werden, um die Invariante vor dem ersten Expansionsschritt
		 * der neuen Phase aufrecht zu erhalten.
		 */
		if (node_table_empty(reducible_nodes) == 0) {

			/* Suche in der Menge der reduzierbaren Endknoten nach einem
			 * Knoten, für den ein passender Pfad existiert.
			 */
			suitable_node = rtsearch(Reduce);
			/* Falls das Suchergebnis leer ist, können die aktuell vorhandenen
			 * reduzierbaren Endknoten nicht zu einer akzeptierenden
			 * Berechnung korrespondieren und können aus dem Graphen entfernt
			 * werden.
			 */
			if (suitable_node == NULL) {
				DEBUG_PRINTF(
						"  Kein gültiger reduzierbarer Endknoten gefunden!\n");

				// Alle reduzierbaren Endknoten werden entfernt:
				while (node_table_empty(reducible_nodes) == 0) {
					kill_node = node_table_pop(reducible_nodes);
					graph_adjust(&kill_node);
				}
			} // if (suitable_node == NULL)

			/* Falls ein reduzierbarer Endknoten mit passendem Pfad gefunden
			 * wurde, muss die entsprechende Reduktion sofort durchgeführt
			 * werden.
			 */
			else if ( items[suitable_node->content].dot_symbol_type == Epsilon ) {
				DEBUG_PRINTF(
						"  Gültiger reduzierbarer Endknoten %ld wurde gefunden!\n",
						suitable_node->content);

				/* Vermerke die Reduktion im Wald der partiellen Ableitungsbäume:
				 */
				PARSE_FOREST_REDUCE(&parse_forest, suitable_node->content);

				// Alle lesbaren Endknoten müssen entfernt werden:
				while (node_table_empty(readable_nodes) == 0) {
					kill_node = node_table_pop(readable_nodes);
					graph_adjust(&kill_node);
				}

				/* Alle reduzierbaren Endknoten ausser suitable_node werden
				 * entfernt:
				 */
				while (node_table_empty(reducible_nodes) == 0) {
					kill_node = node_table_pop(reducible_nodes);
					if (kill_node == suitable_node)
						kill_node = NULL;
					else
						graph_adjust(&kill_node);
				}

				/* Der Graph kann nun auch neue expandierbare Endknoten enthalten,
				 * die ebenfalls entfernt werden müssen:
				 */
				while (node_table_empty(expandable_nodes) == 0) {
					kill_node = node_table_pop(expandable_nodes);
					graph_adjust(&kill_node);
				}

				/* Betrachte den Nichtterminalknoten A, der im Graph einziger
				 * direkter Vorgänger von suitable_node ist, und entferne die
				 * Kante zwischen den beiden:
				 */
				struct node_t *nt_node = 
					node_table_pop(suitable_node->predecessors);
				node_table_remove(nt_node->successors, suitable_node);

				/* Falls ein Itemknoten mehr als einen Vorgänger hatte, liegt ein
				 * Programmierfehler vor:
				 */
				assert(node_table_empty(suitable_node->predecessors) != 0);

				// Entferne den reduzierbaren Knoten aus dem Graphen:
				node_table_free(suitable_node->predecessors);
				node_table_free(suitable_node->successors);
				free(suitable_node);
				suitable_node = NULL;

				/* Nun müssen noch die bisherigen Vorgängerknoten von A behandelt
				 * werden, da aus ihnen die neuen Endknoten des Graphen entstehen.
				 */
				treat_predecessors(&nt_node);

			} // else if (Epsilon)

			/* Ansonsten ist das Suchergebnis ungültig, denn wenn die
			 * Suche auf lesbare Endknoten eingeschränkt wurde, darf das
			 * Ergebnis nur ein Knoten mit dot_symbol_type == Epsilon sein!
			 */
			else {
				fprintf(stderr, "treat_predecessors: Falscher Knoten wurde gefunden!\n");
				abort();
			} // else if (Epsilon) ... else

		} // if (node_table_empty(reducible_nodes) == 0)

		DEBUG_PRINTF("\nWurzelknoten des Graphen nach Ende der Phase:\n");
		LIST_PRINT(graph);
	
#ifdef PARSE_TREE_PHASES
		node_number++;
		printf("\n	{ %ld [label=\"Parsewald nach Phase %ld\"] }\n",
			node_number, phase_number);
		PARSE_FOREST_PRINT(parse_forest);
#endif

		return 0;

	} // if (Epsilon) ... else

} // int reduce_read


/** \brief Rahmenprozedur für den Parser.
 *
 *	Die Funktion \c yyparse erwartet, dass eine Funktion \c yylex vorhanden
 *	ist, die bei jedem Aufruf ein Token in seiner Integerrepräsentation
 *	liefert.
 *	Für eine Eingabefolge x prüft \c yyparse, ob x Teil der Sprache L(G)
 *	ist, die durch die Grammatik G beschrieben wird, die dem Parser
 *	zugrunde liegt.
 *	
 *	\return \c 0, falls \f$ x \in L(G)\f$;
 *		\c 1, falls die Eingabe fehlerhaft ist
 *		(wie Bison).
 */
static
int
yyparse (void)
{
	DEBUG_PRINTF("Betrete Funktion yyparse\n");
#ifdef PARSE_TREE
	atexit(End);
#endif
	// Initialisierung der globalen Knotentabellen:
	DEBUG_PRINTF("Initialisiere globale Knotentabellen... ");
	expandable_nodes      = node_table_init();
	readable_nodes        = node_table_init();
	reducible_nodes       = node_table_init();
	expanded_nodes        = node_table_init();
	expanded_nonterminals = node_table_init();
	DEBUG_PRINTF("erledigt!\n");

	unsigned long i; // Universelle Laufvariable

	// Flag für Erfolg eines Reduktions-/Leseschrittes
	int reduce_read_ok = 0; 

	DEBUG_PRINTF("Fülle Lookahead-Array... ");
	/* Fülle den Eingabepuffer mit den ersten LOOKAHEAD_LENGTH Token,
	 * die yylex liefert.
	 * Sollte die Eingabe vorher enden, liefert yylex für alle weiteren
	 * Aufrufe den Wert 0.
	 */
	for (i = 0; i < LOOKAHEAD_LENGTH; i++) {
		lookahead[i] = call_lexer();
	}
	DEBUG_PRINTF("erledigt!\n");

	DEBUG_PRINTF("Initialisiere Graph...\n");
	/* Zu Beginn der Programmausführung ist das einzige Element des Graphen 
	 * der Knoten für das expandierbare Item [S' -> .S]_0.
	 */
	DEBUG_PRINTF("  Erzeuge Knoten für Startitem [S' -> .S]\n");
	struct node_t *startnode =
		create_node((signed long) startitem_number);
	DEBUG_PRINTF("  Trage Startknoten in Graph ein\n");
	list_add (&graph, startnode);
	DEBUG_PRINTF("  Trage Startknoten in Tabelle expandierbarer Knoten ein\n");
	node_table_add(expandable_nodes, startnode);
	DEBUG_PRINTF("...erledigt!\n");

#ifdef PARSE_TREE
	node_number = 0;
	printf("digraph {");
#endif

	/* Iteriere abwechselnd Expansions- und Reduktions-/Leseschritte, bis
	 * alle Zeichen der Eingabe verbraucht sind oder in einem Reduktions-/
	 * Leseschritt vorzeitig erkannt wird, dass die Eingabe ungültig ist.
	 * Dabei gilt die Eingabe als verbraucht, sobald ein Token mit dem Wert
	 * 0 gefunden wird.
	 */
	DEBUG_PRINTF("Betrete Parserschleife!\n\n");
	reduce_read_ok = 0;
	while ( (lookahead[phase_number % LOOKAHEAD_LENGTH] > 0) &&
			(reduce_read_ok == 0) ) {
		DEBUG_PRINTF("\n");

		DEBUG_PRINTF("Phase %2ld: Lookahead = ", (phase_number + 1));
		for (i=0; i<LOOKAHEAD_LENGTH; i++) {
			DEBUG_PRINTF("%ld, ",
				lookahead[ (phase_number+i) % LOOKAHEAD_LENGTH ]);
		}
		DEBUG_PRINTF("\n");

		DEBUG_PRINTF("Phase %2ld: Expansionsschritt\n", (phase_number + 1));
		expand();

		DEBUG_PRINTF("Phase %2ld: Reduktions-/Leseschritt\n",
			(phase_number + 1));
		reduce_read_ok = reduce_read();
	} // while (lookahead[phase_number % LOOKAHEAD_LENGTH])	

	/* Teste, ob die Parserschleife regulär beendet wurde oder aufgrund
   * eines fehlgeschlagenen Reduktions-/Leseschrittes:
	 */
	if (reduce_read_ok == 0) {
		/* Falls die Eingabe verbraucht ist und der Graph das Item [S' -> S.]_n
		 * enthält: akzeptiere.
		 * Sonst: Beende mit Fehler.
		 *
		 * Wenn der Parsevorgang erfolgreich war, wurde eine der
		 * Zusammenhangskomponenten von \G vollständig auf diesen einen Knoten
		 * reduziert. Es genügt also, die Liste aller Anfangsknoten des Graphen
		 * nach diesem Knoten zu durchsuchen!
		 */
		if (list_contains_node_num(graph, acceptitem_number)) {
#ifndef SUPPRESS_OUTPUT
			fprintf(stderr,
					"\nDie Eingabe wurde erfolgreich geparst! (%ld Token)\n",
					phase_number);
#endif
			clean();

/* Drucke den finalen Ableitungsbaum nur dann aus, wenn nicht schon am
 * Ende jeder Phase der aktuelle Wald ausgegeben wird:
 */			
#ifndef PARSE_TREE_PHASES
			PARSE_FOREST_PRINT(parse_forest);
#endif			
			/** \todo Ist eine clean-Routine auch für den Ableitungswald nötig?
			 *  Eigentlich nicht, weil diese Funktionalität ja nur zu
			 *  Diagnosezwecken eingebaut wurde und im Produktiveinsatz zwecks
			 *  Zeitersparnis nicht aktiviert würde.
			 */
			 return 0;
		}
		else {
			fprintf(stderr,
					"\nFEHLER bei Symbol Nr. %ld: Die Eingabe ist ungültig!\n",
				phase_number);
			clean();
			return 1;
		}
	}
	else {
		fprintf(stderr,
				"\nVorzeitiger Abbruch der Parserschleife bei Symbol %ld: %ld!\n",
				phase_number,
				lookahead[ (phase_number+i) % LOOKAHEAD_LENGTH ]);
		return 1;
	}

} // static int yyparse (void)



