(**************************************************************************
 * file bogenlr.ml                                                        *
 **************************************************************************)

(** Hauptprogramm, das die Funktionen zum Einlesen der Eingabegrammatik
 *	und zur Erzeugung des C-Codes fuer den Parser aufruft.
 *
 *	Funktionsweise:
 *  - Einlesen der Grammatikspezifikation aus einer Datei, deren Name dem
 *    Programm als Kommandozeilenparameter vorliegt.
 *  - Einlesen der Grammatik und Erzeugen einer internen Repraesentation
 *    mittels der Module BogenLexer und BogenParser.
 *  - Durchfuehrung von Tests aus dem Modul BogenGrammar auf der Grammatik,
 *    um sie auf einige notwendige Eigenschaften fuer LR(k)-Grammatiken
 *    zu ueberpruefen.
 *  - Generierung des eigentlichen Parsers mit dem Modul BogenGenerator.
 *  - Ausgabe des C-Code fuer den erzeugten Parser sowie von Header und
 *    Trailer aus der Eingabedatei in die Ausgabedatei.
 *
 *  Dabei greift das Programm [bogenlr] auf folgende Module zurueck:
 *  - BogenGrammar:
 *    Stellt die Datenstrukturen fuer eine interne Repraesentation der
 *    Grammatik zur Verfuegung, zusammen mit grundlegenden Operationen
 *    darauf.
 *  - BogenLexer:
 *    Scanner, der aus der Eingabedatei eine Folge von Token erzeugt und
 *    eine Fehlermeldung ausgibt, falls die Eingabe nicht lexikalisch
 *    korrekt ist.
 *    bogenLexer.ml wird mit [ocamllex] aus der Spezifikationsdatei
 *    bogenLexer.mll automatisch generiert.
 *  - BogenParser:
 *    Ueberfuehrt die Eingabegrammatik, sofern sie syntaktisch korrekt
 *    ist, in eine interne Datenstruktur vom Typ [BogenGrammar.t].
 *    bogenParser.ml wird mit [ocamlyacc] aus der Spezifikationsdatei
 *    bogenParser.mly automatisch generiert.
 *)


(* Zugriff auf Speicherverwaltungsfunktionen, um bei einem Ueberlauf des
 * Stacks seine Groesse auszugeben.
 *)
open Gc


(*************************************************************************
 * Exceptions:                                                           *
 *************************************************************************)

(** Auf der Kommandozeile muss genau eine Eingabedatei spezifiziert
 *  werden.
 *)
exception Wrong_number_of_inputfiles of int


(*************************************************************************
 * Globale Deklarationen:                                                *
 *************************************************************************)

(** Stringreferenz fuer den Namen der Eingabedatei. *)
let in_filename =
  ref ""


(** Stringreferenz fuer den Namen der Ausgabedatei.
 *  Defaultwert: "bogenout.c"
 *)
let out_filename =
  ref "bogenout.c"


(** Flag, ob die Option "-r" auf der Kommandozeile angegeben wurde: *)
let option_r = ref false


(** Definition der erlaubten Kommandozeilen-Optionen. *)
let cmdline_options =
  Arg.align [
    ("-d", Arg.String (fun s -> begin
        BogenGenerator.option_d := true;
        BogenGenerator.token_filename := s
      end),
      " Erzeugt Datei mit angegebenem Namen mit Tokentypen (fuer Lex).");
    ("-l", Arg.Int (fun i -> BogenGenerator.lookahead_length := i),
      " Legt die Lookahead-Länge fest (Default: 10)");
    ("-o", Arg.String (fun s -> out_filename := s),
      " Legt den Namen der Ausgabedatei fest (Default: bogenout.c)");
    ("-r", Arg.Unit (fun _ -> option_r := true),
      " Schreibt Report zu Grammatik und Parser nach \"<eingabedatei>.output\".")
  ]


(** Fehlermeldung bei ungueltiger Kommandozeile oder Aufruf mit Option
 *  "-help" oder "--help": *)
let usage_msg =
  "Beim Aufruf des Programms muss genau eine Eingabedatei angegeben werden."
  ^ "\nGültige Optionen sind:"


(** Zaehler fuer die Anzahl bisher gelesener anonymer Parameter auf der
 *  Kommandozeile.
 *)
let anon_count = ref 0


(**************************************************************************
 * Funktionen:                                                            *
 **************************************************************************)

(** Die Funktion [anon_fun s] verarbeitet anonyme Kommandozeilenparameter,
 *  die also nicht an eine Option gebunden sind.
 *  Davon muss es bei einem korrekten Aufruf genau einen geben,
 *  naemlich den, der den Namen der Eingabedatei bestimmt.
 *  @param s String mit dem Namen der Eingabedatei.
 *  @return [()], da im Erfolgsfall die Stringreferenz [in_filename] auf
 *    den Wert [s] festgelegt wird.
 *)
let anon_fun s =
begin
  anon_count := !anon_count + 1;
  if !anon_count > 1 then
    raise (Wrong_number_of_inputfiles !anon_count)
  else
    in_filename := s
end


(** [in_channel x] stellt einen Eingabekanal fuer die Datei mit dem Namen
 *  [x] bereit.
 *  Wenn es keine Datei mit dem angegebenen Namen gibt oder sie aus
 *	anderen Gruenden nicht geoeffnet werden kann, beendet sich das
 *  Programm mit einer Fehlermeldung und dem Fehlercode 1.
 *	@param x String mit dem Namen der zu oeffnenden Eingabedatei.
 *	@return Eingabekanal.
 *)
let in_channel x =
  if x = "" then
    raise (Wrong_number_of_inputfiles 0);
  try
    open_in x
  with _ ->
    begin
      print_endline "FEHLER:";
      print_endline
        ("Eingabedatei \"" ^ x ^ "\" konnte nicht geöffnet werden!");
      exit 1
    end


(** [out_channel x] stellt einen Ausgabe-Kanal fuer die Datei mit dem Namen
 *  x bereit.
 *  Wenn die Datei bereits existiert, wird sie geloescht und als leere
 *	Datei neu erstellt. Wenn die Datei nicht geoeffnet werden kann,
 *  beendet sich das Programm mit einer Fehlermeldung und dem Fehlercode 1.
 *	@param x String mit dem Namen der zu oeffnenden Zieldatei.
 *	@return Ausgabekanal.
 *)
let out_channel x =
  try
    if Sys.file_exists x
    then Sys.remove x;
    open_out x
  with
    Sys_error e -> begin
      print_endline "bogenlr: FEHLER:";
      print_endline
        ("Ausgabedatei \"" ^ x ^ "\" konnte nicht geöffnet werden:");
      print_endline e;
      exit 1
    end


(** Die Funktion [in_grammar ic] ruft Scanner und Parser auf, um die
 *  Grammatikspezifikation aus dem Eingabekanal [ic] in eine interne
 *  Datenstruktur zu ueberfuehren. Sollte die Eingabedatei enden, bevor
 *  eine gueltige Grammatikspezifikation eingelesen wurde, beendet sich
 *  das Programm mit einer Fehlermeldung.
 *	@param ic Eingabekanal.
 *	@return Interne Grammatikdarstellung vom Typ [BogenGrammar.t].
 *)
let in_grammar ic =
  try
    let lexbuf = Lexing.from_channel ic
    in BogenParser.grammar BogenLexer.token lexbuf
  with
		End_of_file -> 
			begin
        print_endline "bogenlr: FEHLER:";
				print_endline "Vorzeitiges Ende der Eingabedatei!";
        close_in ic;
        exit 1
      end
    | e -> raise e


(** Die Funktion [output_header ic oc h_start h_end] kopiert den Headercode
 *  von Position [h_start] bis [h_end] unveraendert aus dem Quellkanel [ic]
 *  in den Zielkanal [oc].
 *	@param ic Eingabekanal.
 *	@param oc Ausgabekanal.
 *	@param h_start Startposition des Headers im Eingabekanal.
 *	@param h_end Endposition des Headers im Eingabekanal.
 *	@return [()]
 *)
let output_header ic oc h_start h_end =
begin
  output_string oc "\n/* ===== Begin of header section ===== */\n";
  seek_in ic h_start;
  while (pos_in ic < h_end) do
    let c = input_char ic in
      output_char oc c
  done;
  output_string oc "\n/* ===== End of header section ===== */\n\n";
  flush oc
end


(** Die Funktion [output_trailer ic oc t_start] kopiert den Trailercode von
 *  Position [t_start] bis zum Dateiende unveraendert aus	dem Quellkanal
 *  [ic] in den Zielkanal [oc].
 *	@param ic Eingabekanal.
 *	@param oc Ausgabekanal.
 *	@param t_start Startposition des Trailers im Eingabekanal.
 *	@return [()]
 *)
let output_trailer ic oc t_start =
begin
	output_string oc  "\n/* ===== Begin of trailer section ===== */\n";
	seek_in ic t_start;
	let eof = ref false in
	begin
		try
			while not !eof do
				let c = input_char ic in
					output_char oc c
			done
		with
			End_of_file ->
				output_string oc "\n/* ===== End of trailer section ===== */\n";
				eof := true;
				flush oc
	end;
end
 		

(** Ausgabe eines Reports mit Informationen zur Eingabegrammatik und zum
 *  erzeugten Parser:
 *  @param ig Interne Grammatikrepraesentation vom Typ [BogenGrammar.t].
 *  @return [()]
 *)
let output_report ig =
  let filename = !in_filename ^ ".output"

  in let oc2 =
    try
      if Sys.file_exists filename
      then Sys.remove filename;
      open_out filename;
    with
      Sys_error e ->
        begin
          print_endline "bogenlr: FEHLER:";
          print_endline
            ("Reportdatei " ^ filename ^
            " konnte nicht geschrieben werden:");
          print_endline e;
          exit 1
        end

  in begin
    output_string oc2
      ("\n" ^ BogenGrammar.string_of_grammar ig ^ "\n");
    flush oc2;
    close_out oc2;
  end


(** Die Hauptfunktion [main ()] ruft die Funktionen zum Einlesen und
 *  Ueberpruefen der Eingabegrammatik sowie zum Generieren und zur
 *  Ausgabe des Resultatparsers auf.
 *)
let main () =
  (* Verarbeiten der Eingabeparameter und Oeffnen der Ein- und
   * Ausgabekanaele:
   *)
  let _ = Arg.parse cmdline_options anon_fun usage_msg
  and _ = if !anon_count = 0 then
    raise (Wrong_number_of_inputfiles !anon_count)
  and ic = in_channel !in_filename
  and oc = out_channel !out_filename

  (* Einlesen der Eingabegrammatik: *)
  in let intern_grammar = in_grammar ic

  in try
    begin
      (* Test der Grammatik auf einige fuer LR(k) notwendige Kriterien: *)
	  	BogenGrammar.test intern_grammar;

      (* Ausgabe des Headers aus der Eingabedatei in die Zieldatei: *)
  		output_header ic oc
			  intern_grammar.BogenGrammar.header_start
		  	intern_grammar.BogenGrammar.header_end;

 	  	(* Erzeugen des Parsers und Ausgabe in die Zieldatei: *)
  		BogenGenerator.create_parser intern_grammar oc;

		  (* Ausgabe des Trailers aus der Eingabedatei in die Zieldatei: *)
	  	output_trailer ic oc intern_grammar.BogenGrammar.trailer_start;

      (* Schliessen der Kanaele und Programmende: *)
		  close_in ic;
	  	close_out oc;
  		print_endline ("bogenlr: Parser wurde in Datei " ^
	  								 !out_filename ^ " geschrieben.");

      (* Falls die Option "-r" auf der Kommandozeile uebergeben wurde:
       * Generiere Datei "<eingabedatei>.output" mit Informationen zu
       * Grammatik und generiertem Parser.
       *)
      if !option_r then begin
         output_report intern_grammar;
        print_endline("bogenlr: Report wurde in Datei " ^ !in_filename ^
        ".output geschrieben.");
      end;
  
      print_endline ("bogenlr: Fertig!")
    end

  with e ->
    begin
      close_in ic;
      close_out oc;
      raise e
    end


(** Beim Aufruf des Programms [bogenlr], fuehre die Funktion [main] aus:
 *)
let _ =
  try
(*    Gc.set { (Gc.get()) with Gc.stack_limit = 1073741824 }; *)
    main ()
  with
  | Wrong_number_of_inputfiles i->
      begin
        print_endline("Falsche Anzahl von Eingabedateien: "
          ^ (string_of_int i) ^ "\n");
        Arg.usage cmdline_options usage_msg
      end;
  | BogenGrammar.No_rule_for_startsymbol s ->
      begin
        print_endline("bogenlr: FEHLER:");
        print_endline("Keine Regel für Startsymbol " ^ s);
        exit 1
      end;
  | BogenGrammar.Unreachable_nonterminal ->
      begin
        print_endline("bogenlr: FEHLER:");
        print_endline("Nicht alle Nichtterminalsymbole sind erreichbar!");
        exit 1
      end;
  | BogenGrammar.Undefined_nonterminal ->
      begin
        print_endline("bogenlr: FEHLER:");
        print_endline("Nichtterminalsymbol ohne definierende Regel!");
        exit 1
      end;
  | Stack_overflow ->
      begin
        print_endline (string_of_int ((Gc.get ()).stack_limit));
        exit 2
      end;
  | e ->
      raise e
